/* Copyright (c) 1998 L.W.McVoy */
#include "sccs.h"
WHATSTR("@(#)admin.c 1.24");

private	char	*help = "\n\
usage: admin options [-] OR [file file file...]\n\n\
    A very useful thing to note is that\n\
	    find src -name 's.*' -print | admin -what -ever -\n\
    works.  Note the trailing \"-\" which says to read files from stdin.\n\n\
    In general, the options which add things take an upper case\n\
    option which deletes that same thing.\n\n\
    -n			create a new SCCS file\n\
    -i<file>		initial text in <file> or stdin for -n\n\
    -r<rev>		initial release for new file\n\
    -y<com>		initial checkin comment\n\n\
    -a<u>|<g>		add user/group\n\
    -A<u>|<g>		delete user/group\n\n\
    -f<f><val>		add flag (value is optional)\n\
    -F<f>		delete flag\n\n\
    -p<path>:<rev>	set the pathname for rev\n\
    -P<rev>		delete the pathname for rev\n\n\
    -s<sym>:<rev>	set symbol associated with rev (rev can be a release)\n\
    -s<sym>=<rev>	set symbol, replacing any existing symbol\n\
    -S<sym>		delete the symbol\n\n\
    -t<file>		set or (if no file) delete descriptive text\n\
    -T			delete descriptive text\n\n\
    -b			force binary encoding (uuencode)\n\
    -g			force gzipped binary encoding (gzip | uuencode)\n\
    -h			check s.file format for correctness\n\
    -q			run quietly\n\
    -z			recompute checksum\n\n\
    -e<u>|<g>	  	delete user/group (ATT compat)\n\
    -d<f>		delete flag (ATT compat)\n\
    -l<a|r,r>		unlock releases (not implemented)\n\
    -m<mrlist>		MR lists (not implemented)\n\n";

#define	OP(W, V, F) if (next##W < A_SZ-1) { \
			W[next##W].thing = V; \
			W[next##W++].flags = F; \
		  } else { \
		  	fprintf(stderr, "admin: argument overflow.\n"); \
			exit(1); \
		  }

int	do_checkin(char *nm, int enc, int fl, char *rev, char *newf, char *com);

int
main(int ac, char **av, char **ev)
{
	sccs	*sc;
	int	flags = 0;
	char	*rev = 0;
	int	c;
	admin	f[A_SZ], u[A_SZ], s[A_SZ], p[A_SZ];
	int	nextf = 0, nextu = 0, nexts = 0, nextp = 0;
	char	*comment = 0, *text = 0, *newfile = 0;
	char	*name;
	int	encoding = E_ASCII, error = 0;
	int	fastSym;

	debug_main(av);
	if (ac > 1 && streq("--help", av[1])) {
		fprintf(stderr, "%s", help);
		return(1);
	}
	bzero(f, sizeof(f));
	bzero(u, sizeof(u));
	bzero(s, sizeof(s));
	bzero(p, sizeof(p));
	while ((c =
	    getopt(ac, av, "a;A;e;f;F;d;i|np;P;r;y|s;S;Tt|bghqz")) != -1) {
		switch (c) {
		/* user|group */
		    case 'a':	OP(u, optarg, A_ADD); break;
		    case 'A':
		    case 'e':	OP(u, optarg, A_DEL); break;
		/* flags */
		    case 'f':	OP(f, optarg, A_ADD); break;
		    case 'F':	
		    case 'd':	OP(f, optarg, A_DEL); break;
		/* new file options */
		    case 'i':	newfile = optarg ? optarg : "-";
				flags |= NEWFILE; break;
		    case 'n':	flags |= NEWFILE; break;
		    case 'r':	rev = optarg; break;
		    case 'y':	comment = optarg; break;
		/* MR lists - XXX */
		    case 'm':	fprintf(stderr,
				    "admin: MR's not implemented.\n");
				exit(1);
		/* path */
		    case 'p':	OP(p, optarg, A_ADD); break;
		    case 'P':	OP(p, optarg, A_DEL); break;
		/* symbols */
		    case 's':	OP(s, optarg, A_ADD); break;
		    case 'S':	OP(s, optarg, A_DEL); break;
		/* text */
		    case 't':	text = optarg ? optarg : ""; break;
		    case 'T':	text = ""; break;
		/* singletons */
		    case 'b':	encoding = E_UUENCODE; break;
		    case 'g':	encoding = E_UUGZIP; break;
		    case 'h':	flags |= CHECKFILE|INHERIT_DATE; break;
		    case 'q':	flags |= SILENT; break;
		    case 'z':	flags |= NOCKSUM|NEWCKSUM; break;
		    default:	fprintf(stderr, "admin: bad option %c.\n", c);
		    		goto usage;
	    	}
	}
	if ((flags & CHECKFILE) && 
	    ((flags & ~(CHECKFILE|SILENT|INHERIT_DATE))
	    || nextf || nextu || nexts || nextp || rev)) {
		fprintf(stderr, "admin: -h option must be alone.\n");
		goto usage;
	}
	if (comment && !(flags & NEWFILE)) {
		fprintf(stderr,
		    "admin: comment may only be specifed with -i and/or -n\n");
		goto usage;
	}
	if (rev && !(flags & NEWFILE)) {
		fprintf(stderr,
		    "admin: revision may only be specifed with -i and/or -n\n");
		goto usage;
	}
	if ((flags & NEWFILE) && text && !text[0]) {
		fprintf(stderr,
		    "admin: -t must have file arg with -i and/or -n\n");
		goto usage;
	}
	name = sfileFirst("admin", &av[optind], 0);
	if ((flags & NEWFILE) && sfileNext()) {
		fprintf(stderr, "admin: Only one file with -i/-n\n");
		goto usage;
	}
	/*
	 * If we are adding exactly one symbol, do it quickly.
	 */
	fastSym = !(flags & ~SILENT) && !nextf && !nextu && !nextp
	    && !rev && nexts && (s[0].flags == A_ADD) && !s[1].flags;
	if (fastSym) flags |= (MAP_WRITE|NOCKSUM);
	while (name) {
		if (flags & NEWFILE) {
			if (do_checkin(name, encoding, flags&(SILENT|NEWFILE),
			    rev, newfile, comment)) {
				error  = 1;
				name = sfileNext();
				continue;
			}
		}
		sc = sccs_init(name, flags);
		unless (sc) { name = sfileNext(); continue; }
		unless (sc->tree) {
			fprintf(stderr, "admin: can't read delta table in %s\n",
			    sc->sfile);
			sccs_free(sc);
			name = sfileNext();
			error = 1;
			continue;
		}
#ifndef	USE_STDIO
		unless (fastSym &&
		    sc->landingpad && !sccs_addSym(sc, flags, s[0].thing)) {
#endif
			if (sccs_admin(sc, flags, f, u, s, p, text)) {
				unless (BEEN_WARNED(sc)) {
					fprintf(stderr, "admin of %s failed.\n",
					    sc->gfile);
				}
				error = 1;
			}
#ifndef	USE_STDIO
		}
#endif
		sccs_free(sc);
		name = sfileNext();
	}
	sfileDone();
	purify_list();
	return(error);
usage:	fprintf(stderr, "admin: usage error, try `admin --help' for info.\n");
	return (1);
}

/*
 * Handle -n.  Allowable options are:
 *	-i (imples -n)
 *	-r
 *	-y
 */
int
do_checkin(char *name, int encoding,
	int flags, char *rev, char *newfile, char *comment)
{
	delta	*d = 0;
	sccs	*s;
	int	error;

	unless (s = sccs_init(name, flags)) { return (-1); }
	s->encoding = encoding;
	if (HAS_SFILE(s)) {
		fprintf(stderr, "admin: %s exists.\n", s->sfile);
		sccs_free(s);
		return (1);
	}
		
	/*
	 * If they specified a gfile and it is different than the implied
	 * gfile, and the implied gfile exists, that's too weird, fail it.
	 */
	if ((s->state & GFILE) && newfile && strcmp(newfile, s->gfile)) {
		fprintf(stderr,
		    "admin: gfile %s exists and is "
		    "different than specified init file %s\n",
		    s->gfile, newfile);
		sccs_free(s);
		return(-1);
	}
	if (rev) {
		d = sccs_parseArg(d, 'R', rev);
		if ((d->flags & BADREV) ||
		    (d->r[0] && !d->r[1]) || (d->r[2] && ! d->r[3])) {
			sccs_freetree(d);
			fprintf(stderr, "admin: bad revision: %s for %s\n",
			    rev, s->sfile);
			sccs_free(s);
			return(-1);
		}
	}
	if (newfile) {
		free(s->gfile);
		s->gfile = strdup(newfile);
	} else {
		flags |= EMPTY;
	}
	s->state |= GFILE;
	if (comment) { d = sccs_parseArg(d, 'C', comment); }
	if ((error = sccs_delta(s, flags|SAVEGFILE, d, 0))) {
		unless (BEEN_WARNED(s)) {
			fprintf(stderr, "admin: failed to check in %s.\n",
			    s->sfile);
		}
	}
	sccs_free(s);
	return (error);
}
