/* Copyright (c) 1997 L.W.McVoy */
#include "sccs.h"
WHATSTR("@(#)delta.c 1.39");
char	*delta_help = "\n\
usage: delta [-cilnpqs] [-I<f>] [-R<r>] [-y<c>] [files...]\n\n\
    -c		Skip the checksum generation (not advised)\n\
    -i		Initial checkin, create a new revision history\n\
    -l		Follow checkin with a locked checkout like ``get -e''\n\
    -n		Retain the edited g-file, which is normally deleted\n\
    -p		Print differences\n\
    -q		Run silently\n\
    -s		Run silently\n\
    -I<file>	Take the initial date/user/comments/etc from <file>\n\
    		See prs for file format information\n\
    -R<rev>	Force the new revision number to be <rev>.  Use carefully.\n\
    -y<comment>	Sets the revision comment to <comment>.\n\n";

/*
 * Not implemented:
 *  -g<revs>	Specify  a  list of deltas to omit.
 *  -m<mrList>	If the 'v' flag is set, then you must specify an MR number
 *  		with the checkin (MR == modification request number).
 *  -r<rev>
 *
 * TODO -
 *	support ~e escape for dropping into the editor on the comments so far.
 */

/*
 * delta 
 */
int
main(int ac, char **av)
{
	sccs	*s;
	int	flags = FORCE|BRANCHOK;
	int	c;
	char	*rev = 0;
	char	*initFile = 0;
	char	*comment = 0;
	char	*name;

	debug_main(av);
	if (ac > 1 && streq("--help", av[1])) {
help:		fprintf(stderr, delta_help);
		return(1);
	}
	while ((c = getopt(ac, av, "cg;I;ilm;npqR;sy|")) != -1) {
		switch (c) {
		    /* SCCS flags */
		    case 'g': fprintf(stderr, "-g Not implemented.\n");
		    	      goto help;
		    case 'm': fprintf(stderr, "-m Not implemented.\n");
		    	      goto help;
		    case 'n': flags |= SAVEGFILE; break;
		    case 'p': flags |= PRINT; break;
		    case 'r': fprintf(stderr, "-r Not implemented.\n");
		    	      goto help;
		    case 's': flags |= SILENT; break;
		    case 'y': comment = optarg; flags |= DONTASK; break;

		    /* LM flags */
		    case 'c': flags |= NOCKSUM; break;
		    case 'I': initFile = optarg; break;
		    case 'i': flags |= NEWFILE; break;
		    case 'l': flags |= SKIPGET|SAVEGFILE|EDIT; break;
		    case 'q': flags |= SILENT; break;
		    case 'R': flags |= USEREV; rev = optarg; break;

		    default:
usage:			fprintf(stderr, "delta: usage error, try --help.\n");
			return (1);
	    	}
	}
	name = sfileFirst("delta", &av[optind], 0);
	/* They can only have an initFile for one file...  
	 * So we go get the next file and error if there
	 * is one.
	 */
	if (initFile && name && sfileNext()) {
		fprintf(stderr, "delta: only one file "
		    "may be specified with init file.\n");
		goto usage;
	}
	if (initFile && (flags & DONTASK)) {
		fprintf(stderr,
		    "delta: only init file or comment, not both.\n");
		goto usage;
	}

	while (name) {
		delta	*d = 0;

		if (flags & DONTASK) {
			d = sccs_parseArg(d, 'C', comment);
		}
		if (d && (d->flags & DERROR)) {
			sccs_freetree(d);
			goto usage;
		}
		unless (s = sccs_init(name, flags)) {
			if (d) sccs_freetree(d);
			name = sfileNext();
			continue;
		}
		if (rev) {
			if (!d) d = calloc(1, sizeof(*d));
			d->rev = strdup(rev);
		}
		/* XXX - interrupting here leaves unfreed memory */
		if (sccs_delta(s, flags, d, initFile) == -1) {
			fprintf(stderr, "delta of %s failed.\n", s->gfile);
			sfileDone();
			purify_list();
			return(1);
		}
		if (flags & EDIT) {
			int	f = flags & (EDIT|SKIPGET|SILENT);

			unless (s = sccs_restart(s)) {
				fprintf(stderr,
				    "delta: can't restart %s\n", name);
				name = sfileNext();
				continue;
			}
			/* XXX - do the revs match?  What if I was not on
			 * TOT and I do this deledit - where am I? 
			 */
			if (sccs_get(s, 0, 0, 0, f, "-")) {
				unless (BEEN_WARNED(s)) {
					fprintf(stderr,
					"get of %s failed, skipping it.\n",
					name);
				}
			}
		}
		sccs_free(s);
		name = sfileNext();
	}
	sfileDone();
	purify_list();
	return(0);
}
