#include "sccs.h"
#ifdef	DIFF
int
diff(char *lfile, char *rfile, char kind, char *out)
{
	char	*av[6];
	int	status;
	int	i;

	if (fork()) {
		wait(&status);
		return (WEXITSTATUS(status));
	} else {
		if (strcmp("-", out)) {
#undef	open
#undef close
			close(1);
			open(out, O_CREAT|O_WRONLY, 0644);
		}
		av[i = 0] = DIFF;
		switch (kind) {
		    case D_CONTEXT: av[++i] = "-c"; break;
		    case D_UNIFIED: av[++i] = "-u"; break;
		    case D_DIFF: break;
		    default:
			fprintf(stderr,
			    "diff: unknown opt: %c (0x%x)\n", kind, kind);
			break;
		} 
		av[++i] = lfile; 
		av[++i] = rfile;
		av[++i] = 0;
		execve(DIFF, av, 0);
		perror(DIFF);
		exit(19);
	}
}
#else
/*
 * Miller/Meyers diff alg from their paper.
 *
 * With changes to limit memory consumption.
 * With diff style output.
 * Performance seems to be as good as GNU diff on almost everything.
 *
 * Copyright (c) 1997 by Larry McVoy; All rights reserved.
 *
 */
#include "sccs.h"
#define	min(a,b)	((a)<(b)?(a):(b))
#define	max(a,b)	((a)>(b)?(a):(b))
#define	INSERT		1
#define	DELETE		2
#define	eq(a,b)	sameline(l[a], r[b])

typedef struct {
	char	*data;		/* the file data */
	char	*end;		/* pointer after last byte */
	int	size;		/* file size in bytes (end - data) */
	char	**lines;	/* pointers to the heads of each line */
	int	n;		/* number of lines */
	int	start;	/* # of lines at the front that are ident */
} file_t;

typedef struct	edit {
	struct	edit *link;
	int	op;
	int	leftline, rightline;
} edit_t;

static	void put_script(int, int, char **, char **, edit_t *, FILE *);
static	void putline(char prefix, char *s, FILE *out);

static inline int
sameline(register char *left, register char *right)
{
	while (*left == *right && *left != '\n') {
		left++, right++;
	}
	return (*left == *right);
}

static inline void
putline(char prefix, char *s, FILE *out)
{
	assert(s);
	putc(prefix, out);
	putc(' ', out);
	do {
		putc(*s, out);
	} while (*s++ != '\n');
}

static int
init(char *file, file_t *f)
{
	int	fd;
	struct	stat sbuf;

	bzero(f, sizeof(file_t));
	if ((fd = open(file, 0, 0)) == -1) {
		perror(file);
		return (-1);
	}
	if (fstat(fd, &sbuf) == -1) {
		perror("fstat");
		return (-1);
	}
	f->size = sbuf.st_size;
	f->data = mmap(0, f->size, PROT_READ, MAP_SHARED, fd, 0);
	close(fd);
	if ((int)f->data == -1) {
		perror("mmap");
		return (-1);
	}
	return (0);
}

static void
uninit(file_t *f)
{
	if (f->data) {
		munmap(f->data, f->size);
	}
	if (f->lines) free(f->lines);
	f->lines = 0;
	f->data = 0;
}

static void
rows(file_t *f)
{
	char	*p;
	char	*end;
	int	l;

	end = f->end = f->data + f->size;

	/*
	 * Let's assume there are 20 chars per line (matches my source).
	 */
	f->n = f->size / 20;
	if (!f->n) f->n = 20;
	f->lines = malloc(f->n * sizeof(char*));

	for (l = 0, p = f->data; p < f->end; ) {
		if (l >= f->n) {
			char	*tmp;

			f->n *= 2;
			tmp = malloc(f->n * sizeof(char*));
			assert(tmp);
			bcopy(f->lines, tmp, l * sizeof(char*));
			free(f->lines);
			f->lines = (char **)tmp;
		}
		f->lines[l++] = p;
		while ((p < end) && (*p++ != '\n'));
		if (p == end) break;
	}
	f->n = l;
}

/*
 * Better now that I periodically reset.
 */
#define	N	10000
#define	ORIGIN	N
#define	MAXD	(N*2 + 1)
#define	HUNK	50

int
diff(char *lfile, char *rfile, char *output)
{
	int	m, n, e;
	int	lower, upper, d, k, row, col;
	int	mismatches;
	file_t	left, right;
	char	**l, **r;
	int	last_d[MAXD];
	edit_t	*script[MAXD];
	edit_t	edits[MAXD];
	FILE	*out;

	debug((stderr, "diff(%s, %s)\n", lfile, rfile));
	if (init(lfile, &left)) {
		uninit(&left);
		return (2);
	}
	if (init(rfile, &right)) {
		uninit(&left);
		uninit(&right);
		return (2);
	}

	if (strcmp(output, "-") == 0) {
		out = stdout;
	} else {
		out = fopen(output, "w");
	}

	/*
	 * Fast test for exactly same sized files.
	 */
	if (left.size == right.size) {
		if (!left.size || !bcmp(left.data, right.data, left.size)) {
			uninit(&left);
			uninit(&right);
			if (out != stdout) fclose(out);
			return (0);
		}
	}

	/*
	 * Well, gotta do the work so go get some pointers.
	 */
	rows(&left);
	rows(&right);

	/*
	 * This used to be simple - here's the deal.  It's common to have
	 * a wad of matching lines at the begining and end, so we went off
	 * and counted those in match.  We set l/r to point to the first
	 * line that doesn't match and m/n to the block of diffs in the 
	 * middle.  I was drunk when I wrote thish so beshware the goblinsh.
	 */
again:
	debug((stderr, "left: %d,%d    right: %d,%d\n",
	    left.start,left.n,
	    right.start,right.n));
	e = -1;	/* index into edits - replaces new in paper */
	l = &left.lines[left.start];
	mismatches = 0;
	r = &right.lines[right.start];
	m = left.n - left.start;
	m = min(m, N);
	n = right.n - right.start;
	n = min(n, N);
#if 0
	assert(left.start <= left.n);
	assert(left.start >= 0);
	assert(right.start <= right.n);
	assert(right.start >= 0);
	assert(l && r);
#endif
	if (m == 0 && n > 0) {		/* new stuff on the right */
		if (right.start < right.n - 1) {
			fprintf(out, "%da%d,%d\n",
			    left.n, right.start+1, right.n);
		} else {
			fprintf(out, "%da%d\n", left.n, right.start+1);
		}
		for (n = right.start; n < right.n; ++n) {
			putline('>', right.lines[n], out);
		}
		uninit(&left);
		uninit(&right);
		if (out != stdout) fclose(out);
		return (1);
	} else if (n == 0 && m > 0) {	/* deleted stuff on the left */
		if (left.start < left.n - 1) {
			fprintf(out, "%d,%dd%d\n",
			    left.start+1, left.n, right.n);
		} else {
			fprintf(out, "%dd%d\n", left.start+1, right.n);
		}
		for (n = left.start; n < left.n; ++n) {
			putline('<', left.lines[n], out);
		}
		uninit(&left);
		uninit(&right);
		if (out != stdout) fclose(out);
		return (1);
	}
	for (row = 0; row < m && row < n; row++) {
		if (!eq(row,row)) break;
	}
	lower = (row >= m) ? ORIGIN + 1 : ORIGIN - 1;
	upper = (row >= n) ? ORIGIN - 1 : ORIGIN + 1;
	debug((stderr, "m=%d n=%d ", m,n));
	debug((stderr, "low=%d up=%d row=%d\n", lower, upper, row));
    	if (lower > upper) {
		if (row + left.start == left.n &&
		    row + right.start == right.n) {
			uninit(&left);
			uninit(&right);
			if (out != stdout) fclose(out);
			return (0);	/* no diffs */
		}
		left.start += row;
		right.start += row;
		goto again;
	}
	last_d[ORIGIN] = row;
	script[ORIGIN] = 0;

	for (d = 1; d <= MAXD; d++) {
		for (k = lower; k <= upper; k += 2) {
			int	matches = 0;

			e++;
			/*
			 * LMXXX - gcc doesn't like this, wants parens.
			 * Check paper to see what they should be.
			 */
			if ((k == ORIGIN-d) || (k != ORIGIN+d) &&
			    (last_d[k+1] >= last_d[k-1])) {
			    	row = last_d[k+1]+1;
				edits[e].link = script[k+1];
				edits[e].op = DELETE;
			} else {
				row = last_d[k-1];
				edits[e].link = script[k-1];
				edits[e].op = INSERT;
			}
			edits[e].leftline = row;
			edits[e].rightline = col = row + k - ORIGIN;
			script[k] = &edits[e];

			while ((row < m) && (col < n) && eq(row,col)) {
				matches++;
				++row;
				++col;
			}
			    /*
			     * This is not part of the Meyer alg.
			     * It pretends there are more differences
			     * than there are to get larger blocks.
			     */
			    if (matches == 0) mismatches++;
			    else if (mismatches > 2*matches && matches < 6) {
				    row -= matches;
				    col -= matches;
				    matches = 0;
			    }
			last_d[k] = row;
			if (row == m && col == n) {
				put_script(left.start,
				    right.start,
				    l, r, script[k], out);
				if (row + left.start == left.n &&
				    col + right.start == right.n) {
					uninit(&left);
					uninit(&right);
					if (out != stdout) fclose(out);
					return (1);
				}
				left.start += row;
				right.start += col;
				goto again;
			}
			/*
			 * If we just found a big hunk that matches, reset.
			 */
			if (e == MAXD-1 || (matches > HUNK)) {
				debug((stderr, "RESET %d\n", e));
				put_script(left.start,
				    right.start,
				    l, r, script[k], out);
				left.start += row;
				right.start += col;
				goto again;
			}
			if (row == m) lower = k+2;
			if (col == n) upper = k-2;
		}
		lower--;
		upper++;
	}
	/* too many */
	uninit(&left);
	uninit(&right);
	if (out != stdout) fclose(out);
	return (2);
}

static void
put_script(int lbase, int rbase, char **left, char **right, edit_t *start,
	FILE *out)
{
	edit_t	*ep, *behind, *ahead, *a, *b, *c, *d;
	int	change;

	ahead = start;
	ep = 0;
	while (ahead) {
		behind = ep;
		ep = ahead;
		ahead = ahead->link;
		ep->link = behind;
	}

	while (ep) {
		b = ep;
		if (ep->op == INSERT) {
			fprintf(out, "%da", lbase + ep->leftline);
			c = ep;
			do {
				d = c;
				c = c->link;
			} while (c && c->op == INSERT
			    && c->leftline == b->leftline);
			if (d == ep) {
				fprintf(out, "%d\n", rbase + ep->rightline);
			} else {
				fprintf(out, "%d,%d\n", 
				    rbase + ep->rightline,
				    rbase + d->rightline);
			}
		} else {
			do {
				a = b;
				b = b->link;
			} while (b && b->op == DELETE &&
			    b->leftline == a->leftline+1);
			change = b && b->op == INSERT &&
			    b->leftline == a->leftline;
			if (a == ep) {
				fprintf(out, "%d", lbase + ep->leftline);
			} else {
				fprintf(out, "%d,%d",
				    lbase + ep->leftline,
				    lbase + a->leftline);
			}
			if (change) {
				assert(b);
				c = b;
				do {
					d = c;
					c = c->link;
				} while (c && c->op == INSERT
				    && c->leftline == b->leftline);
				if (d == b) {
					fprintf(out, "c%d\n",
					    rbase + b->rightline);
				} else {
					fprintf(out, "c%d,%d\n", 
					    rbase + b->rightline,
					    rbase + d->rightline);
				}
			} else {
				fprintf(out, "d%d\n",
				    rbase + ep->rightline);
			}
			do {
				putline('<',left[ep->leftline-1], out);
				ep = ep->link;
			} while (ep != b);
			if (!change) continue;
			fprintf(out, "---\n");
		}
		do {
			putline('>', right[ep->rightline-1], out);
			ep = ep->link;
		} while (ep && ep->op == INSERT && ep->leftline == b->leftline);
	}
}
#endif
