/*
 * A faster popen - this one skips invoking shell.  Saves a shitload.
 * You have to call it with an absolute pathname, however.
 * And it doesn't pass the environment.
 *
 * But, dig this, it does handle redirection.  But only if it is like
 *	foo > bar
 * the following don't work
 *	foo >bar
 *	foo 2>bar
 * etc.
 */
#undef	PURIFY
#undef	FASTPURIFY
#include "sccs.h"
#include <errno.h>

WHATSTR("@(#)popen.c 1.6");

static struct {
	int	pfd;
	int	pid;
} kids[20];

/*ARGSUSED*/
FILE *
fastPopen(const char *cmd, const char *type)
{
	char	*c = strdup(cmd);
	char	*s;
	int	pfds[2];
	int	pid;
	FILE	*p;
	int	ac = 0;
	int	i;
	char	*output = 0;
	char	*av[100];

	for (i = 0; i < 20; ++i) {
		if (!kids[i].pid) break;
	}
	if (i == 20) {
		fprintf(stderr, "Too many popens without pcloses.\n");
		return (0);
	}
	assert(cmd[0] == '/');
	for (s = strtok(c, " \t\n"); s; s = strtok(0, " \t\n")) {
		av[ac++] = s;
	}
	if (ac == 0) {
		perror(cmd);
		return(0);
	}
	if (ac > 2) {
		if (streq(av[ac-2], ">")) {
			output = av[ac-1];
			ac -= 2;
		}
		av[ac] = 0;
	}

	if (pipe(pfds) == -1) {
		perror("pipe");
		return (0);
	}
	switch (pid = fork()) {
	    case -1:
		perror("fork");
		return (0);

	    case 0:
		if (*type == 'r') {
			dup2(pfds[1], 1);
		} else {
			dup2(pfds[0], 0);
		}
		close(pfds[0]);
		close(pfds[1]);
		if (output) {
			int	fd = creat(output, 0664);

			if (fd == -1) {
				perror(output);
				exit(1);
			}
			dup2(fd, 1);
			close(fd);
		}
		execve(av[0], av, 0);
		perror(av[0]);
		exit(1);
	    
	    default:
		if (*type == 'r') {
			p = fdopen(pfds[0], type);
			close(pfds[1]);
		} else {
			p = fdopen(pfds[1], type);
			close(pfds[0]);
		}
		free(c);
		assert(fileno(p) > 0);
		kids[i].pfd = fileno(p);
		kids[i].pid = pid;
		return (p);
	}
}

int
fastPclose(FILE *f)
{
	int	i, status;

	if (!f || (fileno(f) <= 0)) {
		errno = EBADF;
		return (-1);
	}
	for (i = 0; i < 20; ++i) {
		if (kids[i].pfd == fileno(f)) break;
	}
	if (i == 20) {
		errno = EBADF;
		return (-1);
	}
	fflush(f);
	close(fileno(f));
	if (waitpid(kids[i].pid, &status, 0) == -1) {
		/* XXX - clear the table? */
		perror("waitpid in pclose");
		kids[i].pid = 0;
		return (-1);
	}
	kids[i].pid = 0;
	return (status);
}
