#ifndef	LINT
static	char *what = "@(#)purify.c 1.21";
#endif
#if	defined(PURIFY) || defined(FASTPURIFY)
#include <stdio.h>
#include <unistd.h>
#include <stdlib.h>
#include <assert.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <string.h>

#ifdef	PDEBUG
#	define	pdebug(x)	fprintf x
#else
#	define	pdebug(x)
#endif
#ifdef	MDEBUG
#	define	mdebug(x)	fprintf x
#else
#	define	mdebug(x)
#endif
#define bcopy(s,d,l)    memmove(d,s,l)
#define bcmp(a,b,l)     memcmp(a,b,l)
#define bzero(s,l)      memset(s,0,l)
#define streq(a,b)      !strcmp(a,b)
#define strneq(a,b,n)   !strncmp(a,b,n)
#define unless(e)       if (!(e))
#define next(buf, in)   fgets(buf, sizeof(buf), in)

void	purifyFiles(char *f, int l);
void	purifyCheck(char *file, int line);

/*
 * Memory allocator that keeps track of who allocated what.
 *
 * Memory is laid out like so:
 *
 * [struct mem w/ magic as last word][user data][magic]
 */
#define	MAGIC	0xdeadbeef
#define	BUFFER	4
typedef struct mem {
	char	*file;
	int	line;
	char	*mem;
	int	len;
	struct	mem *next, *prev;
	int	magic[BUFFER];
} mem_t;
static int deadbeef[BUFFER] = { MAGIC, MAGIC, MAGIC, MAGIC };

#define	MAXH	1024	/* big enough for sccs, may be too small for others */
#define	HASH(m)	(((int)(m)>>3)%MAXH)
static	mem_t	*mem_head[MAXH];

char	*
purify_alloc(int bytes, char *file, int line)
{
	int	size = bytes + sizeof(mem_t) + BUFFER * sizeof(int);
    	char	*mem = calloc(1, size);
	mem_t	*m = (mem_t*)mem;
	int	*after;
	int	i;
	
	assert(mem);
	m->file = file;
	m->line = line;
	m->mem = mem + sizeof(mem_t);
	m->len = bytes;
	after = (int*)(m->mem + m->len);
	bcopy(deadbeef, m->magic, sizeof(deadbeef));
	bcopy(deadbeef, after, sizeof(deadbeef));
	i = HASH(m->mem);
	m->next = mem_head[i];
	if (m->next) {
		mem_head[i]->prev = m;
	}
	mem_head[i] = m;
#ifndef	FASTPURIFY
	purifyCheck(file, line);
#endif
	return (m->mem);
}

char *
purify_strdup(char *s, char *file, int line)
{
	char	*t = purify_alloc(strlen(s)+1, file, line);

	if (t) { 
		strcpy(t, s);
	}
	return t;
}

void
purify_free(char *mem, char *file, int line)
{
	mem_t	*m;
	int	i, tmp;

	purifyCheck(file, line);
	for (m = mem_head[tmp = HASH(mem)]; m; m = m->next) {
		if (m->mem == mem) break;
	}
	if (!m) {
		fprintf(stderr, "free(%x) failed.  Called from %s at %d\n",
		    (unsigned int)mem, file, line);
		return;
	}
	for (i = 0; i < m->len; i++) {
		m->mem[i] = 127;
	}
	if (m->next) m->next->prev = m->prev;
	if (m->prev) m->prev->next = m->next;
	if (m == mem_head[tmp]) mem_head[tmp] = m->next;
	free(m);
}

void
purifyList(char *file, int line)
{
	mem_t	*m;
	int	first = 1;
	int	i, j, p;

	for (i = 0; i < MAXH; ++i) {
		for (m = mem_head[i]; m; m = m->next) {
			if (first) {
				fprintf(stderr, 
				    "--- Memory list called from %s:%d ---\n",
				    file, line);
				first = 0;
			}
			fprintf(stderr, "%d bytes alloced by %s:%d",
			    m->len, m->file, m->line);
			for (p = 1, j = 0; j < m->len; j++) {
				if (!isascii(m->mem[j])) p = 0;
			}
			if (p) fprintf(stderr, ": '%.*s'", m->len, m->mem);
			fprintf(stderr, "\n");
		}
	}
	purifyFiles(file, line);
}

void
purifyCheck(char *file, int line)
{
	mem_t	*m;
	int	*after;
	int	first = 1;
	int	i;

	for (i = 0; i < MAXH; ++i) {
		for (m = mem_head[i]; m; m = m->next) {
			after = (int*)(m->mem + m->len);
			if (m->magic[0] != MAGIC || m->magic[1] != MAGIC ||
			    m->magic[2] != MAGIC || m->magic[3] != MAGIC) {
				if (first) {
					fprintf(stderr, 
				    "--- Memory check called from %s:%d ---\n",
					    file, line);
					first = 0;
				}
				fprintf(stderr,
				    "memory underrun for %d"
				    " byte alloc from %s:%d\n",
				    m->len, m->file, m->line);
				assert("UNDER" == (char*)0);
			}
			/*
			 * Has to be bcmp for RISC's.
			 */
			if (bcmp(after, deadbeef, sizeof(deadbeef))) {
				if (first) {
					fprintf(stderr, 
				    "--- Memory check called from %s:%d ---\n",
					    file, line);
					first = 0;
				}
				fprintf(stderr,
				    "memory overrun for %d"
				    " byte alloc from %s:%d\n",
				    m->len, m->file, m->line);
				assert("OVER" == (char*)0);
			}
		}
	}
}

#define	FOPENED		1
#define	POPENED		2
#define	OPENED		3
#define	FDOPENED	4
#define	DUPPED		5
#define	PIPE		6
static	char	*style[] = {
	"closed",
	"fopened", 
	"popened",
	"opened",
	"fdopened",
	"dupped",
	"pipe",
};
static struct	{
	char	*file;
	int	line;
	int	open;
} openfiles[200];

void
purifyFiles(char *file, int line)
{
	int	i;
	int	first = 1;

	for (i = 0; i < 200; ++i) {
		if (openfiles[i].open) {
			if (first) {
				fprintf(stderr, 
				    "--- File check called from %s:%d ---\n",
				    file, line);
				first = 0;
			}
			fprintf(stderr, "%s file %d opened by %s:%d not closed\n",
			    style[openfiles[i].open], i, 
			    openfiles[i].file, openfiles[i].line);
		}
	}
}

int
purify_open(char *p, int mode, int flags, char *file, int line)
{
	int	fd = open(p, mode, flags);

	pdebug((stderr, "open(%s, %o, %o) = %d\n", p,mode,flags,fd));
	if (fd == -1) return (fd);
	openfiles[fd].file = file;
	openfiles[fd].line = line;
	openfiles[fd].open = OPENED;
	return (fd);
}

int
purify_creat(char *p, int mode, char *file, int line)
{
	int	fd = creat(p, mode);

	pdebug((stderr, "creat(%s, %o) = %d\n", p,mode,fd));
	if (fd == -1) return (fd);
	openfiles[fd].file = file;
	openfiles[fd].line = line;
	openfiles[fd].open = OPENED;
	return (fd);
}

int
purify_dup(int f, char *file, int line)
{
	int	fd = dup(f);

	pdebug((stderr, "dup(%d) = %d\n",f,fd));
	if (fd == -1) return (fd);
	openfiles[fd].file = file;
	openfiles[fd].line = line;
	openfiles[fd].open = DUPPED;
	return (fd);
}

int
purify_dup2(int old, int new, char *file, int line)
{
	int	fd = dup2(old, new);

	pdebug((stderr, "dup2(%d,%d) = %d\n",old,new,fd));
	if (fd == -1) return (fd);
	openfiles[fd].file = file;
	openfiles[fd].line = line;
	openfiles[fd].open = DUPPED;
	return (fd);
}

int
purify_pipe(int fds[2], char *file, int line)
{
	int	ret = pipe(fds);

	pdebug((stderr, "pipe() = %d [%d,%d]\n",ret, fds[0],fds[1]));
	if (ret == -1) return (ret);
	openfiles[fds[0]].file = file;
	openfiles[fds[0]].line = line;
	openfiles[fds[0]].open = PIPE;
	openfiles[fds[1]].file = file;
	openfiles[fds[1]].line = line;
	openfiles[fds[1]].open = PIPE;
	return (ret);
}

int
purify_close(int i, char *file, int line)
{
	pdebug((stderr, "close(%d)\n", i));
	if (i < 0) return (-1);
	if (!openfiles[i].open) {
		if (openfiles[i].file) {
			fprintf(stderr, "Duplicate close, %s:%d was first, now %s:%d\n",
			    openfiles[i].file, openfiles[i].line, file, line);
			assert("DUP" == 0);
		}
		fprintf(stderr, "Bad close on fd %d called from %s:%d\n", 
		    i, file, line);
		assert("BAD FD" == 0);
	}
	if ((openfiles[i].open != OPENED) && (openfiles[i].open != DUPPED) &&
	    (openfiles[i].open != PIPE)) {
		fprintf(stderr,
		    "close on file that was %s by %s:%d, closer %s:%d\n",
		    style[openfiles[i].open],
		    openfiles[i].file, openfiles[i].line, file, line);
		assert("MISMATCH" == 0);
	}
	openfiles[i].open = 0;
	return (close(i));
}

FILE *
purify_popen(char *name, char *mode, char *file, int line)
{
	FILE	*ret = popen(name, mode);

	pdebug((stderr, "popen(%s, %s) = %x\n", name,mode,ret));
	if (ret) {
		int	i = fileno(ret);

		openfiles[i].file = file;
		openfiles[i].line = line;
		openfiles[i].open = POPENED;
	}
	return (ret);
}

int
purify_pclose(FILE *f, char *file, int line)
{
	int	i;

	pdebug((stderr, "pclose(%x)\n", f));
	if (!f) return(-1);
	i = fileno(f);
	if (!openfiles[i].open) {
		if (openfiles[i].file) {
			fprintf(stderr, "Duplicate close, %s:%d was first, now %s:%d\n",
			    openfiles[i].file, openfiles[i].line, file, line);
			assert("DUP" == 0);
		}
		fprintf(stderr, "Bad pclose on fd %d called from %s:%d\n", 
		    i, file, line);
		assert("BAD FD" == 0);
	}
	if (openfiles[i].open != POPENED) {
		fprintf(stderr, "pclose on file that was %s by %s:%d, closer %s:%d\n",
		    style[openfiles[i].open],
		    openfiles[i].file, openfiles[i].line, file, line);
		assert("MISMATCH" == 0);
	}
	openfiles[i].open = 0;
	return (pclose(f));
}

FILE *
purify_fopen(char *name, char *mode, char *file, int line)
{
	FILE	*ret = fopen(name, mode);

	pdebug((stderr, "fopen(%s, %s) = %x\n", name,mode,ret));
	if (ret) {
		int	i = fileno(ret);

		openfiles[i].file = file;
		openfiles[i].line = line;
		openfiles[i].open = FOPENED;
	}
	return (ret);
}

FILE *
purify_fdopen(int fd, char *mode, char *file, int line)
{
	FILE	*ret = fdopen(fd, mode);

	pdebug((stderr, "fdopen(%d, %s) = %x\n", fd,mode,ret));
	if (ret) {
		int	i = fileno(ret);

		openfiles[i].file = file;
		openfiles[i].line = line;
		openfiles[i].open = FDOPENED;
	} else {
		fprintf(stderr, "fdopen(%d, %s) = -1 at %s:%d\n",
		    fd, mode, file, line);
	}
	return (ret);
}

FILE *
purify_freopen(char *path, char *mode, FILE *ret, char *file, int line)
{
	ret = freopen(path, mode, ret);

	pdebug((stderr, "freopen(%s, %s, x) = %x\n", path,mode,ret));
	if (ret) {
		int	i = fileno(ret);

		openfiles[i].file = file;
		openfiles[i].line = line;
		openfiles[i].open = FOPENED;
	}
	return (ret);
}

int
purify_fclose(FILE *f, char *file, int line)
{
	int	i;

	pdebug((stderr, "fclose(%x)\n", f));
	if (!f) return(-1);
	i = fileno(f);
	if (i < 0 || i > 199) {
		fprintf(stderr, "fclose on bad FILE * 0x%x from %s:%d\n",
		    (unsigned int)f, file, line);
		return (-1);
	}
	if (!openfiles[i].open) {
		if (openfiles[i].file) {
			fprintf(stderr, "Duplicate close, %s:%d was first, now %s:%d\n",
			    openfiles[i].file, openfiles[i].line, file, line);
			assert("DUP" == 0);
		}
		fprintf(stderr, "Bad fclose on fd %d called from %s:%d\n", 
		    i, file, line);
		if (fileno(f) < 3) return (fclose(f));
		assert("BAD FD" == 0);
	}
	if ((openfiles[i].open != FOPENED) && (openfiles[i].open != FDOPENED)) {
		fprintf(stderr, "fclose on file that was %s by %s:%d, closer %s:%d\n",
		    style[openfiles[i].open],
		    openfiles[i].file, openfiles[i].line, file, line);
		assert("MISMATCH" == 0);
	}
	openfiles[i].open = 0;
	return (fclose(f));
}


#if 0
main()
{
	int	*foo = (int*)purify_alloc(16, __FILE__, __LINE__);
	char	*s = purify_strdup("hi there mom", __FILE__, __LINE__);

	purifyCheck(__FILE__, __LINE__);
	foo[0] = foo[1] = foo[2] = foo[3] = 0;
	purifyCheck(__FILE__, __LINE__);
	/* s[-1] = 0; */
	/* s[14] = 0; */
	/* foo[-1] = 0; */
	/* foo[4] = 0; */
	purifyCheck(__FILE__, __LINE__);
}
#endif

#else
int __dummy_for_ansic(int foo) { return (foo + 12); }
#endif
