/*
 * We thought we were SCCS but we really are RCS.
 * We need to go find the real rcs command and run that instead.
 * I could hardwire the RCS command paths and I have no objection
 * if you want to do that.  However, this code handles a more dynamic
 * solution:  if you stick the SCCS co/ci in your path first and stick
 * the RCS co/ci in your path later, this code will probably find 
 * the right RCS.  The alg is
 *
 * for each dir in your path
 *	skip if .
 *	skip if $HOME/<whatever>
 *	look for $dir/rcscommand
 * Try whatever you find.
 *
 * In addition, we also look for rcs<command>, i.e., "rcsci" "rcsco".
 */
#include "sccs.h"
WHATSTR("@(#)rcs.c 1.4");

/*ARGSUSED*/
void
rcs(char *cmd, int ac, char **av, char **ev)
{
	char	*s, *home = getenv("HOME");
	char	buf[1024];

	for (s = strtok(getenv("PATH"), ":"); s; s = strtok(0, ":")) {
		sprintf(buf, "%s/rcs%s", s, cmd);
		if (executable(buf)) {
			execve(buf, av, ev);
			/* Fall through if the exec fails */
		}
		if (s[0] == '.' && !s[1]) continue;
		if (home) {
			if (!strncmp(home, s, strlen(home))) {
				continue;
			}
		} else {	/* skip anything not in / or /usr */
			char	*slash = index(s, '/');

			if (!slash) continue; 	/* huh?!? */
			slash = index(slash, '/');
			if (slash && strncmp("/usr/", s, 4)) continue;
		}
		sprintf(buf, "%s/%s", s, cmd);
		debug((stderr, "Looking for real RCS in %s\n", buf));
		if (executable(buf)) {
			execve(buf, av, ev);
			perror(buf);
			exit(1);
			/* NOTREACHED */
		}
	}
	fprintf(stderr, "Can't find %s in your path\n", cmd);
	exit(1);
}
