#!/usr/bin/perl -sw

# Copyright (c) 1998 Larry McVoy, all rights reserved.
# 1.14 07/20/98 lm@lm.bitmover.com
# Convert RCS files to SCCS files, preserving the following:
#	user
#	date
#	state
#	symbols
#	parent/child relationship
#	default branch
#	access list of users
#	descriptive text
# XXX - not done
#	comment/user/date/descriptive/user+group/symbols verification
# XXX - not done
#	locks (aka - a file that is checked out -l in the source)
#	there's no reason I couldn't handle this, I just haven't yet.
#
# Because RCS and SCCS branch numbering schemes are different, we can't
# preserve the revision numbers except for fairly simple trees.  We
# approximate by checking the deltas in in the same order as they were
# checked into the RCS file.

$notbig = $verify = 0 if 0;
$x = $x ? " -x" : "";
if ($n) {
	open(SH, "|while read x; do echo \$x; done");
} else {
	open(SH, "|sccssh$x");
}
chop($pwd = `pwd`);
print SH "mkdir SCCS\n";
$stmp = "/tmp/sccs$$";
$rtmp = "/tmp/rcs$$";
$dtmp = "/tmp/diff$$";
@ARGV = <RCS/*,v> if ($#ARGV == -1);
foreach $file (@ARGV) {
	@foo = stat($file);
	if ($notbig && ($foo[7] > 1000000)) {
		warn "Skipping $file, $foo[7] too big.\n";
		next;
	}
	$file =~ s|RCS/||; chop($file); chop($file);
	if (-e "SCCS/s.$file") {
		print SH "echo Skipping $pwd/$file, already done.\n" unless $q;
		next;
	}
	warn "------------ $file -------------\n" if $d;
	&buildtree;
	&printrevs;
}
print SH "/bin/rm -f comment $stmp $rtmp $dtmp\n";
close(SH);
exit 0;

sub checkouts
{
	$sccsrev = $rev = $_[0];
	$sccsrev = $sccsrev{$rev} if defined $sccsrev{$rev};
	print SH "echo D $sccsrev $date{$rev} $author{$rev} " .
	      "> comment-$sccsrev\n";
	if (defined $comments{$rev}) {
		$_ = $comments{$rev};
		foreach $c (split(/\n/, $_)) {
			print SH "echo C ${c} >> comment-$sccsrev\n";
		}
	} else {
		print SH "echo $file: No comments for rev $rev\n" unless $q;
	}
	$f = "$file-$sccsrev";
	if (defined $parent{$rev}) {
		$p = $parent{$rev};
		$p = $sccsrev{$p} if defined $sccsrev{$p};
		print SH "/usr/bin/co -q -kk -p$rev $file > $f\n";
		print SH "chmod +x $f\n" if -x "RCS/${file},v";
	} else {
		print SH "/usr/bin/co -q -kk -p$rev $file > $f\n";
	}
}

sub deltas
{
	$sccsrev = $rev = $_[0];
	if ($rev =~ /\..*\./) {
		return if $pass == 1;
	} else {
		return if $pass == 2;
	}
	$sccsrev = $sccsrev{$rev} if defined $sccsrev{$rev};
	$f = "$file-$sccsrev";
	print SH "mv $f $file\n";
	if (defined $parent{$rev}) {
		$p = $parent{$rev};
		$p = $sccsrev{$p} if defined $sccsrev{$p};
		# They can switch revs in us mid stream, like 1.6 -> 3.1
		@r = split(/\./, $rev);
		@p = split(/\./, $p);
		if ($r[0] != $p[0]) {
			die "$pwd/$file: $p->$rev\n"
			    if ($r[0] < $p[0]);
			print SH
			    "$pwd/$file: $p->$rev\n" if ($r[1] > 1);
			die "$pwd/$file: $p->$rev\n" if ($#r != 1);
			$p = $r[0];
		}
		print SH "get -Fgesr$p $file\n";
		print SH "delta -r$sccsrev -CsfIcomment-$sccsrev $file\n";
	} else {
		print SH "delta -r$sccsrev -CsifIcomment-$sccsrev $file\n";
	}
	print SH "rm comment-$sccsrev\n";
}

sub diffs 
{
	$sccsrev = $rev = $_[0];
	$sccsrev = $sccsrev{$rev} if defined $sccsrev{$rev};
	print SH "echo -n $sccsrev,\n" unless $q;
	print SH "get -Fspk -r$sccsrev $file > $stmp\n";
	print SH "/usr/bin/co -q -p -kk -r$rev $file > $rtmp\n";
	print SH "diff $rtmp $stmp > $dtmp\n";
	print SH "if [ -s $dtmp ]; then echo $pwd/$file bad R $rev S $sccsrev; fi\n";
}

sub printrevs
{
	return unless (-s "RCS/${file},v");
	print SH "echo ========== $file =============\n" unless $q;
	print SH "date\n" if $time;
	for ($i = 0; $i <= $#dates; ++$i) {
		next if $i < $#dates && $dates[$i] eq $dates[$i+1];
		$date = $dates[$i];
		foreach $duprev (split(/ /, $date2rev{$date})) {
			&checkouts($duprev);
		}
	}
	print SH "date\n" if $time;
	$pass = 1;
B:	for ($i = 0; $i <= $#dates; ++$i) {
		next if $i < $#dates && $dates[$i] eq $dates[$i+1];
		$date = $dates[$i];
		foreach $duprev (split(/ /, $date2rev{$date})) {
			&deltas($duprev);
		}
	}
	goto B if (++$pass == 2);
	print SH "admin -qz $file\n";
	if ($#descriptive != -1) {
		$gotone = 0;
		foreach $a (@descriptive) {
			chop($a);
			if ($a =~ /^\s*$/) {
				print SH
				    "echo $pwd/$file: Stripping blank line " .
				    "from description\n";
			} else {
				print SH "echo ${a} >> descriptive\n";
				$gotone++;
			}
		}
		if ($gotone) {
			print SH "admin -tdescriptive $file\n";
			print SH "rm descriptive\n";
		}
	}
	print SH "admin -fb -fR $file\n";
	if ($defbranch ne "") {
		print SH "admin -fd$defbranch $file\n";
	}
	if ($#access != -1) {
		foreach $a (@access) {
			print SH "admin -a$a $file\n";
		}
	}
	if ($#symbols != -1) {
		foreach $a (@symbols) {
			($sym, $rev) = split(/:/, $a);
			if (defined $date{$rev}) {
				$r = $rev;
				$r = $sccsrev{$rev}
				    if defined $sccsrev{$rev};
				print SH "admin -qs${sym}=$r $file\n";
			} else {
				$r = "${rev}.1";
				if (!defined $date{$r}) {
					print SH "echo $pwd/$file: No match for symbol $a.\n";
				} else {
					$r = $sccsrev{$rev}
					    if defined $sccsrev{$rev};
					chop($r);
					chop($r);
					print SH "admin -qs${sym}=$r $file\n";
				}
			}
		}
	}
	print SH "date\n" if $time;
	return unless $verify;
	# Check data contents.
	print SH "echo -n Data:\n" unless $q;
	for ($i = 0; $i <= $#dates; ++$i) {
		next if $i < $#dates && $dates[$i] eq $dates[$i+1];
		$date = $dates[$i];
		foreach $duprev (split(/ /, $date2rev{$date})) {
			&diffs($duprev);
		}
	}
	print SH "echo done\n" unless $q;
}

# Data structures:
#	$author{$rev} = "lm"
#	$comments{$rev} = the comments as one entry w/ newlines.
#	$date2rev{$date} = 1.1.
#	$date{$rev} = the date as "98/03/27 09:12:21";
#	$parent{$rev} = 1.1
#	$state{$rev} = "Exp"
#	$sccsrev{$rev} = sccs version of rev (i.e., 1.3.1.1.1.1 -> 1.3.2.1)
#	$sccsbranches{$rev} = sccs rev of a .1 (top of branch)
#	$rcsrev{$rev} = the other way around
#	@access = (user, user, user...)
#	@dates = (date, date, date...)
#	@descriptive = (line, line, line)
#	@symbols = (SYM:REV, SYM:REV, ...)

sub buildtree
{
	%author = %comments = %date = %date2rev = %parent =
	%sccsbranches = %sccsrev = %state = %rcsrev = ();

	$defbranch = "";
	@access = @dates = @descriptive = @symbols = ();

	return unless (-s "RCS/${file},v");
	open(R, "RCS/${file},v");
	$_ = <R>; 
	unless (/^head\s/) {
		print SH "echo $pwd/$file: bad format, wanted head, $_\n";
		return;
	}
	$_ = <R>; 
	# branch: 1.3.1
	if (/^branch:\s*\d/) {
		@_ = split;
		$defbranch = $_[1];
		warn "Defbranch: $_[1]\n" if $d;
		$_ = <R>;
	}
	unless (/^access/) {
		print SH "echo $pwd/$file: bad format, wanted access, $_\n";
		return;
	}
	if (/^access$/) {	# there are users in the list
		while (<R>) {
			chop;
			s/\s+//g;
			$last = /;/;
			s/;//;
			push(@access, $_);
			last if $last;
		}
	}
	$_ = <R>;
	unless (/^symbols/) {
		print SH "echo $pwd/$file: bad format, wanted symbols, $_\n";
		return;
	}
	if (/^symbols$/) {	# there are symbols
		while (<R>) {
			chop;
			s/\s+//g;
			$last = /;/;
			s/;//;
			push(@symbols, $_);
			last if $last;
		}
	}
	# Ignore locks.
	$_ = <R>; 
	unless (/^locks/) {
		print SH "echo $pwd/$file: bad format, wanted locks, $_\n";
		return;
	}
	if (/^locks$/) {
		while (<R>) {
			last if /^comment/;
			last if /^expand/;
			last if /^\s*$/;
		}
	}
	while (<R>) { last if /^\s*$/; }

	# OK, onto the revs.
	while (<R>) {
#print STDERR "LINE $_";
		next if /^\s*$/;
		last if /^desc$/;
		if (/^[\d\.]+$/) {
			chop;
			$rev = $_;
		} elsif (/^date\s+(\S+);\s+author\s+(\S+);\s+state\s+(\S+);/) {
			$date = $1;
			@d = split(/[\.]/, $date);
			if ($d[0] < 69) {
				$d[0] += 2000;
			} elsif ($d[0] < 100) {
				$d[0] += 1900;
			}
			$date{$rev} =
			    join('/', @d[0..2]) . " " . join(':', @d[3..5]);
			if (defined $date2rev{$date{$rev}}) {
				warn "Duplicate dates: $rev $date2rev{$date{$rev}}\n" unless $q;
				$date2rev{$date{$rev}} .= " $rev";
			} else {
				$date2rev{$date{$rev}} = $rev;
			}
			push(@dates, $date{$rev});
			$author{$rev} = $2;
			$state{$rev} = $3;
		} elsif (/^branches/) {
			if (/^branches$/) {
				while (<R>) {
					s/\s+//g;
					$last = /;/;
					s/;//;
					$parent{$_} = $rev;
warn "A parent($_) = $rev\n" if $d;
					last if $last;
				}
			}
		} elsif (/^next/) {
			if (/next\s+([\d\.]+);/) {
				$tmp = $1;
				if (/\..*\./) {
					$parent{$tmp} = $rev;
warn "B(2) parent($tmp) = $rev\n" if $d;
				} else {
					$parent{$rev} = $tmp;
warn "B(1) parent($rev) = $tmp\n" if $d;
				}
			}
		} else {
			warn "$pwd/$file: ``$_''?";
		}
	}
	$_ = <R>;
	die "$pwd/$file: Bad descript: ``$_''" unless /^@/;
	unless (/^@\@$/) {
		s/^@//;
		push(@descriptive, $_);
		while (<R>) {
			last if /^@/;
			push(@descriptive, $_);
		}
	}

	while (<R>) {
		next if /^\s*$/;
		die "$pwd/$file: needed digits: ``$_'' at $.\n"
		    unless (/^\d\.\d/);
		chop;
		$rev = $_;
		$_ = <R>;
		die "$pwd/$file: needed log: ``$_'' at $.\n" unless /^log$/;
		$_ = <R>;
		die "$pwd/$file: needed log \@: ``$_'' at $.\n" unless /^@/;
		unless (/^@\@$/) {
			s/^@//;
			s/@@/@/g;
			$comments{$rev} = $_;
			while (<R>) {
				last if /^@/;
				s/@@/@/g;
				$comments{$rev} .= $_;
			}
		}
		$_ = <R>;
		die "$pwd/$file: needed text: ``$_'' at $.\n" unless /^text$/;
		$_ = <R>;
		die "$pwd/$file: needed text \@: ``$_'' at $.\n" unless /^@/;
		unless (/^@\@$/) {
			while (<R>) {
				next unless /@/;
				s/@@//g;
				last if /@/;
			}
		}
	}

	@dates = sort(@dates);
	foreach $date (@dates) {
		# XXX - I should worry about what to do if these have the
		# same time.
		warn "$date $date2rev{$date}\n" if $d;
		# Figure out if the rev is not cool in SCCS land and fix it.
		$rev = $date2rev{$date};
		if (!oksccs($rev)) {
			if (defined $sccsrev{$parent{$rev}}) {
				@n = split(/\./, $sccsrev{$parent{$rev}});
				$n[3]++;
				$sccsrev{$rev} = join('.', @n);
				$rcsrev{$sccsrev{$rev}} = $rev;
warn "A $rev -> $sccsrev{$rev}\n" if $d;
			} else {
				# It's a new branch, so go figure out what
				# we can use.
				$p = $rev;
				do {
warn "$p -> $parent{$p}\n" if $d;
					$p = $parent{$p};
				} while ($p =~ /\..*\..*\./);
				$b = 1;
				do {
					$try = "${p}.${b}.1";
warn "Try $try\n" if $d;
					$b++;
				} while (defined $sccsbranches{$try} ||
				         defined $date{$try});
				$sccsbranches{$try} = 1;
				$sccsrev{$rev} = $try;
				$rcsrev{$sccsrev{$rev}} = $rev;
warn "B $rev -> $sccsrev{$rev}\n" if $d;
			}
		}
	}
}

sub oksccs
{
	return ($_[0] !~ /\..*\..*\..*\..*\./);
}
