#include "sccs.h"
WHATSTR("@(#)sccssh.c 1.35");

/*
 * sccssh - a shell for SCCS.  It's just like normal sccs commands except
 * you can feed it as many as you want.
 */
#define	prompt()	if (tty) { fprintf(stderr, "sccs> "); } \
			buf[0] = 0; redirect = 0; myoptind=0;
int	get_main(int ac, char **av, char *out);
int	admin_main(int ac, char **av, char **ev);
int	delta_main(int ac, char **av);
int	cp(int ac, char **av);
int	docmd(char *cmd);
int	echo(int ac, char **av, FILE *out);
int	rm(int ac, char **av);
int	mv(int ac, char **av);
int	datecmd(FILE *out);

int
main(int argc, char **argv, char **ev)
{
	int	ac;
	char	*redirect = 0;
	char	*output;
	int	tty = isatty(0);
	char	*av[200];
	char	buf[2000];
	FILE	*out = stdout;
	int	verbose = mygetopt(argc, argv, "x") == 'x';

	prompt();
	while (fnext(buf, stdin)) {
		if (verbose) fputs(buf, stderr);
		if (!buf[0] || buf[0] == '\n') {
			prompt();
			continue;
		}
		if (strneq("exit", buf, 4)) {
			if (buf[4] && buf[5]) exit(buf[5]);
			exit(0);
		}
		if (!strncmp("for ", buf, 4) ||
		    !strncmp("if ", buf, 3) ||
		    !strncmp("case ", buf, 5)) {
		    	system(buf);
			prompt();
			continue;
		}
		if (strncmp("get ", buf, 4) &&
		    strncmp("admin ", buf, 6) &&
		    strncmp("diff ", buf, 5) &&
		    strncmp("date", buf, 4) &&
		    strncmp("echo ", buf, 5) &&
		    strncmp("cp ", buf, 3) &&
		    strncmp("rm ", buf, 3) &&
		    strncmp("mv ", buf, 3) &&
		    strncmp("cd ", buf, 3) &&
		    strncmp("delta ", buf, 6)) {
			docmd(buf);
			prompt();
			continue;
		}
		av[ac=0] = strtok(buf, " \n");
		while ((av[++ac] = strtok(0, " \n")));
		if (ac > 3) {
			if (streq(">", av[ac-2])) {
				redirect = "w";
			} else if (streq(">>", av[ac-2])) {
				redirect = "a";
			} else if (streq("<", av[ac-2])) {
				redirect = "r";
			}
		}
		if (redirect) {
			output = av[ac-1];
			ac -= 2;
			out = fopen(output, redirect);
			if (!out) {
				perror(output);
				exit(1);
			}
		} else {
			output = "-";
		}
		av[ac] = 0;
		/*
		{ int i;
		for (i = 0; i < ac; ++i) printf("[%s] ",av[i]);
		if (out != stdout) printf("> %s", output);
		puts("");
		}
		*/
		if (streq(av[0], "get")) {
			if (get_main(ac, av, output)) {
				exit(1);
			}
		} else if (strneq("date", buf, 4)) {
			datecmd(out);
		} else if (strneq("echo", buf, 4)) {
			echo(ac, av, out);
		} else if (strneq("rm", buf, 2)) {
			if (rm(ac, av)) exit(1);
		} else if (strneq("mv", buf, 2)) {
			if (mv(ac, av)) exit(1);
		} else if (streq(av[0], "delta")) {
			if (delta_main(ac, av)) {
				exit(1);
			}
		} else if (streq(av[0], "delta")) {
			if (delta_main(ac, av)) {
				exit(1);
			}
		} else if (streq(av[0], "admin")) {
			if (admin_main(ac, av, ev)) {
				exit(1);
			}
		} else if (streq(av[0], "cp")) {
			if (cp(ac, av)) exit(1);
		} else if (streq(av[0], "cd")) {
			if (chdir(av[1])) exit(1);
		} else if (streq(av[0], "diff")) {
			assert(ac == 3);
			diff(av[1], av[2], D_DIFF, output);
		} else {
			fprintf(stderr, "Bad command: %s\n", av[0]);
			exit(1);
		}
		if (redirect) {
			assert(out != stdout);
			fclose(out);
			out = stdout;
			output = 0;
			redirect = 0;
		}
		prompt();
		purify_check();
	}
	purify_list();
	return (0);
}

int
cp(int ac, char **av)
{
	int	from, to, n;
	char	buf[64<<10];

	if (ac != 3) {
		perror("cp - must be 2 args");
		return (-1);
	}
	from = open(av[1], 0, 0);
	if (from == -1) {
		perror(av[1]);
		return (-1);
	}
	to = creat(av[2], 0664);
	if (to == -1) {
		perror(av[1]);
		return (-1);
	}
	while ((n = read(from, buf, sizeof(buf))) > 0) {
		write(to, buf, n);
	}
	close(from);
	close(to);
	return(0);
}

int
echo(int ac, char **av, FILE *out)
{
	int	needspace = 0;
	int	i = 1;
	int	newline = 1;

	if ((ac > 1) && streq(av[1], "-n")) {
		newline = 0;
		i++;
	}
	while (av[i]) {
		if (needspace) fputc(' ', out);
		needspace = 1;
		fputs(av[i++], out);
	}
	if (newline) fputc('\n', out); else fflush(out);
	return (0);
}

int
rm(int ac, char **av)
{
	int	i = 1;

	while (av[i]) {
		if (isreg(av[i])) {
			if (unlink(av[i])) {
				perror(av[i]);
				return (-1);
			}
		} else {
			fprintf(stderr, "builtin rm can't remove %s\n", av[i]);
			return (-1);
		}
		i++;
	}
	return (0);	
}

int
mv(int ac, char **av)
{
	if (ac != 3) {
		fprintf(stderr, "usage: mv reg-file reg-file (ac=%d)\n",ac);
		return (-1);
	}
	if (!isreg(av[1])) {
		fprintf(stderr,
		    "usage: mv reg-file reg-file (%s ! reg)\n",av[1]);
		return (-1);
	}
	unlink(av[2]);
	if (link(av[1], av[2])) {
		perror(av[1]);
		return (-1);
	}
	if (unlink(av[1])) {
		perror(av[1]);
		return (-1);
	}
	return (0);
}

int
datecmd(FILE *out)
{
	time_t	now;
	struct	timeval	tv;
	char	*s;

	time(&now);
	gettimeofday(&tv, 0);
	s = ctime(&now); s[19] = 0;
	fprintf(out, "%s.%03d\n", s, (int)tv.tv_usec/1000);
	return (0);
}

#define main  dummy
#ifndef LINT
#undef	WHATSTR
#define	WHATSTR(x)	static char *getwhat = x
#endif
#define	NOPURIFY
#include "get.c"

#undef main
#define main  delta_main
#ifndef LINT
#undef	WHATSTR
#define	WHATSTR(x)	static char *deltawhat = x
#endif
#include "delta.c"


#undef main
#define main  admin_main
#ifndef LINT
#undef	WHATSTR
#define	WHATSTR(x)	static char *adminwhat = x
#endif
#include "admin.c"

/*
 * This does no argument processing like a shell does.
 * No variables, no quotes, no nothing.
 */
int
docmd(char *cmd)
{
	char	*args[100];
	int	i;
	char	*out = 0, *input = 0, *append = 0;
	

	for (args[i=0] = strtok(cmd, " \t\n");
	    (args[++i] = strtok(0, " \t\n")); );
	i--;
	if (i > 1 && streq(args[i-1], ">")) {
		out = args[i];
		args[i-1] = 0;
	} else if (i > 1 && streq(args[i-1], ">>")) {
		append = args[i];
		args[i-1] = 0;
	} else if (i > 1 && streq(args[i-1], "<")) {
		input = args[i];
		args[i-1] = 0;
	}
	switch (fork()) {
	    case 0:
		if (out) {
#undef	close
			close(1);
			open(out, O_WRONLY|O_CREAT|O_TRUNC, 0664);
		}
		if (append) {
			close(1);
			open(append, O_WRONLY, 0664);
			lseek(1, 0, SEEK_END);
		}
		if (input) {
			close(0);
			open(input, O_RDONLY, 0);
		}
		execvp(args[0], args);
		perror(args[0]);
		exit(1);
	    case -1:
	    	perror("fork");
		return (-1);
	    default:
	    	wait(0);
		return (0);
	}
}

