#include "sccs.h"
#include <dirent.h>
WHATSTR("@(#)sfiles.c 1.14");

/*
 * sfiles.c - s.file file name processing.
 *
 * File name expansion.
 *	<dir> means <dir>/SCCS/s.* if <dir>/SCCS exists
 *	<dir> means <dir>/s.* if <dir>/SCCS doesn't exist
 *	Null means SCCS/s.* if SCCS exists
 *	NULL means ./s.* if SCCS doesn't exist
 *	<files> means <files>	XXX - should check them
 *	- means read files from stdin until EOF.
 *
 * XXX - if you have a file s.foo, "foo" doesn't work.  OK?
 * XXX - if SCCS/SCCS exists, that bites.
 *
 * char	*sfileFirst(av, flags);
 * char	*sfileNext();
 * void sfileDone();
 *
 * It is a fatal error to call sfileFirst() without calling sfileDone() or
 * before sfileNext() returned NULL.
 */
static	int flags;		/* saved flags */
static	char **av;		/* saved copy of argv */
static	int ac;			/* where we are - starts at 0 */
static	FILE *flist;		/* set if getting files from stdin */
static	DIR *d;			/* directory we are reading */
static	char prefix[1024];	/* path/to/dir/SCCS/ */
static	char buf[1024];		/* pathname we actually pass back */
static	int unget;		/* if set, return path again */
static	char *prog;		/* av[0], sort of */

private	int oksccs(char *s, int flags, int complain);

/*
 * Get the next file and munge it into an s.file name.
 */
char *
sfileNext()
{
	char	*name;
	struct	dirent *e;

	if (unget) {
		unget = 0;
		return (buf);
	}
again:
	if (!flist && !d && !av) {
		return (0);
	}
	if (flist) {
		unless (fgets(buf, sizeof(buf), flist)) {
			if (flist != stdin) fclose(flist);
			flist = 0;
			return (0);
		}
		name = buf;
		chop(name);
	} else if (d) {
		while ((e = readdir(d))) {
			unless (oksccs(e->d_name, flags, 0)) continue;
			name = e->d_name;
			sprintf(buf, "%s%s", prefix, e->d_name);
			return (buf);
		}
		closedir(d);
		d = 0;
		return (0);
	} else if (av) {
		unless (name = av[ac++]) {
			av = 0;
			return (0);
		}
		strcpy(buf, name);
	}
	if (!is_sccs(buf)) {
#ifdef	ATT_SCCS
		fprintf(stderr, "Not an SCCS file: %s\n", buf);
		goto again;
#endif
		name = name2sccs(buf);
		strcpy(buf, name);
		free(name);
	}
	if (oksccs(buf, flags, 1)) {
		return (buf);
	}
	goto again;
}

/*
 * Initialization - figure out where we are getting the names from and
 * set it up.  This routine doesn't actually get any names, it calls
 * sfileNext() to do that.
 */
char *
sfileFirst(char *cmd, char **Av, int Flags)
{
	sfileDone();
	prog = cmd;
	flags = Flags;
	if (Av[0]) {
		if (streq("-", Av[0])) {
			if (Av[1]) {
				fprintf(stderr,
				    "%s: - option must be alone.\n", prog);
				sfileDone();
				return (0);
			}
			flist = stdin;
			return (sfileNext());
		}
		if (isdir(Av[0])) {
			if (flags & NOEXPAND) return (0);
			sprintf(prefix, "%s/SCCS", Av[0]);
			if ((d = opendir(prefix))) {
				strcat(prefix, "/");
			} else {
				d = opendir(Av[0]);
				sprintf(prefix, "%s/", Av[0]);
			}
			return (sfileNext());
		} 
		av = Av;
		ac = 0;
		return (sfileNext());
	}
	if (flags & NOEXPAND) return (0);
	if (!d) {
		strcpy(prefix, "SCCS/");
		d = opendir("SCCS");
	}
	if (!d) {
		prefix[0] = 0;
		d = opendir(".");
	}
	return (sfileNext());
}

void
sfileUnget()
{
	unget = 1;
}

void
sfileDone()
{
	if (av) {
		av = 0;
		ac = 0;
	} else if (d) {
		closedir(d);
		d = 0;
	} else if (flist) {
		if (flist != stdin) fclose(flist);
		flist = 0;
	}
	prog = "";
}

private int
oksccs(char *sfile, int flags, int complain)
{
	char	*g;
	char	*s;

	if ((s = rindex(sfile, '/'))) {
		s++;
	} else { 
		s = sfile;
	}
	if (s[0] != 's' || s[1] != '.') {
		if (complain)
			verbose((stderr,
			    "%s: not an s.file: %s\n", prog, sfile));
		return (0);
	}
	g = sccs2name(sfile);
	if ((flags&GFILE) && access(g, R_OK) != 0) {
		if (complain) {
			if (access(sfile, R_OK) != 0) {
				verbose((stderr,
				    "%s: neither '%s' nor '%s' exists.\n",
				    prog, g, sfile));
			} else {
		    		verbose((stderr,
				    "%s: no such file: %s\n", prog, g));
			}
		}
		free(g);
		return (0);
	}
	if ((flags&WRITABLE) && access(g, W_OK) != 0) {
		if (complain)
		    	verbose((stderr,
			    "%s: %s: no write permission\n", prog, g));
		free(g);
		return (0);
	}
	free(g);
	return (1);
}

#ifdef        MAIN
main(int ac, char **av)
{
	char	*name;
	sccs	*s;

	for (name = sfileFirst(&av[1], 0); name; name = sfileNext()) {
		printf("NAME %s\n", name);
		if (s = sccs_init(name, 0)) {
			printf("\t%s\n\t%s\n",
			    s->sfile, s->gfile);
			sccs_free(s);
		}
	}
}
#endif

