/* 
 * SCCS library - all of SCCS is implemented here.  All the other source is
 * no more than front ends that call entry points into this library.
 * It's one big file so I can hide a bunch of stuff as statics (private).
 *
 * XXX - I never check memory allocation returns.  I need to do something
 * about this.
 *
 * Copyright (c) 1997-1998 Larry McVoy.  All rights reserved.
 */
#include "sccs.h"
WHATSTR("@(#)slib.c 1.139");

#define	EACH(line)	for (i=1; (line) && (i < (int)line[0]) && line[i]; ++i)

/*
 * XXX
 *
 * All things that get sent through fputs have to have only one newline in 
 * them.  Check this out.
 */
private	delta	*sfind(sccs *s, int serial);
private	delta	*rfind(sccs *s, char *rev);
private	void	dinsert(sccs *s, delta *d);
private	int	samebranch(delta *a, delta *b);
private	char	*sccsXfile(sccs *sccs, char type);
private	int	badcksum(sccs *s);
private	int	printstate(serlist *state, ser_t *slist);
private void	changestate(register serlist *state, char type, int serial);
private serlist *allocstate(serlist *old, int oldsize, int n);
private	void	end(sccs *, delta *, FILE *, int , int , int , int );
private	void	date(delta *d);
#ifndef	ANSIC
private	int	sig(int what, int sig);
#endif
private	int	flags(sccs *s, char *buf);
private	void	fputflag(delta *d, char flag);
private	void	addhost(delta *d, register char *host);
private	void	addpath(delta *d, register char *path);
private	void	addstate(delta *d, register char *state);
private	void	addzone(delta *d, register char *zone);
private void	inherit(int flags, delta *d);
private void	conflicts(sccs *s, delta *d);
private void	graft(delta *d);
private void	markbranches(delta *d);
private void	move(sccs *s, delta *d);
private void	linktree(delta *l, delta *r);
private void	fputsum(sccs *s, char *buf, FILE *out);
private void	putserlist(sccs *sc, ser_t *s, FILE *out);
private ser_t*	getserlist(sccs *sc, int isSer, char *s, int *ep);
private int	read_pfile(char *who, sccs *s, pfile *pf);

int
executable(char *f)
{
	return (access(f, X_OK) == 0);
}

int
isdir(char *s)
{
	struct	stat sbuf;

	if (stat(s, &sbuf) ==-1) return 0;
	return (S_ISDIR(sbuf.st_mode));
}

int
isreg(char *s)
{
	struct	stat sbuf;

	if (stat(s, &sbuf) ==-1) return 0;
	return (S_ISREG(sbuf.st_mode));
}

int
mkexecutable(FILE *f, char *fname)
{
	struct	stat sbuf;

#ifdef	ANSIC
	if (stat(fname, &sbuf)) {
		perror("stat");
		return(-1);
	}
	sbuf.st_mode |= S_IXUSR|S_IXGRP|S_IXOTH;
	return (chmod(fname, sbuf.st_mode & 07777));
#else
	if (fstat(fileno(f), &sbuf)) {
		perror("fstat");
		return(-1);
	}
	sbuf.st_mode |= S_IXUSR|S_IXGRP|S_IXOTH;
	return (fchmod(fileno(f), sbuf.st_mode & 07777));
#endif
}

int
Chmod(FILE *f, char *fname, mode_t mode)
{
#ifndef	ANSIC
	return (fchmod(fileno(f), mode));
#else
	return (chmod(fname, mode));
#endif
}

/*
 * Dup up to but not including the newline.
 */
private char	*
strnonldup(char *s)
{
	register char *t = s;
	int	len;

	while (*t++ && (t[-1] != '\n'));
	len = t - s;
	len--;
	t = malloc(len + 1);
	assert(t);
	strncpy(t, s, len);
	t[len] = 0;
	return (t);
}

void
chop(register char *s)
{
	while (*s++);
	s[-2] = 0;
}

/*
 * Return the length of the buffer until a newline.
 */
private int
linelen(char *s)
{
	char	*t = s;

	while (*t && (*t++ != '\n'));
	return (t-s);
}

/*
 * Save a line in an array.  If the array is out of space, reallocate it.
 * The size of the array is in array[0].
 */
char **
addLine(char **space, char *line)
{
	int	i;

	if (!space) {
		space = calloc(32, sizeof(char *));
		assert(space);
		space[0] = (char *)32;
	} else if (space[(int)space[0]-1]) {	/* full up, dude */
		int	size = (int)space[0];
		char	**tmp = calloc(size*2, sizeof(char*));

		assert(tmp);
		bcopy(space, tmp, size*sizeof(char*));
		tmp[0] = (char *)(size * 2);
		free(space);
		space = tmp;
	}
	for (i = 1; (i < (int)space[0]) && space[i]; ++i);
	assert(i < (int)space[0]);
	assert(space[i] == 0);
	space[i] = line;
	return (space);
}

private void
freeLines(char **space)
{
	int	i;

	if (!space) return;
	for (i = 1; (i < (int)space[0]) && space[i]; ++i) {
		free(space[i]);
	}
	space[0] = 0;
	free(space);
}

private int
removeLine(char **space, char *s)
{
	int	i, found, n = 0;

	do {
		for (found = 0, i = 1; (i < (int)space[0]) && space[i]; ++i) {
			if (streq(space[i], s)) {
				free(space[i]);
				while ((++i < (int)space[0]) && space[i]) {
					space[i-1] = space[i];
					space[i] = 0;
				}
				n++;
				found = 1;
				break;
			}
		}
	} while (found);
	return (n > 0);
}

/*
 * Compare up to and including the newline.  Both have to be on \n to match.
 */
private int
strnleq(register char *s, register char *t)
{
	while (*t == *s) {
		if (!*t || (*t == '\n')) return (1);
		t++, s++;
	}
	return (0);
}

private char	*
getuser(void)
{
	static	char	*s;

	if (s) return (s);
	s = getenv("USER");
	if (!s) {
		s = getlogin();
	}
	if (!s) {
		s = "UNKNOWN";
	}
	return (s);
}

/*
 * Convert a serial to an ascii string.
 */
void
sertoa(register char *buf, ser_t val)
{
	char    reverse[6];
	int     i, j;

	for (i = 0; val; i++, val /= 10) {
		reverse[i] = '0' + val % 10;
	}
	for (j = 0; i--; ) buf[j++] = reverse[i];
	buf[j] = 0;
}

#define	atoi	myatoi
private inline int
atoi(register char *s)
{
	register int val = 0;

	if (!s) return (0);
	while (*s && isdigit(*s)) {
		val = val * 10 + *s++ - '0';
	}
	return(val);
}

private inline int
atoi2(char **sp)
{
	register int val = 0;
	register char *s = *sp;

	if (!s) return (0);
	while (*s && isdigit(*s)) {
		val = val * 10 + *s++ - '0';
	}
	*sp = s;
	return(val);
}

/*
 * Free the delta tree.
 */
void
sccs_freetree(delta *tree)
{
	if (!tree) return;

	debug((stderr, "freetree(%s %s %d)\n",
	    tree->rev, tree->sdate, tree->serial));
	sccs_freetree(tree->siblings);
	sccs_freetree(tree->kid);
	if (tree->comments) {
		freeLines(tree->comments);
	}
	if (tree->rev) free(tree->rev);
	if (tree->user) free(tree->user);
	if (tree->sdate) free(tree->sdate);
	if (tree->include) free(tree->include);
	if (tree->exclude) free(tree->exclude);
	if (tree->ignore) free(tree->ignore);
	if (tree->hostname && !(tree->flags & DUPHOST)) free(tree->hostname);
	if (tree->pathname && !(tree->flags & DUPPATH)) free(tree->pathname);
	if (tree->zone && !(tree->flags & DUPZONE)) free(tree->zone);
	if (tree->state && !(tree->flags & DUPSTATE)) free(tree->state);
	free(tree);
}

/*
 * Insert the delta in the (ordered) tree.
 * A little weirdness when it comes to removed deltas, 
 * we want them off to the side if possible (it makes rfind work better).
 */
private void
dinsert(sccs *s, delta *d)
{
	delta	*p;

	debug((stderr, "dinsert(%s)", d->rev));
	if (!s->tree) {
		s->tree = d;
		s->lastinsert = d;
		debug((stderr, " -> ROOT\n")); 
	    	return;
	}
	if (s->lastinsert && (s->lastinsert->serial == d->pserial)) {
		p = s->lastinsert;
		debug((stderr, " (fast)"));
	} else {
		p = sfind(s, d->pserial);
	}
	assert(p);
	s->lastinsert = d;
	d->parent = p;
	if (!p->kid) {
		p->kid = d;
		debug((stderr, " -> %s (kid)\n", p->rev)); 
		return;
	}
	if ((p->kid->type == 'D') &&
	    samebranch(p, p->kid)) {	/* in right place */
		d->siblings = p->kid->siblings;
		p->kid->siblings = d;
		debug((stderr, " -> %s (sib)\n", p->rev)); 
	    	return;
	} /* else not in right place, put the new delta there. */
	debug((stderr, "kid type %c, %d.%d.%d.%d vs %d.%d.%d.%d\n",
	    p->kid->type, p->r[0], p->r[1], p->r[2], p->r[3],
	    p->kid->r[0], p->kid->r[1], p->kid->r[2], p->kid->r[3]));
	d->siblings = p->kid;
	p->kid = d;
	debug((stderr, " -> %s (kid, moved sib %s)\n",
	    p->rev, d->siblings->rev)); 
}

/*
 * Find the delta referenced by the serial number.
 */
private delta *
sfind(sccs *s, int serial)
{
	delta	*t;

	assert(serial <= s->numdeltas);
	if (serial >= s->ser2dsize) goto realloc;
	if (s->ser2delta && s->ser2delta[serial]) return (s->ser2delta[serial]);
	if (s->ser2delta) {
		for (t = s->table; t; t = t->next) {
			if (t->serial == serial) {
				assert(serial < s->ser2dsize);
				s->ser2delta[serial] = t;
				return (t);
			}
		}
		return (0);
	}

realloc:
	if (s->ser2delta) free(s->ser2delta);
	/* We leave a little extra room for sccs_delta. */
	s->ser2dsize = s->numdeltas+10;
	s->ser2delta = calloc(s->ser2dsize, sizeof(delta*));
	for (t = s->table; t; t = t->next) {
		if (t->serial >= s->ser2dsize) {
			fprintf(stderr, "%s %d %d\n",
			    s->sfile, t->serial, s->ser2dsize);
		}
		assert(t->serial < s->ser2dsize);
		s->ser2delta[t->serial] = t;
	}
	return (s->ser2delta[serial]);
}

private char days[13] = { 0, 31,28,31,30,31,30,31,31,30,31,30,31 };

/*
 * XXX - this doesn't work for years < 1752 but I fail to see that as an
 * issue for source management :-)
 */
int
monthDays(int year, int month)
{
	if (month != 2) return (days[month]);
	if ((!((year) % 4) && ((year) % 100)) || !((year) % 400)) return (29);
	return (28);
}

private time_t
getDate(delta *d)
{
	if (!d->date) d->date = sccs_date2time(d->sdate, d->zone, EXACT);
	return (d->date);
}

/*
 * Take 93/07/25 21:14:11 and return a time_t that is useful.
 * For years < 70, assume 20xx which means add 100.
 *
 * roundup is ROUNDDOWN, EXACT, or ROUNDUP.  Which do the implied
 * adjustments for the unspecified fields.  Fields which are incorrectly
 * specified, i.e., 31 for a month that has 30 days, are truncated back
 * down to legit values.
 *
 * This is a little weird because tm_ does not translate directly from
 * yy/mm/dd hh:mm:ss - see mktime(3) for details.
 */
time_t
sccs_date2time(char *asctime, char *z, int roundup)
{
	struct	tm tm;
	int	sw;
	int	neg = 0;
	int	mday;
	extern	long timezone;

#define	gettime(field) \
	if (isdigit(asctime[1])) { \
		tm.field = ((asctime[0] - '0') * 10 + (asctime[1] - '0')); \
		asctime += 2; \
	} else { \
		tm.field = *asctime++ - '0'; \
	} \
	for ( ; *asctime && !isdigit(*asctime); asctime++); 

	/* this is weird, but all I care about here is up or !up */
	roundup = (roundup == ROUNDUP);
	bzero(&tm, sizeof(tm));
	tm.tm_mday = 1;
	if (roundup) {
		tm.tm_mday = 31;
		tm.tm_mon = 11;
		tm.tm_hour = 23;
		tm.tm_min = tm.tm_sec = 59;
	}
	/* Adjust for year 200 problems */
	gettime(tm_year); if (tm.tm_year < 70) tm.tm_year += 100;
	unless (*asctime) goto correct;

	/* tm_mon counts 0..11; we count 1..12 */
	gettime(tm_mon); tm.tm_mon--; unless (*asctime) goto correct;
	gettime(tm_mday); unless (*asctime) goto correct;
	gettime(tm_hour); unless (*asctime) goto correct;
	gettime(tm_min); unless (*asctime) goto correct;
	gettime(tm_sec);
	if (((asctime[-1] == '-') || (asctime[-1] == '+')) && !z) z = --asctime;

correct:
	/*
	 * Truncate down oversized fields.
	 */
	if (tm.tm_mon > 11) tm.tm_mon = 11;
	mday = monthDays(tm.tm_year, tm.tm_mon + 1);
	if (mday < tm.tm_mday) tm.tm_mday = mday;
	if (tm.tm_hour > 23) tm.tm_hour = 23;
	if (tm.tm_min > 59) tm.tm_min = 59;
	if (tm.tm_sec > 59) tm.tm_sec = 59;

	/*
	 * If they didn't give us a correction, assume localtime.
	 * This will do the wrong thing for stuff created in a 
	 * different timezone or a different daylight savings time,
	 * but that's what they get for using a bad SCCS implementation.
	 */
	if (!z) {
		struct	tm *tm;
		time_t	tt = time(0);
		static	time_t last_tm = 0;
		static	long int mytimezone = -1;
	
		/*
		 * This bit of weirdness is so that we don't call localtime()
		 * over and over again.  GNU libc wacks the timezone variable
		 * as a side effect of mktime() (I think, it's getting wacked
		 * somewhere), so I cache it.  But I refresh the cache once an
		 * hour.
		 */
		if ((tt - last_tm) > 3600) {
			last_tm = tt;
			tm = localtime(&tt);
			mytimezone = timezone;
		}
		sw = mytimezone;
	} else {
		if ((neg = (*z == '-'))) z++;
		if (*z == '+') z++;
#define	N(c)	(c - '0')
		sw = atoi(z) * 3600;
		assert(sw <= (24 * 3600));
		while (*z++ != ':');
		assert((atoi(z) * 60) <= 3600);
		sw += atoi(z) * 60;
		if (neg) sw = -sw;
	}
#if 0
	{
	struct	tm tm2 = tm;
	fprintf(stderr, "%02d/%02d/%02d %02d:%02d:%02d%c%02d:%02d = %u\n",
	    tm.tm_year,
	    tm.tm_mon + 1,
	    tm.tm_mday,
	    tm.tm_hour,
	    tm.tm_min,
	    tm.tm_sec, 
	    sw < 0 ? '-' : '+',
	    ((sw < 0 ? -sw : sw) / 60) / 60,
	    ((sw < 0 ? -sw : sw) / 60) % 60,
	    mktime(&tm2) + sw
	    );
	}
#endif
	return (mktime(&tm) + sw);
}

/*
 * Diff can give me
 *	10a12,14	-> 10 a
 *	10,12d13	-> 10 d
 *	10,12c12,14	-> 10 c
 */
private inline int
scandiff(char *s, int *where, char *what)
{
	if (!isdigit(*s)) return (-1);
	*where = atoi2(&s);
	if (*s == ',') {
		s++;
		(void)atoi2(&s);
	}
	if (*s != 'a' && *s != 'c' && *s != 'd') {
		return (-1);
	}
	*what = *s;
	return (0);
}

/*
 * Convert ascii 1.2.3.4 -> 1, 2, 3, 4
 */
private inline int
scanrev(char *s, int *a, int *b, int *c, int *d)
{
	if (!isdigit(*s)) return (0);
	*a = atoi2(&s);
	if (b && *s == '.') {
		s++;
		*b = atoi2(&s);
		if (c && *s == '.') {
			s++;
			*c = atoi2(&s);
			if (d && *s == '.') {
				s++;
				*d = atoi(s);
				return (4);
			} else return (3);
		} else return (2);
	} else return (1);
}

private inline void
explode_rev(delta *d)
{
	register char *s = d->rev;
	int	dots = 0;

	while (*s) { if (*s++ == '.') dots++; }
	if (dots > 3) d->flags |= DERROR|BADREV;
	d->r[0] = d->r[1] = d->r[2] = d->r[3] = 0;
	scanrev(d->rev,&d->r[0],&d->r[1],&d->r[2],&d->r[3]);
}

/*
 * This one assumes SCCS style branch numbering, i.e., x.y.z.d
 */
private int
samebranch(delta *a, delta *b)
{
	if (!a->r[2] && !b->r[2]) return (1);
	return ((a->r[0] == b->r[0]) &&
		(a->r[1] == b->r[1]) &&
		(a->r[2] == b->r[2]));
}

/*
 * XXX - why isn't this just return (d->rev)????
 */
private char *
revname(delta *d)
{
	int	a1 = 0, a2 = 0, a3 = 0, a4 = 0;
	static	char buf[25];

	scanrev(d->rev, &a1, &a2, &a3, &a4);
	if (a3) {
		sprintf(buf, "%d.%d.%d.%d", a1,a2,a3,a4);
	} else {
		sprintf(buf, "%d.%d", a1,a2);
	}
	return (buf);
}

private char *
branchname(delta *d)
{
	int	a1 = 0, a2 = 0, a3 = 0, a4 = 0;
	static	char buf[6];

	scanrev(d->rev, &a1, &a2, &a3, &a4);
	if (a3) {
		sprintf(buf, "%d", a3);
	} else {
		sprintf(buf, "0");
	}
	return (buf);
}

char *
basenm(char *s)
{
	char	*t;

	for (t = s; *t; t++);
	do {
		t--;
	} while (*t != '/' && t > s);
	if (*t == '/') t++;
	return (t);
}

private inline int
samefile(char *a, char *b)
{
	struct	stat sa, sb;

	if (stat(a, &sa) == -1) return 0;
	if (stat(b, &sb) == -1) return 0;
	return ((sa.st_dev == sb.st_dev) && (sa.st_ino == sb.st_ino));
}

/*
 * Translate SCCS/s.foo.c to /u/lm/smt/sccs/SCCS/s.foo.c
 *
 * XXX - this is slow.  
 * XXX - I made it faster by looking at $PWD.  This works as long as we
 * only do operations on `pwd`/SCCS.
 */
char *
fullname(sccs *s, int withsccs)
{
	static	char new[1024];
	char	here[1024];
	char	*t;

	assert(s);
	assert(s->sfile);
	assert(s->gfile);

	/*
	 * If they have a full path name, then just use that.
	 */
	if (s->gfile[0] == '/') {
		return (withsccs ? s->sfile : s->gfile);
	}

	if ((strncmp("SCCS/", s->sfile, 5) == 0) &&
	    (t = getenv("PWD")) && samefile(".", t)) {
	    	sprintf(new, "%s/%s", t, withsccs ? s->sfile : s->gfile);
		return (new);
	}

	getcwd(here, sizeof(here));

	/*
	 * If there is no slash in gfile, it's here.
	 */
	if (!strchr(s->gfile, '/')) {
		strcpy(new, here);
		strcat(new, "/");
		strcat(new, withsccs ? s->sfile : s->gfile);
		return (new);
	}

	/*
	 * We have a partial name like foo/bar/blech or
	 * ../fs/ufs/SCCS/bmap.c
	 */
	for (t = s->gfile; *t; t++);
	while (*--t != '/');
	*t = 0;
	if (chdir(s->gfile)) {
		/* shouldn't happen, if it does, we have bigger problems */
		perror(s->gfile);
		return (0);
	} else {
		getcwd(new, sizeof(new));
		chdir(here);
	}
	*t = '/';
	strcat(new, t);
	return (new);
}

/*
 * Revision number/branching theory of operation:
 *	rfind() does an exact match - use this if you want a specific delta.
 *	findrev() does an inexact match - it will find top of trunk for the
 *		trunk or a branch if it is specified as -r1 or -r1.2.1.
 *		Note that -r1 will not find tot if tot is 2.1; use a null
 *		rev to get that.
 *	getedit() finds the revision and gets the new revision number for a
 *		new delta.  The find logic is findrev().  The new revision
 *		will be the next available on that line, unless there is
 *		a conflict or the forcebranch flag is set.  In either of
 *		those cases, the new revision will be a branch.
 *	        If the revision is on the trunk and they wanted to bump
 *		the release, do so.
 */
private inline int
samerev(int a[4], int b[4])
{
	return ((a[0] == b[0]) &&
		(a[1] == b[1]) &&
		(a[2] == b[2]) &&
		(a[3] == b[3]));
}

private	int	R[4];
/*
 * This one uses the globals set up below.  This hack makes this library
 * !MT safe.  Which it wasn't anyway.
 */
private delta *
_rfind(delta *d)
{
	delta	*t;

	debug((stderr,
	    "\t_rfind(%d.%d.%d.%d)\n",d->r[0],d->r[1],d->r[2],d->r[3]));
	if (samerev(d->r, R)) {
		return (d);
	}
	if (d->kid && (t = _rfind(d->kid))) {
		return (t);
	}
	if (d->siblings && (t = _rfind(d->siblings))) {
		return (t);
	}
	return (0);
}

private inline int
name2rev(sccs *s, char **rp)
{
	char	*rev = *rp;
	symbol	*sym;

	if (!rev) return (0); 
	if (isdigit(rev[0])) return (0);
again:	for (sym = s->symbols; sym; sym = sym->next) {
		if (streq(rev, sym->name)) {
			rev = sym->rev;
			if (isdigit(rev[0])) {
				*rp = rev;
				return (0);
			}
			goto again;
		}
	}
	return (-1);
}

private delta *
sym2delta(sccs *s, char *sym)
{
	if (name2rev(s, &sym) == -1) return (0);
	return (rfind(s, sym));
}

/*
 * Find the delta referenced by rev.  It must be an exact match.
 */
private delta *
rfind(sccs *s, char *rev)
{
	delta	*d;

	debug((stderr, "rfind(%s) ", rev));
	name2rev(s, &rev);
	R[0] = R[1] = R[2] = R[3] = 0;
	scanrev(rev, &R[0], &R[1], &R[2], &R[3]);
	debug((stderr, "aka %d.%d.%d.%d\n", R[0],R[1],R[2],R[3]));
	d = _rfind(s->tree);
	debug((stderr, "rfind(%s) got %s\n", rev, d ? d->rev : "Not found"));
	return (d);
}

/*
 * Find the specified delta.  The rev can be
 *	(null)	get top of the default branch
 *	d	get top of trunk for that release
 *	d.d	get that revision or error if not there
 *	d.d.d	get top of branch that matches those three digits
 *	d.d.d.d	get that revision or error if not there.
 *	XXX - symbolic names
 */
delta *
findrev(sccs *s, char *rev)
{
    	int	a = 0, b = 0, c= 0, d = 0;
	delta	*e;
	char	buf[20];

	if (!s->tree) return (0);
	if (!rev || !*rev) rev = s->defbranch ? s->defbranch : "100000";
	if (name2rev(s, &rev)) return (0);
	switch (scanrev(rev, &a,&b,&c,&d)) {
	    case 1:
		for (e = s->tree;
		    e->kid && (e->kid->type == 'D') && 
		    (e->kid->r[2] == 0) &&
		    (e->kid->r[0] <= a);
		    e = e->kid)
			;
		debug((stderr, "findrev(%s) =  %s\n", rev, e->rev));
		return (e);
	    case 2:
	    case 4:
		e = rfind(s, rev);	/* well formed */
		debug((stderr,
		    "findrev(%s) =  %s\n", rev, e ? e->rev : "Not found"));
		return (e);
	    case 3:
		/* XXX - should this be a while loop, decrementing the 
		 * third digit until it find it???  Nahh.
		 */
		sprintf(buf, "%s.1", rev);
		unless (e = rfind(s, buf)) { /* well formed */
			debug((stderr, "findrev(%s) =  Not found\n", rev));
			return (0);
		}
		for ( ;
		    e->kid && e->kid->type == 'D' && samebranch(e, e->kid);
		    e = e->kid)
			;
		debug((stderr, "findrev(%s) =  %s\n", rev, e->rev));
		return (e);
	    default:
		fprintf(stderr, "Malformed revision: %s\n", rev);
		debug((stderr, " BAD\n", e->rev));
		return (0);
	}
}

/*
 * Find the first delta (in this LOD) that is > date.
 *
 * XXX - if there are two children in the same LOD, this finds only one of them.
 */
private delta *
findDate(delta *d, time_t date)
{
	time_t	d2;
	delta	*tmp, *tmp2;
	
	if (!d) return (0);
	d2 = DATE(d);
	if (d2 >= date) return (d);
	for (tmp = d->kid; tmp; tmp = tmp->siblings) {
		if (samebranch(tmp, d) && (tmp2 = findDate(tmp, date))) {
			return (tmp2);
		}
	}
	return (0);
}

/*
 * Take either a revision or date/symbol and return the delta.
 *
 * All tokens may have a prefix or a postfix of "," or ".".
 * Dots are inclusive, commas are exclusive.  
 * A, means first thing after A but 
 * .B means up to and including "B".
 * Dates are rounded down if postfixed and up if prefixed.  If there is
 * neither a postfix or a prefix, then the roundup variable is used.
 *
 * XXX - TODO [rev rev) notation
 */
delta *
sccs_getrev(sccs *sc, char *rev, char *dateSym, int roundup)
{
	delta	*tmp, *d = 0;
	time_t	date, d2;
	char	*s = rev ? rev : dateSym;
	char	*last;
	int	inclusive = 1;
	char	pre = 0, post = 0;

	/*
	 * Allow null revisions to mean TOT.
	 */
	if (!s || !*s) {
		return (findrev(sc, 0));
	}

	/*
	 * Strip off any prefix/postfix.
	 * XXX - I don't error check for ".foo."
	 */
	for (last = s; last[1]; last++);
	if ((s[0] == ',') || (s[0] == '.')) {
		roundup = ROUNDUP;
		inclusive = s[0] == '.';
		pre = s[0];
		s++;
	} else if ((last[0] == ',') || (last[0] == '.')) {
		roundup = ROUNDDOWN;
		inclusive = last[0] == '.';
		post = last[0];
		last[0] = 0;
	} else if ((s[0] == '(') || (s[0] == '[')) {
		roundup = ROUNDDOWN;
		inclusive = s[0] == '[';
		pre = s[0];
		s++;
	} else if ((last[0] == ')') || (last[0] == ']')) {
		roundup = ROUNDUP;
		inclusive = last[0] == ']';
		post = last[0];
		last[0] = 0;
	}

	/*
	 * If it's a revision, go find it and use it, being careful to
	 * go down the same branch if we need to go forward.
	 * XXX - if 1.5 is TOT and we ask for ,1.6 this fails.
	 */
	if (rev) {
		unless (d = rfind(sc, s)) {
			if (post) last[0] = post;
			return (0);
		}
		if (post) last[0] = post;
		if (inclusive) return (d);
		if (roundup == ROUNDUP) {
			d = d->parent;
			return (d);
		}
		for (tmp = d->kid;
		    tmp && !samebranch(tmp, d); tmp = tmp->siblings);
		return (tmp);
	}
	
	/*
	 * Get a date to work with.  If it's a symbol, convert that.
	 */
	if (isdigit(*s)) {
		date = sccs_date2time(s, 0, roundup);
	} else {
		delta	*d = sym2delta(sc, s);

		if (post) last[0] = post;
		if (!d) return(0);
		if (d->date) {
			date = d->date;
		} else {
			date = DATE(d);
		}
	}
	if (post) last[0] = post;

	if (roundup == ROUNDUP) {
		/*
		 * Go get TOT (XXX - needs to respect LOD markers)
		 */
		d = findrev(sc, 0);	
		if (!inclusive) date--;	

		/*
		 * We are working backwards from dates that are larger than
		 * what we have.  
		 */
		for (;;) {
			if (!d) return (0);
			d2 = DATE(d);
			if (d2 <= date) {
				return (d);
			}
			d = d->parent;
		}
	} else if (roundup == ROUNDDOWN) {
		/*
		 * We are looking for the first delta after date.  
		 */
		if (!inclusive) date++;	
		return (findDate(sc->tree, date));
	} else /* EXACT */ {
		/*
		 * Go get TOT (XXX - needs to respect LOD markers)
		 */
		d = findrev(sc, 0);	
		for (;;) {
			if (!d) return (0);
			d2 = DATE(d);
			if (d2 == date) return (d);
			d = d->parent;
		}
	}
}

private inline int
isbranch(delta *d)
{
	return (d->r[2]);
}

private inline int
morekids(delta *d)
{
	return (d->kid && (d->kid->type != 'R') && samebranch(d, d->kid));
}

/*
 * Get the delta that is the basis for this edit.
 * Get the revision name of the new delta.
 */
private delta *
getedit(sccs *s, char **revp, int branch)
{
	char	*rev = *revp;
    	int	n, a = 0, b = 0, c= 0, d = 0;
	delta	*e, *t;
	static	char buf[20];

	/*
	 * use the findrev logic to get to the delta.
	 */
	unless (e = findrev(s, rev)) {
		/*
		 * Special case: we're at 1.x and they want 2.1 or 3.1.
		 * We allow that.
		 */
		scanrev(rev,&a,&b,&c,&d);
		if ((c == 0) && (b == 1)) {
			e = findrev(s, 0);
			assert(e);
			if (e->r[0] < a) goto ok;
		}
		return (0);
	}
ok:
	/*
	 * Just continue trunk/branch
	 * Because the kid may be a branch, we have to be extra careful here.
	 */
	if (!branch && !morekids(e)) {
		if (scanrev(e->rev, &a,&b,&c,&d) == 2) { 
			/* Seems weird but makes -e -r2 -> 2.1 when tot is 1.x
			 */
			int	release = rev ? atoi(rev) : 1;

			if (release > a) { 
				a = release;
				b = 1;
			} else {
				b++;
			}
			sprintf(buf, "%d.%d", a,b);
		} else {
			sprintf(buf, "%d.%d.%d.%d",a,b,c,d+1);
		}
		debug((stderr, "getedit(%s) -> %s\n", rev, buf));
		*revp = buf;
		return (e);
	}

	/* 
	 * For whatever reason (they asked, or there is a kid in the way),
	 * we need a branch.
	 * Branches are all based, in their /name/, off of the closest
	 * trunk node.  Go backwards up the tree until we hit the trunk
	 * and then use that rev as a basis.
	 */
	for (t = e; isbranch(t); t = t->parent) ;
	n = scanrev(t->rev, &a,&b,0,0);
	c = 1;
	do {
		sprintf(buf, "%d.%d.%d.1", a, b, c);
		c++;
	} while (rfind(s, buf));	/* well formed */
	debug((stderr, "getedit(%s) -> %s\n", rev, buf));
	*revp = buf;
	return (e);
}

#ifndef	USE_STDIO
private char	*
fastnext(sccs *s)
{
	register char *t = s->where;
	register char *tmp = s->mmap + s->size;

	if (s->where >= tmp) return (0);
	/* I tried unrolling this a couple of ways and it got worse */
	while (t < tmp && *t++ != '\n');
	tmp = s->where;
	s->where = t;
	return (tmp);
}
#endif

/*
 * This does standard SCCS expansion, it's almost 100% here.
 * New stuff added:
 * lm	user@host 
 */
private char *
expand(sccs *s, delta *d, char *l)
{
	static	char buf[1024];
	char	*t = buf;
	char	*tmp;
	time_t	now = 0;
	struct	tm *tm;
	int	a[4];

	while (*l != '\n') {
		if (l[0] != '%' || l[1] == '\n' || l[2] != '%') {
			*t++ = *l++;
			continue;
		}
		switch (l[1]) {
		    case 'A':	/* @(#)%Y% slib.c 1.139@(#) */
			/* XXX - no tflag */
			strcpy(t, "@(#) "); t += 5;
			tmp = basenm(s->gfile);
			strcpy(t, tmp); t += strlen(tmp); *t++ = ' ';
			tmp = revname(d); strcpy(t, tmp); t += strlen(tmp);
			strcpy(t, "@(#)"); t+= 4;
			break;

		    case 'B':	/* branch name: XXX */
			tmp = branchname(d); strcpy(t, tmp); t+= strlen(tmp);
			break;

		    case 'C':	/* line number - XXX */
			*t++ = '%'; *t++ = 'C'; *t++ = '%'; break;

		    case 'D':	/* today: 97/06/22 */
			if (!now) { time(&now); tm = localtime(&now); }
			assert(tm);
			if (s->state & YEAR4) {
				int	y = tm->tm_year;

				if (y < 70) y += 2000; else y += 1900;
				sprintf(t, "%4d/%02d/%02d",
				    y, tm->tm_mon+1, tm->tm_mday);
				t += 10;
			} else {
				sprintf(t, "%02d/%02d/%02d",
				    tm->tm_year, tm->tm_mon+1, tm->tm_mday);
				t += 8;
			}
			break;
			
		    case 'E':	/* most recent delta: 97/06/22 */
			if (s->state & YEAR4) {
				if (atoi(d->sdate) > 69) {
					*t++ = '1'; *t++ = '9';
				} else {
					*t++ = '2'; *t++ = '0';
				}
			}
			strncpy(t, d->sdate, 8); t += 8;
			break;

		    case 'F':	/* s.file name */
			strcpy(t, "SCCS/"); t += 5;
			tmp = basenm(s->sfile);
			strcpy(t, tmp); t += strlen(tmp);
			break;

		    case 'G':	/* most recent delta: 06/22/97 */
			*t++ = d->sdate[3]; *t++ = d->sdate[4]; *t++ = '/';
			*t++ = d->sdate[6]; *t++ = d->sdate[7]; *t++ = '/';
			if (s->state & YEAR4) {
				if (atoi(d->sdate) > 69) {
					*t++ = '1'; *t++ = '9';
				} else {
					*t++ = '2'; *t++ = '0';
				}
			}
			*t++ = d->sdate[0]; *t++ = d->sdate[1]; 
			break;

		    case 'H':	/* today: 06/22/97 */
			if (!now) { time(&now); tm = localtime(&now); }
			assert(tm);
			if (s->state & YEAR4) {
				int	y = tm->tm_year;

				if (y < 70) y += 2000; else y += 1900;
				sprintf(t, "%4d/%02d/%02d",
				    y, tm->tm_mon+1, tm->tm_mday);
				sprintf(t, "%02d/%02d/%04d",
				    tm->tm_mon+1, tm->tm_mday, y);
				t += 10;
			} else {
				sprintf(t, "%02d/%02d/%02d",
				    tm->tm_mon+1, tm->tm_mday, tm->tm_year);
				t += 8;
			}
			break;

		    case 'I':	/* name of revision: 1.1 or 1.1.1.1 */
			tmp = revname(d); strcpy(t, tmp); t += strlen(tmp);
			break;

		    case 'L':	/* 1.2.3.4 -> 2 */
			scanrev(d->rev, &a[0], &a[1], 0, 0);
			sprintf(t, "%d", a[1]); t+= strlen(t);
			break;

		    case 'M':	/* mflag or filename: slib.c */
			tmp = basenm(s->gfile);
			strcpy(t, tmp); t += strlen(tmp);
			break;
			
		    case 'P':	/* full: /u/lm/smt/sccs/SCCS/s.slib.c */
			tmp = fullname(s, 1);
			strcpy(t, tmp); t += strlen(tmp);
			break;

		    case 'Q':	/* qflag */
			*t++ = '%'; *t++ = 'Q'; *t++ = '%'; break;

		    case 'R':	/* release 1.2.3.4 -> 1 */
			scanrev(d->rev, &a[0], 0, 0, 0);
			sprintf(t, "%d", a[0]); t+= strlen(t);
			break;

		    case 'S':	/* rev number: 1.2.3.4 -> 4 */
			a[3] = 0;
			scanrev(d->rev, &a[0], &a[1], &a[2], &a[3]);
			sprintf(t, "%d", a[3]); t+= strlen(t);
			break;

		    case 'T':	/* time: 23:04:04 */
			if (!now) { time(&now); tm = localtime(&now); }
			assert(tm);
			sprintf(t, "%02d:%02d:%02d",
			    tm->tm_hour, tm->tm_min, tm->tm_sec);
			t += 8;
			break;

		    case 'U':	/* newest delta: 23:04:04 */
			strcpy(t, &d->sdate[9]); t += 8;
			break;

		    case 'W':	/* @(#)slib.c 1.139: @(#)slib.c 1.1 */
			strcpy(t, "@(#) "); t += 4;
			tmp = basenm(s->gfile);
			strcpy(t, tmp); t += strlen(tmp); *t++ = ' ';
			tmp = revname(d); strcpy(t, tmp); t += strlen(tmp);
			break;

		    case 'Y':	/* tflag */
			*t++ = '%'; *t++ = 'Y'; *t++ = '%'; break;

		    case 'Z':	/* @(#) */
			strcpy(t, "@(#)"); t += 4; break;

		    case '@':	/* user@host */
		    	strcpy(t, d->user);
			t += strlen(d->user);
			if (d->hostname) {
				*t++ = '@';
				strcpy(t, d->hostname);
				t += strlen(d->hostname);
			}
			break;

		    default:	t[0] = l[0];
				t[1] = l[1];
				t[2] = l[2];
				t += 3;
				break;
		}
		l += 3;
	} 
	*t++ = '\n'; *t = 0;
	return (buf);
}

/*
 * This does standard RCS expansion.
 * Keywords to expand:
 *	$Revision$
 *	$Id$
 *	$Author$
 *	$Date$
 *	$Header$
 *	$Locker$
 *	$Log$		Not done 
 *	$Name$
 *	$RCSfile$
 *	$Source$
 *	$State$
 */
private char *
rcsexpand(sccs *s, delta *d, char *l)
{
	static	char buf[1024];
	char	*t = buf;
	char	*tmp;
	delta	*h;

	while (*l != '\n') {
		if (l[0] != '$') {
			*t++ = *l++;
			continue;
		}
		if (strneq("$Author$", l, 8)) {
			strcpy(t, "$Author: "); t+= 9;
		    	strcpy(t, d->user);
			t += strlen(d->user);
			for (h = d; h && !h->hostname; h = h->parent);
			if (h && h->hostname) {
				*t++ = '@';
				strcpy(t, h->hostname);
				t += strlen(h->hostname);
			}
			*t++ = ' ';
			*t++ = '$';
			l += 8;
		} else if (strneq("$Date$", l, 6)) {
			strcpy(t, "$Date: "); t+= 7;
			strncpy(t, d->sdate, 17); t += 17;
			for (h = d; h && !h->zone; h = h->parent);
			if (h && h->zone) {
				*t++ = '-';
				strcpy(t, h->zone);
				t += strlen(h->zone);
			}
			*t++ = ' ';
			*t++ = '$';
			l += 6;
		} else if (strneq("$Header$", l, 8)) {
			strcpy(t, "$Header: "); t+= 9;
			tmp = s->sfile;
			strcpy(t, tmp); t += strlen(tmp); *t++ = ' ';
			tmp = revname(d); strcpy(t, tmp); t += strlen(tmp);
			*t++ = ' ';
			strncpy(t, d->sdate, 17); t += 17;
			for (h = d; h && !h->zone; h = h->parent);
			if (h && h->zone) {
				*t++ = '-';
				strcpy(t, h->zone);
				t += strlen(h->zone);
			}
			*t++ = ' ';
		    	strcpy(t, d->user);
			t += strlen(d->user);
			for (h = d; h && !h->hostname; h = h->parent);
			if (h && h->hostname) {
				*t++ = '@';
				strcpy(t, h->hostname);
				t += strlen(h->hostname);
			}
			*t++ = ' ';
			for (h = d; h && !h->state; h = h->parent);
			if (h && h->state) {
				*t++ = ' ';
				strcpy(t, h->state);
				t += strlen(h->state);
				*t++ = ' ';
			}
			*t++ = '$';
			l += 8;
		} else if (strneq("$Id$", l, 4)) {
			strcpy(t, "$Id: "); t+= 5;
			tmp = basenm(s->sfile);
			strcpy(t, tmp); t += strlen(tmp); *t++ = ' ';
			tmp = revname(d); strcpy(t, tmp); t += strlen(tmp);
			*t++ = ' ';
			strncpy(t, d->sdate, 17); t += 17;
			for (h = d; h && !h->zone; h = h->parent);
			if (h && h->zone) {
				*t++ = '-';
				strcpy(t, h->zone);
				t += strlen(h->zone);
			}
			*t++ = ' ';
		    	strcpy(t, d->user);
			t += strlen(d->user);
			for (h = d; h && !h->hostname; h = h->parent);
			if (h && h->hostname) {
				*t++ = '@';
				strcpy(t, h->hostname);
				t += strlen(h->hostname);
			}
			*t++ = ' ';
			for (h = d; h && !h->state; h = h->parent);
			if (h && h->state) {
				*t++ = ' ';
				strcpy(t, h->state);
				t += strlen(h->state);
				*t++ = ' ';
			}
			*t++ = '$';
			l += 4;
		} else if (strneq("$Locker$", l, 8)) {
			strcpy(t, "$Locker: <Not implemented> $"); t+= 28;
			l += 8;
		} else if (strneq("$Log$", l, 5)) {
			strcpy(t, "$Log: <Not implemented> $"); t+= 25;
			l += 5;
		} else if (strneq("$Name$", l, 6)) {
			strcpy(t, "$Name: <Not implemented> $"); t+= 26;
			l += 6;
		} else if (strneq("$RCSfile$", l, 9)) {
			strcpy(t, "$RCSfile: "); t += 10;
			tmp = basenm(s->sfile);
			strcpy(t, tmp); t += strlen(tmp);
			*t++ = ' '; *t++ = '$';
			l += 9;
		} else if (strneq("$Revision$", l, 10)) {
			strcpy(t, "$Revision: "); t+= 11;
			tmp = revname(d); strcpy(t, tmp); t += strlen(tmp);
			*t++ = ' ';
			*t++ = '$';
			l += 10;
		} else if (strneq("$Source$", l, 8)) {
			strcpy(t, "$Source: "); t += 9;
			tmp = fullname(s, 1);
			strcpy(t, tmp); t += strlen(tmp);
			*t++ = ' '; *t++ = '$';
			l += 8;
		} else if (strneq("$State$", l, 7)) {
			strcpy(t, "$State: "); t += 8;
			for (h = d; h && !h->state; h = h->parent);
			*t++ = ' ';
			if (h && h->state) {
				strcpy(t, h->state);
				t += strlen(h->state);
			} else {
				strcpy(t, "<unknown>");
				t += 9;
			}
			*t++ = ' '; *t++ = '$';
			l += 7;
		} else {
			*t++ = *l++;
		}
	} 
	*t++ = '\n'; *t = 0;
	return (buf);
}

/*
 * We couldn't initialize an SCCS file.  Tell the user why not.
 * Possible reasons we handle:
 *	no s.file
 *	bad checksum
 *	couldn't build the graph
 *
 * XXX - use this for 100% of the error messages.
 */
void
sccs_whynot(char *who, sccs *s)
{
	if (BEEN_WARNED(s)) return;
	unless (HAS_SFILE(s)) {
		fprintf(stderr, "%s: No such file: %s\n", who, s->sfile);
		return;
	}
	unless (s->cksumok) {
		fprintf(stderr, "%s: bad checksum in %s\n", who, s->sfile);
		return;
	}
	unless (s->tree) {
		fprintf(stderr, "%s: couldn't decipher delta table in %s\n",
		    who, s->sfile);
		return;
	}
	if (HAS_ZFILE(s)) {
		fprintf(stderr, "%s: %s is zlocked\n", who, s->gfile);
		return;
	}
	if (HAS_PFILE(s)) {
		fprintf(stderr, "%s: %s is edited\n", who, s->gfile);
		return;
	}
	fprintf(stderr, "%s: %s: unknown error.\n", who, s->sfile);
}

/*
 * Read in an sfile and build the delta table graph.
 * Scanf caused problems here when there are corrupted files.
 */
private void
mkgraph(sccs *s)
{
	delta	*d = 0, *t;
	char	rev[100], date[9], time[9], user[100];
	char	*p;
	char	tmp[100];
	int	i;
	int	line = 1;
	char	*expected = "?";
	BUF(	buf);

	seekto(s, 0);
	next(buf, s);			/* checksum */
	line++;
	debug((stderr, "mkgraph(%s)\n", s->sfile));
	for ( ;; ) {
nextdelta:	unless (next(buf, s)) {
bad:
			fprintf(stderr,
			    "%s: bad delta on line %d, expected `%s', "
			    "line follows:\n\t",
			    s->sfile, line, expected);
			fprintf(stderr, "``%.*s''\n", linelen(buf)-1, buf);
			sccs_freetree(s->table);
			s->table = 0;
			return;
		}
		line++;
		if (strneq(buf, "\001u\n", 3)) break;
		t = calloc(sizeof(*t), 1);
		assert(t);
		s->numdeltas++;
		t->kid = d;
		if (d)
			d->next = t;
		else
			s->table = t;
		d = t;
		/* ^As 00001/00000/00011 */
		if (buf[0] != '\001' || buf[1] != 's' || buf[2] != ' ') {
			expected = "^As ";
			goto bad;
		}
		d->added = atoi(&buf[3]);
		d->deleted = atoi(&buf[9]);
		d->same = atoi(&buf[15]);
		/* ^Ad D 1.2.1.1 97/05/15 23:11:46 lm 4 2 */
		/* ^Ad R 1.2.1.1 97/05/15 23:11:46 lm 4 2 */
		next(buf, s);
		line++;
		if (buf[0] != '\001' || buf[1] != 'd' || buf[2] != ' ') {
			expected = "^Ad ";
			goto bad;
		}
	    /* D|R */
		d->type = buf[3];
	    /* 1.1 or 1.2.2.2, etc. */
		p = &buf[5];
		for (i = 0, p = &buf[5]; *p != ' '; rev[i++] = *p++);
		rev[i] = 0;
		if (*p != ' ') { expected = "^AD 1.1 "; goto bad; }
	    /* 98/03/17 */
		for (i = 0, p++; *p != ' '; date[i++] = *p++);
		date[i] = 0;
		if (*p != ' ') { expected = "^AD 1.1 98/03/17 "; goto bad; }
	    /* 18:32:39[.mmm] */
		for (i = 0, p++; *p != ' ' && *p != '.'; time[i++] = *p++);
		time[i] = 0;
		if (*p != ' ') {
			expected = "^AD 1.1 98/03/17 18:32:39";
			goto bad;
		}
	    /* user */
		for (i = 0, p++; *p != ' '; user[i++] = *p++);
		user[i] = 0;
		if (*p != ' ') {
			expected = "^AD 1.1 98/03/17 18:32:39 user ";
			goto bad;
		}
		p++;
	    /* 10 11 */
		d->serial = atoi2(&p);
		if (*p != ' ') {
			expected = "^AD 1.1 98/03/17 18:32:39 user 12 ";
			goto bad;
		}
		p++;
		d->pserial = atoi(p);
		if (d->serial >= s->nextserial) s->nextserial = d->serial + 1;
		debug((stderr, "mkgraph(%s)\n", rev));
		d->rev = strdup(rev);
		explode_rev(d);
		if (d->flags & BADREV) {
			expected = "1.2 or 1.2.3.4, too many dots";
			goto bad;
		}
		sprintf(tmp, "%s %s", date, time);
		d->sdate = strdup(tmp);
		d->user = strdup(user);
		for ( ;; ) {
			if (!next(buf, s) || buf[0] != '\001') {
				expected = "^A";
				goto bad;
			}
			line++;
			switch (buf[1]) {
			    case 'c':
				i = 0; goto comment;
			    case 'i':
			    	d->include = getserlist(s, 1, &buf[3], 0);
			    	break;
			    case 'x':	
			    	d->exclude = getserlist(s, 1, &buf[3], 0);
			    	break;
			    case 'g':	
			    	d->ignore = getserlist(s, 1, &buf[3], 0);
				break;
			    case 'e':
				goto nextdelta;
			    default:
				fprintf(stderr, "Bad file format: ");
				expected = "^A{c,i,x,g,e}";
			    	goto bad;
			}
		}
		/* ^Ac branch. */
		for ( ;; ) {
			if (!next(buf, s) || buf[0] != '\001') {
				expected = "^A";
				freeLines(d->comments);
				goto bad;
			}
			line++;
comment:		switch (buf[1]) {
			    case 'e': goto done;
			    case 'c':
				d->comments = 
				    addLine(d->comments, strnonldup(&buf[3]));
				break;
			    default:
			    	expected = "^A{e,c}";
				freeLines(d->comments);
				goto bad;
			}
		}
done:		;
	}

	/* The deltas are all here, typically lowest to highest */
	s->tree = d;
	d = d->kid;
	s->tree->kid = 0;
	while (d) {
		delta	*therest = d->kid;

		assert(d->serial <= s->numdeltas);
		d->kid = 0;
		dinsert(s, d);
		d = therest;
	}
	debug((stderr, "mkgraph() done\n"));
}

/*
 * Read the stuff after the ^au (which has already been read).
 * ^au
 * user
 * user
 * group name (ATT SCCS allows only numbers, maybe we should too)
 * ^aU
 * ^af whatever
 * ^at
 * Descriptive text (pass through).
 * ^aT
 */
private int
misc(sccs *s)
{
	BUF(	buf);

	/* Save the users / groups list */
	for ( ; next(buf, s) && !strneq(buf, "\001U\n", 3); ) {
		if (buf[0] == '\001') {
			fprintf(stderr, "%s: corrupted user section.\n",
			    s->sfile);
			return (-1);
		}
		s->usersgroups = addLine(s->usersgroups, strnonldup(buf));
	}
			
	/* Save the flags.  Some are handled by the flags routine; those
	 * are not handled here.
	 */
	for ( ; next(buf, s) && !strneq(buf, "\001t\n", 3); ) {
		if (strneq(buf, "\001f R\n", 5)) {	/* XXX - obsolete */
			s->state |= RCS;
			continue;
		} else if (strneq(buf, "\001f b\n", 5)) {
			s->state |= BRANCHOK;
			continue;
		} else if (strneq(buf, "\001f Y\n", 5)) {/* XXX - obsolete */
			s->state |= YEAR4;
			continue;
		} else if (strneq(buf, "\001f x ", 5)) {
			int	bits = atoi(&buf[5]);

			if (bits & X_YEAR4) s->state |= YEAR4;
			if (bits & X_RCSEXPAND) s->state |= RCS;
		} else if (strneq(buf, "\001f &", 4) ||
		    strneq(buf, "\001f z ", 5)) {
#ifndef	USE_STDIO
			s->landingpad = buf;
#endif
			continue;
		} else if (strneq(buf, "\001f e ", 5)) {
			switch (atoi(&buf[5])) {
			    case E_ASCII:
			    case E_UUENCODE:
			    case E_UUGZIP:
			    	s->encoding = atoi(&buf[5]);
				break;
			    default:
			    	fprintf(stderr,
				    "sccs: don't know encoding %d, "
				    "assuming ascii\n",
				    atoi(&buf[5]));
			    	s->encoding = E_ASCII;
				return (-1);
			}
			continue;
		} 
		if (!flags(s, buf)) {
			s->flags = addLine(s->flags, strnonldup(&buf[3]));
		}
	}
			
	/* Save descriptive text. */
	for ( ; next(buf, s) && !strneq(buf, "\001T\n", 3); ) {
		s->text = addLine(s->text, strnonldup(buf));
	}
	s->data = tell(s);
	return (0);
}

/*
 * Reads until a space without a preceeding \.
 * XXX - counts on s[-1] being a valid address.
 */
private char *
gettok(char **sp)
{
	static char	copy[1024];
	char	*t;
	char	*s = *sp;

	t = copy;
	*t = 0;
	while ((*t = *s++)) {
		if (*t == '\n') {
			*t = 0;
			*sp = s;
			return (copy);
		}
		if (*t == ' ') {
			if (t[-1] == '\\') {
				t--;
				*t = ' ';
			} else {
				*t = 0;
				*sp = s;
				return (copy);
			}
		}
		t++;
	}
	*sp = s;
	return (copy);
}

/*
 * Look at the flag and if we handle in this routine, return true.
 * Otherwise, return false so it gets passed through.
 *
 * We handle here:
 *	b	<rev> default branch
 *	h	<rev> <host> hostname where checkin happened.
 *	p	<rev> <path> pathname of the revision.
 *	s	<rev> <symbol> symbolic name for the rev.
 *	S	<rev> <state> Named state of the revision.
 *	w	<rev> <hh:mm> minutes west of GMT where checkin happened.
 */
private int
flags(sccs *s, char *buf)
{
	char	*p = &buf[5];
	char	f = buf[3];
	char	*t;
	delta	*d;
	symbol	*sym;
	char	rev[MAXREV];

	if (buf[0] != '\001' || buf[1] != 'f' || buf[2] != ' ') return (0);
	t = gettok(&p);
	assert(t);
	if ((f != 's') && (f != 'd') &&
	    (!s->tree || !(d = rfind(s, t)))) {
		/* ignore it. but keep it. */
		return (0);
	}
	if (f == 'd') {
		if (s->defbranch) free(s->defbranch);
		s->defbranch = strdup(t);
		return (1);
	}
	strcpy(rev, t);
	t = gettok(&p);
	if (!t) return (0);
	switch (f) {
	    case 'h':	/* hostname */
		addhost(d, t);
		return (1);
	    case 'p':	/* pathname */
		addpath(d, t);
		return (1);
	    case 's':	/* symbol */
		sym = calloc(1, sizeof(*sym));
		assert(sym);
		/*
		 * OK, this is the new way, to be consistent with the others.
		 * But I accept the old way.
		 */
		if (isdigit(rev[0])) {
			sym->rev = strdup(rev);
			sym->name = strdup(t);
		} else {
			sym->name = strdup(rev);
			sym->rev = strdup(t);
		}
		sym->next = s->symbols;
		s->symbols = sym;
		return (1);
	    case 'S':	/* state */
		addstate(d, t);
		return (1);
	    case 'w':	/* minutes west of GMT */
		addzone(d, t);
		return (1);
	}
	return (0);
}

private void
addhost(delta *d, register char *host)
{
	register delta	*p;

	if (d->flags & NOHOST) return;
	for (p = d->parent; p; p = p->parent) {
		if (p->hostname) {
			if (streq(p->hostname, host)) return;
			break;
		}
	}
	if (d->hostname) free(d->hostname);
	d->hostname = strnonldup(host); return;
}

private void
addpath(delta *d, register char *path)
{
	register delta	*p;

	if (d->flags & NOPATH) return;
	for (p = d->parent; p; p = p->parent) {
		if (p->pathname) {
			if (streq(p->pathname, path)) return;
			break;
		}
	}
	if (d->pathname) free(d->pathname);
	d->pathname = strnonldup(path); return;
}

private void
addstate(delta *d, register char *state)
{
	register delta	*p;

	if (d->flags & NOSTATE) return;
	for (p = d->parent; p; p = p->parent) {
		if (p->state) {
			if (streq(p->state, state)) return;
			break;
		}
	}
	if (d->state) free(d->state);
	d->state = strnonldup(state); return;
}

private void
addzone(delta *d, register char *zone)
{
	register delta	*p;

	if (d->flags & NOZONE) return;
	for (p = d->parent; p; p = p->parent) {
		if (p->zone) {
			if (streq(p->zone, zone)) return;
			break;
		}
	}
	if (d->zone) free(d->zone);
	d->zone = strnonldup(zone); return;
}

sccs *
check_gfile(sccs *s, int flags)
{
	struct	stat sbuf;

	if (stat(s->gfile, &sbuf) == 0) {
		if (!S_ISREG(sbuf.st_mode)) {
			verbose((stderr, "Not a regular file: %s\n", s->gfile));
			free(s->gfile);
			free(s->sfile);
			free(s);
			return (0);
		}
		s->state |= GFILE;
		if (sbuf.st_mode & 0222) s->state |= WRITABLE;
		if (sbuf.st_mode & 0111) s->state |= GEXECUTABLE;
	} else {
		s->state &= ~(GFILE|WRITABLE|GEXECUTABLE);
	}
	return (s);
}

/*
 * Initialize an SCCS file.  Do this before anything else.
 * If the file doesn't exist, the graph isn't set up.
 * It should be OK to have multiple files open at once.
 */
sccs*
sccs_init(char *name, int flags)
{
	sccs	*s = calloc(1, sizeof(*s));
	struct	stat sbuf;

	assert(s);
	if (is_sccs(name)) {
		s->sfile = strdup(name);
		s->gfile = sccs2name(name);
	} else {
		fprintf(stderr, "Not an SCCS file: %s\n", name);
		free(s);
		return (0);
	}
	unless (check_gfile(s, flags)) return (0);
	if (stat(s->sfile, &sbuf) == 0) {
		if (!S_ISREG(sbuf.st_mode)) {
			verbose((stderr, "Not a regular file: %s\n", s->sfile));
			free(s->gfile);
			free(s->sfile);
			free(s);
			return (0);
		}
		if (sbuf.st_size == 0) {
			verbose((stderr, "Zero length file: %s\n", s->sfile));
			free(s->gfile);
			free(s->sfile);
			free(s);
			return (0);
		}
		s->state |= SFILE;
#ifndef	USE_STDIO
		s->size = sbuf.st_size;
#endif
		if (sbuf.st_mode & 0111) s->state |= SEXECUTABLE;
	}
	s->pfile = strdup(sccsXfile(s, 'p'));
	s->zfile = strdup(sccsXfile(s, 'z'));
	if (isreg(s->pfile)) s->state |= PFILE;
	if (isreg(s->zfile)) s->state |= ZFILE;
	debug((stderr, "init(%s) -> %s, %s\n", name, s->sfile, s->gfile));
	s->nextserial = 1;
#ifdef	USE_STDIO
	s->file = fopen(s->sfile, "r");
	if (!s->file) {
		s->cksumok = 1;
		return(s);
	}
#else
	if (flags & MAP_WRITE) {
		sbuf.st_mode |= 0200;
		chmod(s->sfile, sbuf.st_mode & 0777);
		s->fd = open(s->sfile, 2, 0);
		s->mmap = mmap(0,
		    s->size, PROT_READ|PROT_WRITE, MAP_SHARED, s->fd, 0);
		s->state |= CHMOD;
	} else {
		s->fd = open(s->sfile, 0, 0);
		s->mmap = mmap(0, s->size, PROT_READ, MAP_SHARED, s->fd, 0);
	}
	if ((int)s->mmap == -1) {
		s->cksumok = 1;
		return(s);
	}
	debug((stderr, "mapped %s for %d at 0x%x\n",
	    s->sfile, s->size, s->mmap));
#endif
	s->state |= SOPEN;
	if (((flags&NOCKSUM) == 0) && badcksum(s)) {
		fprintf(stderr, "Bad checksum for %s\n", s->sfile);
		return (s);
	} else {
		s->cksumok = 1;
	}
	mkgraph(s);
	debug((stderr, "mkgraph found %d deltas\n", s->numdeltas));
	if (s->tree) {
		if (misc(s)) {
			sccs_free(s);
			return (0);
		}
		if (flags & (INHERIT|INHERIT_DATE)) {
			if (!s->tree->pathname && !(flags & NODEFAULT)) {
				s->tree->pathname = strdup(fullname(s, 0));
			}
			inherit(flags, s->tree);
		}
	}
#ifdef	ANSIC
	signal(SIGINT, SIG_IGN);
#else
	sig(CATCH, SIGINT);
	sig(BLOCK, SIGINT);
#endif
	return (s);
}

/*
 * Restart an sccs_init because we've changed the state of the file.
 *
 * This does not reread the delta table, if you want that, open and close the
 * file.
 */
sccs*
sccs_restart(sccs *s)
{
	struct	stat sbuf;

	assert(s);
	if (stat(s->gfile, &sbuf) == 0) {
		if (!S_ISREG(sbuf.st_mode)) {
			free(s->gfile);
			free(s->sfile);
			free(s);
			return (0);
		}
		s->state |= GFILE;
		if (sbuf.st_mode & 0222) s->state |= WRITABLE;
		if (sbuf.st_mode & 0111) s->state |= GEXECUTABLE;
	}
	if (stat(s->sfile, &sbuf) == 0) {
		if (!S_ISREG(sbuf.st_mode)) {
			free(s->gfile);
			free(s->sfile);
			free(s);
			return (0);
		}
		if (sbuf.st_size == 0) {
			free(s->gfile);
			free(s->sfile);
			free(s);
			return (0);
		}
		s->state |= SFILE;
		if (sbuf.st_mode & 0111) s->state |= SEXECUTABLE;
	}
#ifndef	USE_STDIO
	if (s->size != sbuf.st_size) {
		BUF(	buf);

		if (s->mmap) munmap(s->mmap, s->size);
		s->size = sbuf.st_size;
		s->mmap = mmap(0, s->size, PROT_READ, MAP_SHARED, s->fd, 0);
		seekto(s, 0);
		for ( ; next(buf, s) && !strneq(buf, "\001T\n", 3); );
		s->data = tell(s);
	}
#endif
	if (isreg(s->pfile)) s->state |= PFILE;
	if (isreg(s->zfile)) s->state |= ZFILE;
	return (s);
}

/*
 * Free up all resources associated with the file.
 * This is the last thing you do.
 */
void
sccs_free(sccs *s)
{
	symbol	*sym, *t;

	assert(s);
	assert(s->sfile);
	assert(s->gfile);
	sccsXfile(s, 0);
	if (s->tree) sccs_freetree(s->tree);
	for (sym = s->symbols; sym; sym = t) {
		t = sym->next;
		if (sym->name) free(sym->name);
		if (sym->rev) free(sym->rev);
		free(sym);
	}
	free(s->sfile);
	free(s->gfile);
	free(s->zfile);
	free(s->pfile);
#ifdef	USE_STDIO
	if (s->file) fclose(s->file);
#else
	if (s->mmap) munmap(s->mmap, s->size);
	if (s->state & CHMOD) {
		struct	stat sbuf;

		if (fstat(s->fd, &sbuf) == 0) {
			sbuf.st_mode &= ~0200;
#ifdef	ANSIC
			chmod(s->sfile, sbuf.st_mode & 0777);
#else
			fchmod(s->fd, sbuf.st_mode & 0777);
#endif
		}
	}
#endif
	close(s->fd);
	if (s->defbranch) free(s->defbranch);
	if (s->ser2delta) free(s->ser2delta);
	freeLines(s->usersgroups);
	freeLines(s->flags);
	freeLines(s->text);
	free(s);
#ifdef	ANSIC
	signal(SIGINT, SIG_DFL);
#else
	sig(UNCATCH, SIGINT);
	sig(UNBLOCK, SIGINT);
#endif
}

/*
 * We want SCCS/s.foo or path/to/SCCS/s.foo
 * ATT allows s.foo or path/to/s.foo.
 * Since we do a pretty good job of putting stuff in SCCS/s.*, we skip
 * that check here.
 */
int
is_sccs(char *name)
{
	char	*s = rindex(name, '/');

	if (!s) {
		if (name[0] == 's' && name[1] == '.') return (1);
		return (0);
	}
	if (s[1] == 's' && s[2] == '.') return (1);
	return (0);
}

/*
 * remove SCCS/s. or just the s.
 * It's up to the caller to free() the resulting name.
 */
char *
sccs2name(char *sfile)
{
	char	*s, *t;
	char	*new = strdup(sfile);

	s = rindex(new, '/');
	if (!s) {
		assert(new[0] == 's' && new[1] == '.');
		t = new;
		s = &new[2];
	} else {
		assert(s[1] == 's' && s[2] == '.');
		if ((s >= (new + 4)) &&
		    (s[-1] == 'S') && (s[-2] == 'C') &&
		    (s[-3] == 'C') && (s[-4] == 'S')) {
			t = &s[-4];
			s = &s[3];
		} else {
			t = &s[1];
			s = &s[3];
		}
	}
	while ((*t++ = *s++));
	return (new);
}

/*
 * Make the sccs dir if we need one.
 */
void
mksccsdir(sccs *sc)
{
	char	*s = rindex(sc->sfile, '/');

	if (!s) return;
	if ((s >= sc->sfile + 4) &&
	    s[-1] == 'S' && s[-2] == 'C' && s[-3] == 'C' && s[-4] == 'S') {
	    	*s = 0;
		mkdir(sc->sfile, 0775);
		*s = '/';
	}
}

/*
 * Take a file name such as foo.c and return SCCS/s.foo.c
 * Also works for /full/path/foo.c -> /fullpath/SCCS/s.foo.c.
 * It's up to the caller to free() the resulting name.
 */
char *
name2sccs(char *name)
{
	int	len = strlen(name);
	char	*s, *newname;

	/* maybe it has the SCCS in it already */
	s = rindex(name, '/');
	if ((s >= name + 4) && strneq(s - 4, "SCCS/", 5)) {
	    	return (strdup(name));
	}
	newname = malloc(len + 8);
	assert(newname);
	strcpy(newname, name);
	if ((s = rindex(newname, '/'))) {
		s++;
		strcpy(s, "SCCS/s.");
		s += 7;
		strcpy(s, rindex(name, '/') + 1);
	} else {
		strcpy(s = newname, "SCCS/s.");
		s += 7;
		strcpy(s, name);
	}
	return (newname);
}

/*
 * create SCCS/<type>.foo.c
 */
private int
lock(sccs *sccs, char type)
{
	char	*s;
	int	islock;

	s = sccsXfile(sccs, type);
	islock = open(s, O_CREAT|O_WRONLY|O_EXCL, type == 'z' ? 0444 : 0644);
	close(islock);
	debug((stderr, "lock(%s) = %d\n", sccs->sfile, islock > 0));
	return (islock > 0);
}

/*
 * Take SCCS/s.foo.c and unlink SCCS/<type>.foo.c
 */
private int
unlock(sccs *sccs, char type)
{
	char	*s;

	debug((stderr, "unlock(%s, %c)\n", sccs->sfile, type));
	s = sccsXfile(sccs, type);
	return (unlink(s));
}

/*
 * Take SCCS/s.foo.c, type and return a temp copy of SCCS/<type>.foo.c
 */
private char *
sccsXfile(sccs *sccs, char type)
{
	static	char	*s;
	static	int	len;
	char	*t;

	if (type == 0) {	/* clean up so purify doesn't barf */
		if (len) free(s);
		len = 0;
		return (0);
	}
	if (!len) {
		len = strlen(sccs->sfile) + 50;
		s = malloc(len + 50);
		assert(s);
	} else if (len < strlen(sccs->sfile) + 3) {
		free(s);
		len = strlen(sccs->sfile) + 50;
		s = malloc(len);
		assert(s);
	}
	if (!index(sccs->sfile, '/')) {
		strcpy(s, sccs->sfile);
		s[0] = type;
		return (s);
	}
	strcpy(s, sccs->sfile);
	t = rindex(s, '/') + 1;
	*t = type;
	return (s);
}

/*
 * Get the date as YY/MM/DD HH:MM:SS.mmm 
 * and get timezone as minutes west of GMT
 */
private void
date(delta *d)
{
	struct	timeval tv;
	struct	timezone tz;
	struct	tm *tm;
	time_t	tt;
	char	tmp[50];
	extern	long timezone;

	gettimeofday(&tv, &tz);
	tt = (time_t)tv.tv_sec;
	tm = localtime(&tt);
	strftime(tmp, sizeof(tmp), "%y/%m/%d %H:%M:%S", tm);
	d->sdate = strdup(tmp);
	sprintf(tmp, "%02d:%02d", (int)(timezone/3600), (int)(timezone%3600));
	d->zone = strdup(tmp);

}

private char *
now()
{
	struct	timeval tv;
	struct	timezone tz;
	struct	tm *tm;
	time_t	tt;
	static	char	tmp[50];

	gettimeofday(&tv, &tz);
	tt = (time_t)tv.tv_sec;
	tm = localtime(&tt);
	strftime(tmp, sizeof(tmp), "%y/%m/%d %H:%M:%S", tm);
	return (tmp);
}

private char	*
gethost(void)
{
	static	char host[100];
	static	done = 0;
	struct	utsname uts;
	char	*s;

	if (done) return(host[0] ? host : 0);
	host[0] = 0;
	bzero(&uts, sizeof(uts));
	uname(&uts);
	if (!uts.nodename) return (0);
	strcpy(host, uts.nodename);
	/* bail out early if it looks like a domain is there */
	for (s = host; *s; s++) {
		if (*s == '.') return (host);
	}
#ifdef	HAVE_GETDOMAINNAME
	getdomainname(s, sizeof(host) - (s-host));
	if (*s && !streq(s, "(none)")) s[-1] = '.';
	if (streq(host, "localhost") || streq(host, "localhost.localdomain")) {
		host[0] = 0;
		return(0);
	}
#endif
	return (host);
}

/*
 * Save a serial in an array.  If the array is out of space, reallocate it.
 * The size of the array is in array[0].
 */
ser_t *
addSerial(ser_t *space, ser_t s)
{
	int	i;

	if (!space) {
		space = calloc(16, sizeof(ser_t));
		assert(space);
		space[0] = (ser_t)16;
	} else if (space[(int)space[0]-1]) {	/* full up, dude */
		int	size = (int)space[0];
		ser_t	*tmp = calloc(size*2, sizeof(ser_t));

		assert(tmp);
		bcopy(space, tmp, size*sizeof(ser_t));
		tmp[0] = (ser_t)(size * 2);
		free(space);
		space = tmp;
	}
	for (i = 1; (i < (int)space[0]) && space[i]; ++i);
	assert(i < (int)space[0]);
	assert(space[i] == 0);
	space[i] = s;
	return (space);
}

/*
 * Parse the list of revisions passed, sort of like strtok().
 * The first call passes in the list, the next passes in 0.
 * Handle rev,rev,rev,rev-rev,rev...
 * "rev" can be a symbol.
 * We're careful to have the list unchanged if we return an error.
 */
private delta *
walkList(sccs *s, char *list, int *errp)
{
	static	char *next;	/* left pointing at next rev */
	static	delta *d;	/* set if we are working up the tree */
	static	delta *stop;	/* where to stop in the tree, inclusive */
	delta	*tmp;
	char	save, *t, *rev;

	if (list) {
		next = list;
		d = stop = 0;
	}
	if (d) {
		tmp = d;
		d = d == stop ? 0 : d->parent;
		return (tmp);
	}
	if (!next) return (0);
	/* XXX - this will screw up if they have symbols w/ - in the name */
	for (t = next; *t && *t != ',' && *t != '-'; t++);
	if (!*t || (*t == ',')) {
		save = *t;
		*t++ = 0;
		rev = next;
		next = save ? t : 0;
		name2rev(s, &rev);
		tmp = rfind(s, rev);
		t[-1] = save;
		return (tmp);
	}
	if (*t != '-') {
		*errp = 1;
		return (0);
	}
	/*
	 * OK, it's a range.  Find the bottom/top and and start walking them.
	 * We insist that the ordering is 1.3-1.9, not the other way, and 
	 * that there is a path from the second rev to the first through
	 * the parent pointers (i.e., it doesn't go around a corner down
	 * a branch).
	 */
	*t++ = 0;
	rev = next;
	next = t;
	name2rev(s, &rev);
	stop = rfind(s, rev);
	t[-1] = '-';
	for (t = next; *t && *t != ',' && *t != '-'; t++);
	if (*t && (*t != ',')) {
		*errp = 1;
		return (0);
	}
	save = *t;
	*t++ = 0;
	rev = next;
	next = save ? t : 0;
	name2rev(s, &rev);
	d = rfind(s, rev);
	t[-1] = save;

	if (!stop || !d) {
		*errp = 2;
		return (0);
	}
	for (tmp = d; tmp && tmp != stop; tmp = tmp->parent);
	if (tmp != stop) {
		*errp = 2;
		return (0);
	}
	tmp = d;
	d = d == stop ? 0 : d->parent;
	return (tmp);
}

/*
 * Read the list of serials in s and put them in list.
 * S can either be list of serials or a list of revisions.
 * In neither case does this recurse on the list, it just parses the list.
 * This is quite different than serialmap() which uses this routine but
 * stores the results in a different data structure.
 * XXX - this poorly named, it doesn't use "struct serlist".
 */
private ser_t *
getserlist(sccs *sc, int isSer, char *s, int *ep)
{
	delta	*t;
	ser_t	*l = 0;

	debug((stderr, "getserlist(%.*s)\n", linelen(s), s));
	if (isSer) {
		while (*s && *s != '\n') {
			l = addSerial(l, atoi(s));
			while (*s && *s != '\n' && isdigit(*s)) s++;
			while (*s && *s != '\n' && !isdigit(*s)) s++;
		}
		return (l);
	}
	for (t = walkList(sc, s, ep); !*ep && t; t = walkList(sc, 0, ep)) {
		l = addSerial(l, t->serial);
	}
	return (l);
}

/*
 * XXX - this poorly named, it doesn't use "struct serlist".
 */
private void
putserlist(sccs *sc, ser_t *s, FILE *out)
{
	int	first = 1, i;
	char	buf[20];

	if (!s) return;
	/* This is not EACH because I want to go backwards */
	for (i = (int)s[0] - 1; i > 0; i--) {
		if (s[i]) {
			sertoa(buf, s[i]);
			if (!first) fputsum(sc, " ", out);
			fputsum(sc, buf, out);
			first = 0;
		}
	}
}

/*
 * Include/exclude processing.   
 *
 * The temporary list has two elements, {ser, what}.
 * "ser" is the serial number that has caused "what" to be set.
 * "what" is 0, S_INC, or S_EXCL.
 * When adding serial numbers, if there is a conflict, the higher number
 * wins since that one was a more recent addition to the file.
 * This should work in all cases except, perhaps, after a smoosh.
 */
private void
inex(sccs *s, ser_t doer, ser_t ser, ielist *list, int what)
{
	delta	*t;
	int	i;

	debug((stderr, "inex(%s, %d, %d, %s)\n",
	    s->gfile, doer, ser, what == S_INC ? "inc" : "excl"));
	if (!list[ser].ser) {
		list[ser].ser = doer;
		list[ser].what = what;
	} else if (list[ser].ser < doer) {
		list[ser].ser = doer;
		list[ser].what = what;
	}
	t = sfind(s, ser);
	EACH(t->include) {
		inex(s, t->serial, t->include[i], list, what);
	}
	what = what == S_INC ? S_EXCL : S_INC;
	EACH(t->exclude) {
		inex(s, t->serial, t->exclude[i], list, what);
	}
}

/*
 * Generate a list of serials to use to get a particular delta and
 * allocate & return space with the list in the space.
 * The 0th entry is contains the maximum used entry unless it got excluded.
 * Note that the error pointer is to be used only by walkList, it's null if
 * the lists are null.
 * Note we don't have to worry about growing tables here, the list isn't saved
 * across calls.
 */
private ser_t *
serialmap(sccs *s, delta *d, int flags, char *iLst, char *xLst, int *errp)
{
	ser_t	*slist;
	ielist	*ie;
	delta	*t;
	int	i;

	ie = calloc(s->nextserial, sizeof(ielist));
	assert(ie);
	assert(d);
	for (t = d; t; t = t->parent) {
		assert(t->serial <= s->numdeltas);
		inex(s, t->serial, t->serial, ie, S_INC);
	}
	if (iLst) {
		verbose((stderr, "Included:"));
	    	for (t = walkList(s, iLst, errp); 
		    !*errp && t; t = walkList(s, 0, errp)) {
			verbose((stderr, " %s", t->rev));
			inex(s, s->nextserial, t->serial, ie, S_INC);
	    	}
		verbose((stderr, "\n"));
		if (*errp) goto bad;
	}
	if (xLst) {
		verbose((stderr, "Excluded:"));
	    	for (t = walkList(s, xLst, errp); 
		    !*errp && t; t = walkList(s, 0, errp)) {
			verbose((stderr, " %s", t->rev));
			inex(s, s->nextserial, t->serial, ie, S_EXCL);
	    	}
		verbose((stderr, "\n"));
		if (*errp) goto bad;
	}
	for (i = s->nextserial; i--; ) {
		if (ie[i].ser) break;
	}
	assert(i >= 0);
	slist = calloc(s->nextserial, sizeof(ser_t));
	debug((stderr, "map %s (max=%d): ", d->rev, i));
	slist[0] = i;
	for ( ; i > 0; i--) {
		if (ie[i].what == S_INC) {
			slist[i] = 1;
			debug((stderr, "[%d] ", i));
		}
	}
	debug((stderr, "\n"));
	free(ie);
	return (slist);
bad:	free(slist);
	return (0);
}

private void
changestate(register serlist *state, char type, int serial)
{
	register serlist *s;
	register serlist *n;
	int	i;

	debug2((stderr, "chg(%c, %d)\n", type, serial));
	/*
	 * Find the item and delete it if it is an 'E'.
	 */
	if (type == 'E') {	/* free this item */

		for (i=0, s = state[SLIST].next; s; s = s->next, i++) {
			if (s->serial == serial) break;
		}
		assert(s && (s->serial == serial));
		if (s->prev) {
			s->prev->next = s->next;
		} else {
			state[SLIST].next = s->next;
		}
		if (s->next) {
			s->next->prev = s->prev;
		}
		s->next = state[SFREE].next;
		state[SFREE].next = s;
		return;
	} 
	assert(state[SFREE].serial);
	n = state[SFREE].next;
	state[SFREE].next = n->next;
	for (i=0, s = state[SLIST].next; s; s = s->next, i++) {
		if ((s->serial > serial) || !s->next) break;
	}
	/*
	 * We're either at the head of the (empty) list,
	 * at the right place (insert before s), or
	 * or we're at the end of the list (insert after).
	 */
	if (!s) {
		state[SLIST].next = n;
		n->prev = 0;
		n ->next = 0;
	} else if (s->serial > serial) {
		n->next = s;
		if (s->prev) {
			s->prev->next = n;
		} else {
			state[SLIST].next = n;
		}
		n->prev = s->prev;
		s->prev = n;
	} else {
		assert(!s->next);
		n->next = 0;
		n->prev = s;
		s->next = n;
	}
	n->serial = serial;
	n->type = type;
}

/*
 * Allocate (or realocate) a serial list array.  The way this works is like
 * so:
 * [SLIST] is the allocated list, linked through .next and .prev.
 * [SLIST].serial is the number of allocated nodes.
 * [SFREE] is the free list, linked only through .next.
 * [SFREE].serial is the number of free nodes.
 */
private serlist *
allocstate(serlist *old, int oldsize, int n)
{
	serlist	*s;
	int	i;

	assert(!old);	/* XXX - realloc not done */
	s = malloc(n * sizeof(serlist));
	assert(s);
	for (i = 2; i < n-1; ++i) {
		s[i].next = &s[i+1];
		s[i].prev = 0;
	}
	s[SFREE].next = &s[2];
	s[SFREE].serial = n-2;
	s[SLIST].next = 0;
	s[SLIST].serial = 0;
	return (s);
}

private int
printstate(serlist *state, ser_t *slist)
{
	register serlist *s;
	int	ourI = 0, theirI = 0, ourD = 0;

	for (s = state[SLIST].next; s; s = s->next) {
		if (s->type == 'I') {
			if (slist[s->serial]) { 
				ourI = s->serial;
			} else {
				theirI = s->serial;
			}
		}
		else if ((s->type == 'D') && slist[s->serial]) ourD = s->serial;
	}
	/* we are in delete mode, it matters not what the rest are doing */
	debug2((stderr, "{oI %d tI %d oD %d} ", ourI, theirI, ourD));
	if (ourD >= ourI) {
		debug2((stderr, " = 0\n"));
		return 0;
	}
	if (ourI > theirI) {
		debug2((stderr, " = %d\n", ourI));
		return ourI;
	}
	debug2((stderr, " = 0\n"));
	return 0;
}

private void inline
fnlputs(char *buf, FILE *out)
{
	register char	*t = buf;
	char	fbuf[1024];
	register char *p = fbuf;

	do {
		*p++ = *t;
		if (p == &fbuf[1023]) {
			*p = 0;
			p = fbuf;
			fputs(fbuf, out);
		}
	} while (*t && (*t++ != '\n'));
	if (p != fbuf) {
		*p = 0;
		fputs(fbuf, out);
	}
}

private void 
fputsum(sccs *s, char *buf, FILE *out)
{
	register char	*t = buf;
	register unsigned short sum = s->cksum;
	char	fbuf[1024];
	register char *p = fbuf;

	for (; *t; t++) {
		sum += *t;
		*p++ = *t;
		if (p == &fbuf[1023]) {
			*p = 0;
			p = fbuf;
			fputs(fbuf, out);
		}
		if (*t == '\n') break;
	}
	s->cksum = sum;
	if (p != fbuf) {
		*p = 0;
		fputs(fbuf, out);
	}
}

typedef	unsigned char uchar;
#define ENC(c)  ((((uchar)c) & 0x3f) + ' ')
#define DEC(c)  ((((uchar)c) - ' ') & 0x3f)

inline int
uuencode1(register uchar *from, register char *to, int n)
{
	int     space[4];
	register int *c = space;
	register int i;
	char	*save = to;

	*to++ = ENC(n);
	for (i = 0; i < n; i += 3) {
		c[0] = from[i] >> 2;
		c[1] = ((from[i]<<4)&0x30) | ((from[i+1]>>4)&0xf);
		c[2] = ((from[i+1]<<2)&0x3c) | ((from[i+2]>>6)&3);
		c[3] = from[i+2] & 0x3f;
		*to++ = ENC(c[0]);
		*to++ = ENC(c[1]);
		*to++ = ENC(c[2]);
		*to++ = ENC(c[3]);
	}
	*to++ = '\n';
	*to = 0;
	return (to - save);
}

int
uuencode_sum(sccs *s, FILE *in, FILE *out)
{
	uchar	ibuf[450];
	char	obuf[80];
	register uchar *p;
	register int n;
	register int length;
	int	added = 0;

	while ((length = fread(ibuf, 1, 450, in)) > 0) {
		p = ibuf;
		while (length > 0) {
			n = (length > 45) ? 45 : length;
			length -= n;
			uuencode1(p, obuf, n);
			fputsum(s, obuf, out);
			p += n;
			added++;
		}
	}
	fputsum(s, " \n", out);
	return (++added);
}

int
uuencode(FILE *in, FILE *out)
{
	uchar	ibuf[450];
	char	obuf[650];
	register uchar *buf;
	register char *p = obuf;
	register int n;
	register int length;
	int	added = 0;

	while ((length = fread(ibuf, 1, 450, in)) > 0) {
		p = obuf;
		buf = ibuf;
		while (length > 0) {
			n = (length > 45) ? 45 : length;
			length -= n;
			p += uuencode1(buf, p, n);
			added++;
			buf += n;
		}
		*p = 0;
		fputs((char *)obuf, out);
	}
	fputs(" \n", out);
	return (++added);
}

inline int
uudecode1(register char *from, register uchar *to)
{
	register int	length = DEC(*from++);
	int	save = length;

	assert(length <= 50);
	if (!length) return (0);
	while (length > 0) {
		if (length-- > 0)
			*to++ = (uchar)((DEC(from[0])<<2) | (DEC(from[1])>>4));
		if (length-- > 0)
			*to++ = (uchar)((DEC(from[1])<<4) | (DEC(from[2])>>2));
		if (length-- > 0)
			*to++ = (uchar)((DEC(from[2]) << 6) | DEC(from[3]));
		from += 4;
	}
	return (save);
}

private int
openOutput(int encode, char *file, FILE **op)
{
	char	buf[100];
	int	toStdout = streq(file, "-");

	switch (encode) {
	    case E_ASCII: 
	    case E_UUENCODE: 
	    	*op = toStdout ? stdout : fopen(file, "w");
		break;
	    case E_UUGZIP: 
		if (toStdout) {
			*op = popen("gunzip", "w");
		} else {
			sprintf(buf, "gunzip > %s", file);
	    		*op = popen(buf, "w");
		}
		break;
	    default:
	    	*op = NULL;
		return (-1);
	}
	return (encode == E_UUGZIP);
}

/*
 * get the specified revision.
 * The output file is passed in so that callers can redirect it.
 */
int
sccs_get(sccs *s, char *rev, char *iLst, char *xLst, int flags, char *printOut)
{
	serlist	*state = 0;
	ser_t	*slist = 0;
	delta	*d;
	int	print = 0;
	int	lines = 0;
	FILE	*out = 0;
	int	popened = 0;
	int	encoding = (flags&FORCEASCII) ? E_ASCII : s->encoding;
	int	error = 0;
	char	*base;
	BUF	(buf);

	debug((stderr, "get(%s, %s, %s, %s, %x, %s)\n",
	    s->sfile, rev, iLst, xLst, flags, printOut));
	unless (s->state & SOPEN) {
		fprintf(stderr, "get: couldn't open %s\n", s->sfile);
		return (-1);
	}
	unless (s->cksumok) {
		fprintf(stderr, "get: bad chksum on %s\n", s->sfile);
		return (-1);
	}
	unless (s->tree) {
		fprintf(stderr, "get: no/bad delta tree in %s\n", s->sfile);
		return (-1);
	}
	if (flags & EDIT) {
		int	f = (s->state & BRANCHOK) ? flags&FORCEBRANCH : 0;
		d = getedit(s, &rev, f);
		if (!d) {
			fprintf(stderr, "get: can't create revision %s in %s\n",
			    rev, s->sfile);
			s->state |= WARNED;
		}
	} else {
		d = findrev(s, rev);
		if (!d) {
			fprintf(stderr,
			    "get: can't find revision like %s in %s\n",
			rev, s->sfile);
			s->state |= WARNED;
		}
	}
	unless (d) return (-1);
	/* moved this up above the opens so that I can bail easily */
	unless (flags & SKIPGET) {
		slist = serialmap(s, d, flags, iLst, xLst, &error);
		if (error == 1) {
			assert(!slist);
			fprintf(stderr,
			    "Malformed include/exclude list for %s\n",
			    s->sfile);
			s->state |= WARNED;
			return (-1);
		}
		if (error == 2) {
			assert(!slist);
			fprintf(stderr,
		"Can't find specified rev in include/exclude list for %s\n",
			    s->sfile);
			s->state |= WARNED;
			return (-1);
		}
	}
	if (flags & EDIT) {
		int	fd;
		char	tmp[200];

		if (IS_WRITABLE(s) && !(flags & SKIPGET)) {
			fprintf(stderr,
			    "Writeable %s exists, skipping it.\n", s->gfile);
			s->state |= WARNED;
			if (slist) free(slist);
			return (-1);
		}
		if (!lock(s, 'z')) {
			fprintf(stderr, "get: can't zlock %s\n", s->gfile);
			if (slist) free(slist);
			return (-1);
		}
		if (!lock(s, 'p')) {
			fprintf(stderr, "get: can't plock %s\n", s->gfile);
			unlock(s, 'z');
			if (slist) free(slist);
			return (-1);
		}
		fd = open(s->pfile, 2, 0);
		sprintf(tmp, "%s %s %s %s",
		    d->rev, rev, getuser(), now());
		if (iLst) {
			strcat(tmp, " -i");
			strcat(tmp, iLst);
		}
		if (xLst) {
			strcat(tmp, " -x");
			strcat(tmp, xLst);
		}
		strcat(tmp, "\n");
		write(fd, tmp, strlen(tmp));
		close(fd);
		if (flags&SKIPGET) goto skip_get;
		unlink(s->gfile);
		popened = openOutput(encoding, s->gfile, &out);
		if (!out) {
			fprintf(stderr, "Can't open %s for writing\n",
			    s->gfile);
			unlock(s, 'p');
			unlock(s, 'z');
			if (slist) free(slist);
			return (-1);
		}
	} else if (flags&SKIPGET) {
		goto skip_get;
	} else if (!(flags & PRINT)) {
		if (IS_WRITABLE(s)) {
			fprintf(stderr, "Writeable %s exists\n", s->gfile);
			s->state |= WARNED;
			if (slist) free(slist);
			return (-1);
		}
		unlink(s->gfile);
		popened = openOutput(encoding, s->gfile, &out);
	} else {
		popened = openOutput(encoding, printOut, &out);
	}
	if ((s->state & RCS) && (flags & EXPAND)) flags |= RCSEXPAND;
	/* Think carefully before changing this */
	if (s->encoding != E_ASCII) {
		flags &= ~(REVNUMS|PREFIX_DATE|USER|EXPAND|RCSEXPAND);
	}
	state = allocstate(0, 0, 300);
	if (flags & MODNAME) base = basenm(s->gfile);
	seekto(s, s->data);
	while (next(buf, s)) {
		register char *e;

		if (buf[0] != '\001') {
			if (!print) continue;
			if (flags & (PREFIX_DATE|REVNUMS|USER|MODNAME)) {
				delta *tmp = sfind(s, print);
				
				if (flags&MODNAME)
					fprintf(out, "%s\t", base);
				if (flags&PREFIX_DATE)
					fprintf(out, "%.8s\t", tmp->sdate); 
				if (flags&USER)
					fprintf(out, "%s\t", tmp->user);
				if (flags&REVNUMS)
					fprintf(out, "%s\t", tmp->rev); 
			}
			e = buf;
			if (flags & EXPAND) {
				for (e = buf; *e != '%' && *e != '\n'; e++);
				if (*e == '%') {
					e = expand(s, d, buf);
				} else {
					e = buf;
				}
			}
			if (flags & RCSEXPAND) {
				char	*t;

				for (t = buf; *t != '$' && *t != '\n'; t++);
				if (*t == '$') {
					e = rcsexpand(s, d, e);
				}
			}
			if (encoding != E_ASCII) {
				uchar	obuf[50];
				int	n = uudecode1(e, obuf);

				fwrite(obuf, n, 1, out);
			} else {
				fnlputs(e, out);
			}
			lines++;
			continue;
		}

		debug2((stderr, "%.*s", linelen(buf), buf));
		changestate(state, buf[1], atoi(&buf[3]));
		print = printstate(state, slist);
	}
	debug((stderr, "GET done\n"));
	if (popened) {
		pclose(out);
	} else if (flags & PRINT) {
		unless (streq("-", printOut)) fclose(out);
	} else {
		fclose(out);
	}
	if (flags&EDIT) {
		if (IS_SEXEC(s)) 
			chmod(s->gfile, 0777);
		else
			chmod(s->gfile, 0666);
	} else if (!(flags&PRINT)) {
		if (IS_SEXEC(s)) 
			chmod(s->gfile, 0555);
		else
			chmod(s->gfile, 0444);
	}
skip_get:
	if (flags&EDIT) {
		unlock(s, 'z');
	}
	if (!(flags&SILENT)) {
		fprintf(stderr, "%s %s", s->gfile, d->rev);
		if (flags & EDIT)
			fprintf(stderr, " -> %s", rev);
		if (!(flags & SKIPGET)) fprintf(stderr, ": %d lines", lines);
		fprintf(stderr, "\n");
	}
	if (slist) free(slist);
	if (state) free(state);
	return (0);
}

/*
 * Return true if bad cksum
 */
#ifdef	USE_STDIO
private int
badcksum(sccs *s)
{
	char	buf[1024];
	register sum_t sum = 0;
	int	filesum;

	assert(s);
	seekto(s, 0);
	next(buf, s);
	if (sscanf(buf, "\001h%05d\n", &filesum) != 1) return(1);
	while (next(buf, s)) {
		char	*t = buf;

		while (*t) sum += *t++;
	}
	if (sum == filesum) s->cksumok = 1;
	debug((stderr,
	    "%s has %s cksum\n", s->sfile, s->cksumok ? "OK" : "BAD"));
	return (sum != filesum);
}
#else
private int
badcksum(sccs *s)
{
	register char *t;
	register char *end = s->mmap + s->size;
	register sum_t sum = 0;
	int	filesum;

	debug((stderr, "Checking sum from %x to %x (%d)\n",
	    s->mmap, end, end - s->mmap));
	assert(s);
	seekto(s, 0);
	filesum = atoi(&s->mmap[2]);
	debug((stderr, "File says sum is %d\n", filesum));
	t = s->mmap + 8;
	while (t < (end - 16)) {
		sum += t[0] + t[1] + t[2] + t[3] + t[4] + t[5] + t[6] + t[7]
		  + t[8] + t[9] + t[10] + t[11] + t[12] + t[13] + t[14] + t[15];
		t += 16;
	}
	while (t < end) sum += *t++;
	if (sum == filesum) s->cksumok = 1;
	debug((stderr,
	    "%s has %s cksum\n", s->sfile, s->cksumok ? "OK" : "BAD"));
	return (sum != filesum);
}
#endif /* !USE_STDIO */

/*
 * Check for bad characters in the file.
 */
#ifdef	USE_STDIO
private int
badchars(sccs *s)
{
	char	buf[1024];

	assert(s);
	seekto(s, 0);
	while (next(buf, s)) {
		register char	*t = buf;

		while (*t) {
			unless (isascii(*t) || ((*t == '\001') && t == buf)) {
				fprintf(stderr,
				    "admin: bad line in %s follows\n%s", buf);
				return (1);
			}
			t++;
		}
	}
	return (0);
}
#else
private int
badchars(sccs *s)
{
	register char *t = s->mmap;
	register char *end = s->mmap + s->size;

	assert(s);
	assert(s->mmap);
	if ((t[0] != '\001') || (t[1] != 'h') || (t[7] != '\n')) {
		fprintf(stderr, "admin: bad checksum format in %s\n",
		    s->sfile);
		return (1);
	}
	t += 8;
	for (t = s->mmap; t < end; t++) {
		unless (isascii(*t) || ((*t == '\001') && (t[-1] == '\n'))) {
			char	*r = t;

			while ((r > s->mmap) && (r[-1] != '\n')) r--;
			fprintf(stderr,
			    "admin: bad line in %s follows\n%.*s", 
			    s->sfile, linelen(r), r);
			return (1);
		}
	}
	return (0);
}
#endif /* !USE_STDIO */

private	FILE *not_on_stack_out;
private	sccs *not_on_stack_sccs;

void
putflags(delta *d)
{
	if (d->hostname) fputflag(d, 'h');
	if (d->pathname) fputflag(d, 'p');
	if (d->state) fputflag(d, 'S');
	if (d->zone) fputflag(d, 'w');
	if (d->kid) {
		putflags(d->kid);
	}
	if (d->siblings) {
		putflags(d->siblings);
	}
}

private void
fputflag(delta *d, char flag)
{
	char	*p, *c;
	char	buf[1024+20];

	assert(d);
	c = buf;
	*c++ = '\001';
	*c++ = 'f';
	*c++ = ' ';
	*c++ = flag;
	*c++ = ' ';
	for (p = d->rev; *p; *c++ = *p++);
	*c++ = ' ';
	switch (flag) {
	    case 'h':	assert(d->hostname);
	    		if (d->flags & NOHOST) return;
			p = d->hostname;
			break;
	    case 'p':	assert(d->pathname);
	    		if (d->flags & NOPATH) return;
			p = d->pathname;
			break;
	    case 'S':	assert(d->state);
			p = d->state;
			break;
	    case 'w':	assert(d->zone);
	    		if (d->flags & NOZONE) return;
			p = d->zone;
			break;
	    default:
	    	assert(!"Larry forgot something in fputflag");
	}
	for ( ; *p; *c++ = *p++) {
		if (*p == ' ') *c++ = '\\';
	}
	*c++ = '\n';
	*c = 0;
	fputsum(not_on_stack_sccs, buf, not_on_stack_out);
}

/*
 * The table is all here in order, just print it.
 */
private void
delta_table(sccs *s, FILE *out, int willfix)
{
	delta	*d;
	int	i;	/* used by EACH */
	int	first = willfix;
	char	buf[1024];
	symbol	*sym;
	int	bits = 0;

	fprintf(out, "\001hXXXXX\n");
	s->cksum = 0;
	for (d = s->table; d; d = d->next) {
		sprintf(buf, "\001s %05d/%05d/%05d\n",
		    d->added, d->deleted, d->same);
		if (first)
			fputs(buf, out);
		else
			fputsum(s, buf, out);
		first = 0;
		sprintf(buf, "\001d %c %s %s %s %d %d\n",
		    d->type, d->rev, d->sdate, d->user, 
		    d->serial, d->pserial);
		fputsum(s, buf, out);
		if (d->include) {
			fputsum(s, "\001i ", out);
			putserlist(s, d->include, out);
			fputsum(s, "\n", out);
		}
		if (d->exclude) {
			fputsum(s, "\001x ", out);
			putserlist(s, d->exclude, out);
			fputsum(s, "\n", out);
		}
		if (d->ignore) {
			fputsum(s, "\001g ", out);
			putserlist(s, d->ignore, out);
			fputsum(s, "\n", out);
		}
		EACH(d->comments) {
			fputsum(s, "\001c ", out);
			fputsum(s, d->comments[i], out);
			fputsum(s, "\n", out);
		}
		fputsum(s, "\001e\n", out);
	}
	fputsum(s, "\001u\n", out);
	EACH(s->usersgroups) {
		fputsum(s, s->usersgroups[i], out);
		fputsum(s, "\n", out);
	}
	fputsum(s, "\001U\n", out);
	sprintf(buf, "\001f e %d\n", s->encoding);
	fputsum(s, buf, out);
	not_on_stack_out = out;
	not_on_stack_sccs = s;
	putflags(s->tree);
	not_on_stack_out = 0;
	not_on_stack_sccs = 0;
	if (s->state & BRANCHOK) {
		fputsum(s, "\001f b\n", out);
	}
	if (s->state & YEAR4) bits |= X_YEAR4;
	if (s->state & RCS) bits |= X_RCSEXPAND;
	if (bits) {
		char	buf[40];

		sprintf(buf, "\001f x %u\n", bits);
		fputsum(s, buf, out);
	}
	if (s->defbranch) {
		fputsum(s, "\001f d ", out);
		fputsum(s, s->defbranch, out);
		fputsum(s, "\n", out);
	}
	EACH(s->flags) {
		fputsum(s, "\001f ", out);
		fputsum(s, s->flags[i], out);
		fputsum(s, "\n", out);
	}
	for (sym = s->symbols; sym; sym = sym->next) {
		if (!sym->name) continue;
		fputsum(s, "\001f s ", out);
		fputsum(s, sym->rev, out);
		fputsum(s, " ", out);
		fputsum(s, sym->name, out);
		fputsum(s, "\n", out);
	}
	/* Landing pad for more flags/symbols */
	fputsum(s, "\001f z ________________________"
	    "__________________________________________________\n", out);
	fputsum(s, "\001t\n", out);
	EACH(s->text) {
		fputsum(s, s->text[i], out);
		fputsum(s, "\n", out);
	}
	fputsum(s, "\001T\n", out);
}

/*
 * If we are trying to compare with expanded strings, do so.
 */
private inline int
expandnleq(sccs *s, delta *d, char *fbuf, char *sbuf, int flags)
{
	char	*e = fbuf;

	if (!(flags & (EXPAND|RCSEXPAND))) return 0;
	if (flags & EXPAND) {
		e = expand(s, d, e);
	}
	if (flags & RCSEXPAND) {
		e = rcsexpand(s, d, e);
	}
	return strnleq(e, sbuf);
}

/*
 * This is an expensive call but not as expensive as running diff.
 */
private int
has_diffs(sccs *s, int flags)
{
	FILE	*tmp = fopen(s->pfile, "r");
	serlist	*state = 0;
	ser_t	*slist = 0;
	int	print = 0, different;
	delta	*d;
	char	rev[40];
	BUF	(fbuf);
	char	sbuf[1024];

#define	RET(x)	different = x; goto out;

	unless (tmp) return (-1);
	if (fscanf(tmp, "%s", rev) != 1) {
		verbose((stderr,
		    "can't get revision info from %s\n", s->pfile));
		RET(-1);
	}
	fclose(tmp); tmp = 0;
	unless (d = findrev(s, rev)) {	
		verbose((stderr, "can't find %s in %s\n", rev, s->gfile));
		RET(-1);
	}
	unless (tmp = fopen(s->gfile, "r")) {
		verbose((stderr, "can't open %s\n", s->gfile));
		RET(-1);
	}
	assert(s->state & SOPEN);
	slist = serialmap(s, d, 0, 0, 0, 0);
	state = allocstate(0, 0, 300);
	seekto(s, s->data);
	while (next(fbuf, s)) {
		if (fbuf[0] != '\001') {
			if (!print) continue;
			unless (fnext(sbuf, tmp)) {
				debug((stderr, "diff because EOF on gfile\n"));
				RET(1);
			}
			unless (strnleq(fbuf, sbuf) ||
			    expandnleq(s, d, fbuf, sbuf, flags)) {
				debug((stderr, "diff because diff data\n"));
				RET(1);
			}
			debug2((stderr, "SAME %.*s", linelen(fbuf), fbuf));
			continue;
		}
		changestate(state, fbuf[1], atoi(&fbuf[3]));
		debug2((stderr, "%.*s\n", linelen(fbuf), fbuf));
		print = printstate(state, slist);
	}
	unless (fnext(sbuf, tmp)) {
		debug((stderr, "same\n"));
		RET(0);
	}
	debug((stderr, "diff because not EOF on both %s %s\n",
	    eof(s) ? "sfile done" : "sfile not done",
	    feof(tmp) ? "gfile done" : "gfile not done"));
	RET(1);
out:
	if (tmp) fclose(tmp);
	if (slist) free(slist);
	if (state) free(state);
	return (different);
}

/*
 * Apply the encoding to the gfile and leave it in tmpfile.
 */
int
deflate(sccs *s, char *tmpfile)
{
	FILE	*in, *out;
	char	cmd[1024];
	int	n;

	unless (out = fopen(tmpfile, "w")) return (-1); 
	switch (s->encoding) {
	    case E_UUENCODE:
		in = fopen(s->gfile, "r");
		n = uuencode(in, out);
		fclose(in);
		fclose(out);
		break;
	    case E_UUGZIP:
		sprintf(cmd, "gzip -nq4 < %s", s->gfile);
		in = popen(cmd, "r");
		uuencode(in, out);
		pclose(in);
		fclose(out);
		break;
	    default:
		assert("Bad encoding" == 0);
	}
	return(0);
}

/*
* Returns:
* 	-1 for some already bitched about error
*	0 if there were differences
*	1 if no differences
*/
private int
diff_gfile(sccs *s, pfile *pf, char *tmpfile)
{
	char	old[100];	/* the version from the s.file */
	char	new[100];	/* the new file, usually s->gfile */
	int	ret;

	debug((stderr, "diff_gfile(%s, %s)\n", pf->oldrev, s->gfile));
	if (s->encoding > E_ASCII) {
		sprintf(new, "/tmp/getU%d", getpid());
		if (IS_WRITABLE(s)) {
			if (deflate(s, new)) {
				unlink(new);
				return (-1);
			}
		} else {
/* XXX - I'm not sure when this would ever be used. */
			if (sccs_get(s, 0,0,0, FORCEASCII|SILENT|PRINT, new)) {
				unlink(new);
				return (-1);
			}
		}
	} else {
		strcpy(new, s->gfile);
	}
	sprintf(old, "/tmp/get%d", getpid());
	if (sccs_get(s, pf->oldrev, pf->iLst, pf->xLst,
	    FORCEASCII|SILENT|PRINT, old)) {
		unlink(old);
		return (-1);
	}
	ret = diff(old, new, D_DIFF, tmpfile);
	unlink(old);
	unless (streq(new, s->gfile)) unlink(new);	/* careful */
	switch (ret) {
	    case 0:	/* no diffs */
		return (1);
	    case 1:	/* diffs */
		return (0);
	    case 2:	/* diff ran into problems */
		fprintf(stderr, "Arrrrg.  Diff errored\n");
		return (-1);
	    default:	/* unknown? */
		fprintf(stderr, "Unknown exit from diff.\n");
		return (-1);
    	}
}

private void
unedit(sccs *s, int flags)
{
	unlink(s->pfile);
	unless (flags&SAVEGFILE)
		unlink(s->gfile);	/* Careful */
}

private void
pdiffs(char *gfile, char *left, char *right, FILE *diffs)
{
	int	first = 1;
	char	buf[1024];

	while (fnext(buf, diffs)) {
		if (first) {
			printf("===== %s %s vs %s =====\n", gfile, left, right);
			first = 0;
		}
		fputs(buf, stdout);
	}
	fclose(diffs);
}

void
free_pfile(pfile *pf)
{
	if (pf->iLst) free(pf->iLst);
	if (pf->xLst) free(pf->xLst);
	if (pf->user) free(pf->user);
	bzero(pf, sizeof(*pf));
}

/*
 * clean / unedit the specified file.
 *
 * Return codes are passed out to exit() so don't error on warnings.
 * If listing, don't do any cleans.
 */
int
sccs_clean(sccs *s, int flags)
{
	pfile	pf;
	char	tmpfile[50];

	unless (HAS_SFILE(s)) {
		verbose((stderr, "%s not under SCCS control\n", s->gfile));
		return (0);
	}
	unless (s->tree) return (-1);
	unless (HAS_PFILE(s)) {
		if (!IS_WRITABLE(s)) {
			verbose((stderr, "Clean %s\n", s->gfile));
			unlink(s->gfile);
			return (0);
		}
		fprintf(stderr, "%s writable but not edited?\n", s->gfile);
		return (1);
	}
	if (flags & UNEDIT) {
		unedit(s, flags);
		return (0);
	}
	unless (HAS_GFILE(s)) {
		verbose((stderr, "%s not checked out\n", s->gfile));
		return (0);
	}
	/*
	 * XXX - there is a bug somewhere that leaves stuff edited when it
	 * isn't.  I suspect some interactions with make, but I'm not
	 * sure.  The difference ends up being on a line with the keywords.
	 */
	if (access(s->gfile, W_OK)) {
		verbose((stderr, "%s edited but not writeable?\n", s->gfile));
		flags |= EXPAND;
		if (s->state & RCS) flags |= RCSEXPAND;
	}
	if (read_pfile("clean", s, &pf)) return (1);
	sprintf(tmpfile, "/tmp/diffg%d", getpid());
	switch (diff_gfile(s, &pf, tmpfile)) {
	    case 1: 		/* no diffs */
		verbose((stderr, "Clean %s\n", s->gfile));
		unedit(s, flags);
		free_pfile(&pf);
		unlink(tmpfile);
		return (0);
	    case 0: 		/* diffs */
		unless (flags & PRINT) {
			fprintf(stderr,
			    "%s has been modified, needs delta.\n", s->gfile);
		} else {
			pdiffs(s->gfile,
			    pf.oldrev, "edited", fopen(tmpfile, "r"));
		}
		free_pfile(&pf);
		unlink(tmpfile);
		return (2);
	    default:		/* error */
		fprintf(stderr,
		    "couldn't compute diffs on %s, skipping", s->gfile);
		free_pfile(&pf);
		unlink(tmpfile);
		return (1);
	}
}

/*
 * provide information about the editing status of a file.
 * XXX - does way too much work for this, shouldn't sccs init.
 *
 * Return codes are passed out to exit() so don't error on warnings.
 */
int
sccs_info(sccs *s, int flags)
{
	FILE	*f;
	char	buf[200];

	unless (HAS_SFILE(s)) {
		verbose((stderr, "%s not under SCCS control\n", s->gfile));
		return (0);
	}
	GOODSCCS(s);
	if (!HAS_PFILE(s)) {
		return (0);
	}
	unless (HAS_GFILE(s)) {
		verbose((stderr, "%s not checked out\n", s->gfile));
		return (0);
	}
	sprintf(buf, "%s:", s->gfile);
	printf("%-16s", s->gfile);
	f = fopen(s->pfile, "r");
	if (fgets(buf, sizeof(buf), f)) {
		char	*s;
		for (s = buf; *s && *s != '\n'; ++s) 
			;
		*s = 0;
		printf(buf);
	}
	fclose(f);
	switch (has_diffs(s, flags)) {
	    case 1:
		printf(" (modified, needs delta)\n");
		return (1);
	    case -1:
		fprintf(stderr,
		    "couldn't compute diffs on %s, skipping\n", s->gfile);
		return (1);
	    default: printf("\n");
	}
	return (0);
}

/*
 * Work backwards and count all the lines for this delta.
 */
private int
count_lines(delta *d)
{
	if (!d) return (0);
	return (count_lines(d->parent) + d->added - d->deleted);
}

/*
 * Check the file for binary data.
 */
int
ascii(FILE *f)
{
	char	buf[4096];
	int	n = fread(buf, 1, sizeof(buf), f);
	int	i;

	rewind(f);
	for (i = 0; i < n; ++i) {
		unless (isascii(buf[i])) return (0);
	}
	return (1);
}

/*
 * Open the input for a checkin/delta.
 * The set of options we have are:
 *	{empty, stdin, file} | {cat,gzip|uuencode}
 */
private int
openInput(sccs *s, int flags, FILE **inp)
{
	char	*file = (flags&EMPTY) ? "/dev/null" : s->gfile;
	char	buf[1024+80];

	unless (flags & EMPTY) {
		unless (HAS_GFILE(s)) {
			*inp = NULL;
			return (-1);
		}
	}
	switch (s->encoding) {
	    default:
	    case E_ASCII:
	    case E_UUENCODE:
		if (streq("-", file)) {
			*inp = stdin;
			return (0);
		}
		*inp = fopen(file, "r");
		if ((s->encoding == E_ASCII) && ascii(*inp)) return (0);
		s->encoding = E_UUENCODE;
		return (0);
	    case E_UUGZIP:
		/*
		 * Some very seat of the pants testing showed that -4 was
		 * the best time/space tradeoff.
		 */
		if (streq("-", file)) {
			*inp = popen("gzip -nq4", "r");
		} else {
			sprintf(buf, "gzip -nq4 < %s", file);
			*inp = popen(buf, "r");
		}
		return (1);
	}
}

/*
 * Check in initial gfile.
 */
/* ARGSUSED */
private int
checkin(sccs *s, int flags, delta *prefilled)
{
	FILE	*sfile, *gfile;
	delta	*n;
	int	added = 0;
	char	buf[1024];
	int	popened;

	assert(s);
	debug((stderr, "checkin %s %x\n", s->gfile, flags));
	unless (flags & NEWFILE) {
		verbose((stderr,
		    "%s not checked in, use -i flag.\n", s->gfile));
		unlock(s, 'z');
		if (prefilled) sccs_freetree(prefilled);
		return (1);
	}
	popened = openInput(s, flags, &gfile);
	unless (gfile) {
		perror(s->gfile);
		unlock(s, 'z');
		if (prefilled) sccs_freetree(prefilled);
		return (1);
	}
	if (exists(s->sfile)) {
		fprintf(stderr, "delta: lost checkin race on %s\n", s->sfile);
		if (gfile != stdin) {
			if (popened) pclose(gfile); else fclose(gfile);
		}
		return (1);
	}
	sfile = fopen(s->sfile, "w");
	if (prefilled) {
		n = prefilled;
	} else {
		n = calloc(sizeof(*n), 1);
		assert(n);
	}
	n->type = 'D';
	if (!n->sdate) date(n);
	if (!n->user) n->user = strdup(getuser());
	if (!n->hostname) n->hostname = gethost() ? strdup(gethost()) : 0;
	if (!n->pathname) n->pathname = strdup(fullname(s, 0));
	n->serial = s->nextserial++;
	n->next = s->table;
	s->table = n;
	if (!n->rev) n->rev = strdup("1.1");
	explode_rev(n);
	if (n->flags & BADREV) {
		unlock(s, 'z');
		if (prefilled) sccs_freetree(prefilled);
		fprintf(stderr, "checkin: bad revision: %s for %s\n", 
		    n->rev, s->sfile);
		return (1);
	}
	if (!n->comments) {
		sprintf(buf, "%s created on %s by %s",
		    s->gfile, n->sdate, n->user);
		n->comments = addLine(0, strdup(buf));
	}
	dinsert(s, n);
	s->numdeltas++;
	delta_table(s, sfile, 1);
	fputsum(s, "\001I 1\n", sfile);
	if (s->encoding > E_ASCII) {
		added = uuencode_sum(s, gfile, sfile);
	} else {
		while (fnext(buf, gfile)) {
			fputsum(s, buf, sfile);
			added++;
		}
	}
	fputsum(s, "\001E 1\n", sfile);
	end(s, n, sfile, flags, added, 0, 0);
#ifdef	PARANOID
	unless (flags&SAVEGFILE)
		rename(s->gfile, sccsXfile(s, 'G'));
#else
	unless (flags&SAVEGFILE)
		unlink(s->gfile);	/* Careful */
#endif
	Chmod(sfile, s->sfile, 0444);
	if (IS_GEXEC(s)) mkexecutable(sfile, s->sfile);
	if (gfile != stdin) {
		if (popened) pclose(gfile); else fclose(gfile);
	}
	fclose(sfile);
	unlock(s, 'z');
	return (0);
}

private int
checkrevs(char *file, delta *d)
{
	int	error = 0;

	if (d->flags & BADREV) {
		fprintf(stderr, "admin: bad rev '%s' in %s\n",
		    d->rev, file);
	}

	/*
	 * Make sure that the revision is well formed.
	 */
	if (!d->r[0] || !d->r[1] || (d->r[2] && !d->r[3])) {
		fprintf(stderr, "%s: bad revision %s (parent = %s)\n",
		    file, d->rev, d->parent?d->parent->rev:"Root");
		error = 1;
	}

	/*
	 * Two checks here.  
	 * If they are on the same branch, is the sequence numbering
	 * correct?  Handle 1.9 -> 2.1 properly.  
	 */
	if (!d->parent) goto done;
	/* If a x.y.z.q release, then it's trunk node should be x.y */
	if (d->r[2]) {
		delta	*p;

		for (p = d->parent; p && p->r[3]; p = p->parent);
		if (!p) {
			fprintf(stderr, "%s: rev %s not connected to trunk\n",
			    file, d->rev);
			error = 1;
		}
		if ((p->r[0] != d->r[0]) || (p->r[1] != d->r[1])) {
			fprintf(stderr, "%s: rev %s has incorrect parent %s\n",
			    file, d->rev, p->rev);
			error = 1;
		}
		/* if it's a x.y.z.q and not a .1, then check parent */
		if ((d->r[3] > 1) && (d->parent->r[3] != d->r[3]-1)) {
			fprintf(stderr, "%s: rev %s has incorrect parent %s\n",
			    file, d->rev, p->rev);
			error = 1;
		}
		/* OK */
		goto done;
	}
	/* If on the trunk and release numbers are the same, 
	 * then the revisions should be in sequence.
	 */
	if (d->r[0] == d->parent->r[0]) {
		if (d->r[1] != d->parent->r[1]+1) {
			fprintf(stderr, "%s: rev %s has incorrect parent %s\n",
			    file, d->rev, d->parent->rev);
			error = 1;
		}
	} else {
		/* Otherwise, this should be a .1 node */
		if (d->r[1] != 1) {
			fprintf(stderr, "%s: rev %s should be a .1 rev"
			    " since parent %s is a different release\n",
			    file, d->rev, d->parent->rev);
			error = 1;
		}
	}
	/* If there is a parent, make sure the dates increase. */
	if (d->parent && (d->date < d->parent->date)) {
		fprintf(stderr, 
		    "%s: time goes backwards between %s and %s\n",
		    file, d->rev, d->parent->rev);
		fprintf(stderr, "\t%s: %s    %s: %s -> %d seconds\n",
		    d->rev, d->sdate, d->parent->rev, d->parent->sdate,
		    (int)(d->date - d->parent->date));
		error = 1;
	}
	
done:	if (d->kid) error |= checkrevs(file, d->kid);
	if (d->siblings) error |= checkrevs(file, d->siblings);
	return (error);
}

private int
getval(char *arg)
{
	if (!arg || !*arg || !isdigit(*arg)) return (-1);
	return (atoi(arg));
}

/*
 * Accept dates in any of the following forms and convert to the 
 * SCCS form, filling out the delta struct.
 *	1998-01-11 20:00:00-08		(RCS -zLT)
 *	1990/01/12 04:00:00		(RCS UTC)
 *	98/01/11 20:00:00.000-8:00	(LMSCCS fully qualified)
 *	98/01/11 20:00:00.000     	(LMSCCS localtime)
 *	98/01/11 20:00:00		(SCCS, local time)
 */
private delta *
dateArg(delta *d, char *arg)
{
	char	*save = arg;
	struct	timeval tv;
	struct	timezone tz;
	struct	tm *tm;
	time_t	tt;
	char	tmp[50];
	extern	long timezone;
	int	year, month, day, hour, minute, second, msec, hwest, mwest;
	int	rcs = 0;

	if (!d) {
		d = (delta *)calloc(1, sizeof(*d));
		assert(d);
	}
	if (!arg || !*arg) { d->flags = DERROR; return (d); }
	year = getval(arg);
	if (year == -1) {
out:		fprintf(stderr, "sccs: can't parse date format %s at %s\n",
		    save, arg && *arg ? arg : "(unknown)");
		d->flags = DERROR;
		return (d);
	}
	if (year >= 100) {
		rcs++;
		if (year > 2069 || year < 1970) {
			goto out;
		}
		if (year < 2000) year -= 1900;
		if (year >= 2000) year -= 2000;
	}
#define	move(a)	{while (*a && isdigit(*a)) a++; if (*a && (*a++ == '-')) rcs++;}
#define	getit(what) {move(arg); what = getval(arg); if (what == -1) goto out;}
	getit(month);
	getit(day);
	getit(hour);
	getit(minute);
	getit(second);
	hwest = mwest = 0;
	msec = 0;
	if (*arg && arg[2] == '.') {
		getit(msec);
		if (arg[-1] == '-') {
			getit(hwest);
			getit(mwest);
		}
	} else if (*arg && arg[2] == '-') {
		getit(hwest);
		/* I don't know if RCS ever puts in the minutes, but I'll
		 * take 'em if they give 'em.
		 */
		if (*arg && arg[2] == ':') getit(mwest);
	} else if (rcs) {	/* then UTC */
		/* This is a bummer because we can't figure out in which
		 * timezone the delta was performed.
		 * So we assume here.
		 */
		gettimeofday(&tv, &tz);
		tt = (time_t)tv.tv_sec;
		tm = localtime(&tt);
		hwest = timezone/3600;
		mwest = timezone%3600;
	} else {		/* then local time */
		gettimeofday(&tv, &tz);
		tt = (time_t)tv.tv_sec;
		tm = localtime(&tt);
		hwest = timezone/3600;
		mwest = timezone%3600;
	}
	sprintf(tmp, "%02d/%02d/%02d %02d:%02d:%02d",
	    year, month, day, hour, minute, second);
	d->sdate = strdup(tmp);
	if (hwest || mwest) {
		sprintf(tmp, "%02d:%02d", hwest, mwest);
		d->zone = strdup(tmp);
	}
	/* Note: I don't fill in d->date; it's used only part of the time */
	return (d);
}
#undef	getit
#undef	move

delta *
userArg(delta *d, char *arg)
{
	char	*save = arg;

	if (!d) {
		d = (delta *)calloc(1, sizeof(*d));
		assert(d);
	}
	if (!arg || !*arg) { d->flags = DERROR; return (d); }
	while (*arg && (*arg++ != '@'));
	if (arg[-1] == '@') {
		arg[-1] = 0;
		d->hostname = strdup(arg);
	}
	d->user = strdup(save);		/* has to be after we null the @ */
	return (d);
}

/*
 * Note: null host names are accepted on purpose.
 */
delta *
hostArg(delta *d, char *arg)
{
	if (!d) {
		d = (delta *)calloc(1, sizeof(*d));
		assert(d);
	}
	if (!arg || !*arg) {
		d->flags |= NOHOST;
	} else {
		d->hostname = strdup(arg);
	}
	return (d);
}

/*
 * Note: null pathnames are accepted on purpose.
 */
delta *
pathArg(delta *d, char *arg)
{
	if (!d) {
		d = (delta *)calloc(1, sizeof(*d));
		assert(d);
	}
	if (!arg || !*arg) {
		d->flags |= NOPATH;
	} else {
		d->pathname = strdup(arg);
	}
	return (d);
}

/*
 * Take a string with newlines in it and split it into lines.
 * Note: null comments are accepted on purpose.
 */
delta *
commentArg(delta *d, char *arg)
{
	char	*tmp;

	if (!d) {
		d = (delta *)calloc(1, sizeof(*d));
		assert(d);
	}
	if (!arg) {
		/* don't call me unless you want one. */
		d->comments = addLine(0, strdup(""));
		return (d);
	}
	while (arg && *arg) {
		tmp = arg;
		while (*arg && *arg++ != '\n');
		if (arg[-1] == '\n') arg[-1] = 0;
		d->comments = addLine(d->comments, strdup(tmp));
	}
	return (d);
}

/*
 * Explode the rev.
 */
delta *
revArg(delta *d, char *arg)
{
	if (!d) {
		d = (delta *)calloc(1, sizeof(*d));
		assert(d);
	}
	d->rev = strdup(arg);
	explode_rev(d);
	return (d);
}

/*
 * Partially fill in a delta struct.  If the delta is null, allocate one.
 * Follow all the conventions used for delta creation such that this delta
 * can be added to the tree and freed later.
 */
delta *
sccs_parseArg(delta *d, char what, char *arg) 
{
	switch (what) {
	    case 'D':	/* any part of 1998/03/09 18:23:45.123-08:00 */
	    	return (dateArg(d, arg));
	    case 'U':	/* user or user@host */
	    	return (userArg(d, arg));
	    case 'H':	/* host */
	    	return (hostArg(d, arg));
	    case 'P':	/* pathname */
		return (pathArg(d, arg));
	    case 'C':	/* comments - one string, possibly multi line */
		return (commentArg(d, arg));
	    case 'R':	/* 1 or 1.2 or 1.2.3 or 1.2.3.4 */
		return (revArg(d, arg));
	    default:
	    	fprintf(stderr, "Unknown user arg %c ignored.\n", what);
		return (0);
	}
}

private char*
escape_sym(char *s)
{
	char	*buf = malloc(strlen(s) * 2);
	char	*t;

	for (t = buf; *s; *t++ = *s++) {
		if (*s == ' ') *t++ = '\\';
	}
	*t = 0;
	return (buf);

}

private int
addSym(sccs *s, char *name, char *rev, int replace)
{
	symbol	*sym;

	for (sym = s->symbols; sym; sym = sym->next) {
		if (streq(sym->name, name)) break;
	}
	if (!sym) {
		sym = (symbol*)calloc(1, sizeof(*sym));
		assert(sym);
		sym->name = strdup(name);
		sym->rev = strdup(rev);
		sym->next = s->symbols;
		s->symbols = sym;
	} else if (replace) {
		free(sym->name);
		free(sym->rev);
		sym->name = strdup(name);
		sym->rev = strdup(rev);
	} else {
		return (-1);
	}
	return (0);
			
}

private int
deleteSym(sccs *s, char *name)
{
	symbol	*sym, *b;
	int	foundone;
	int	n = 0;

	do {
		for (sym = s->symbols, b = 0, foundone = 0;
		    sym; b = sym, sym = sym->next) {
			if (streq(sym->name, name)) {
				free(sym->name);
				free(sym->rev);
				if (b) {
					b->next = sym->next;
				} else {
					s->symbols = sym->next;
				}
				free(sym);
				foundone++;
				n++;
				break;
			}
		}
	} while (foundone);
	return (n > 0);
}

#ifndef USE_STDIO
/*
 * Try and stuff the symbol into the landing pad,
 * return 1 if it fails.
 */
int
sccs_addSym(sccs *sc, int flags, char *s)
{
	symbol	*sym;
	char	*rev;
	int	force = 0;
	delta	*d;
	int	space;
	char	*t, *r;
	sum_t	sum;

	assert(sc->landingpad > sc->mmap);
	assert(sc->landingpad < sc->mmap + sc->data);
	s = escape_sym(s);
	if ((rev = strrchr(s, ':'))) {
		*rev++ = 0;
		force = 0;
	} else if ((rev = strrchr(s, '='))) {
		*rev++ = 0;
		force = 1;
	}
	for (sym = sc->symbols; sym; sym = sym->next) {
		if (streq(sym->name, s)) break;
	}
	/*
	 * We could rip the symbol out of the flags section and rewrite it.
	 * For now, I bail.
	 */
	if (sym) {
nope:		free(s);
		return (1);
	}
	unless (d = findrev(sc, rev)) goto nope;
	if (!rev || !*rev) rev = d->rev;
	/*
	 * ^Ad f s rev sym\n^Ad f &\n
	 */
	space = strlen(s) + strlen(rev) + 16;
	for (t = sc->landingpad; *t != '\n'; t++);
	if ((t - sc->landingpad) <= space) goto nope;
	sum = atoi(&sc->mmap[2]);
	t = &sc->landingpad[3];
#define	ADD(c)	{ sum -= *t; *t++ = c; sum += c; }
	ADD('s');
	ADD(' ');
	for (r = rev; *r; r++) ADD(*r);
	ADD(' ');
	/* XXX - escape spaces */
	for (r = s; *r; r++) ADD(*r);
	ADD('\n');
	ADD('\001');
	ADD('f');
	ADD(' ');
	ADD('&');
	sprintf(sc->mmap, "\001h%05u", sum);
	sc->mmap[7] = '\n';
	verbose((stderr, "admin: fast add symbol %s->%s in %s\n",
	    s, rev, sc->sfile));
	free(s);
	return (0);
}
#endif

/*
 * admin the specified file.
 *
 * Note: flag values are optional.
 *
 * XXX - this could do the insert hack for paths/users/whatever.
 * For large files, this is a win.
 */
int
sccs_admin(sccs *sc, int flags, admin *f,
	admin *u, admin *s, admin *p, char *text)
{
	FILE	*sfile = 0;
	int	error = 0, locked, i;
	char	*t;
	BUF	(buf);

	GOODSCCS(sc);
	unless (locked = lock(sc, 'z')) {
		fprintf(stderr, "admin: can't get lock on %s\n", sc->sfile);
		error = -1; sc->state |= WARNED;
out:		
		if (sfile) fclose(sfile);
		if (locked) unlock(sc, 'z');
		debug((stderr, "delta returns %d\n", error));
		return (error);
	}
#define	OUT	error = -1; sc->state |= WARNED; goto out;

	unless (HAS_SFILE(sc)) {
		verbose((stderr, "admin: no SCCS file: %s\n", sc->sfile));
		OUT;
	}

	if (flags & CHECKFILE) {
		if (checkrevs(sc->sfile, sc->tree) || badchars(sc)) {
			OUT;
		}
		verbose((stderr, "admin: %s checks out OK\n", sc->sfile));
		goto out;
	}

	/*
	 * "sym" means TOT.
	 * "sym:" means TOT.
	 * "sym:1.2" means that rev.
	 * "sym:1" or "sym:1.2.1" means TOT of that branch.
	 *
	 * We allow the same stuff for pathnames, et al.
	 */
	for (i = 0; s[i].flags; ++i) {
		char	*rev;
		char	*sym = escape_sym(s[i].thing);
		int	force = 0;

		if ((rev = strrchr(sym, ':'))) {
			*rev++ = 0;
			force = 0;
		} else if ((rev = strrchr(sym, '='))) {
			*rev++ = 0;
			force = 1;
		}
		if (s[i].flags & A_ADD) {
			delta	*d;

			unless (d = findrev(sc, rev)) {
				verbose((stderr,
				    "admin: can't find %s in %s\n",
				    rev, sc->sfile));
				error = 1; sc->state |= WARNED;
				free(sym);
				continue;
			}
			if (!rev || !*rev) rev = d->rev;
			if (isdigit(s[i].thing[0])) {
				fprintf(stderr,
				    "admin: %s: can't start with a digit.\n",
				    sym);
				error = 1; sc->state |= WARNED;
				free(sym);
				continue;
			}
			if (addSym(sc, sym, rev, force)) {
				fprintf(stderr,
				    "admin: symbol %s exists in %s, "
				    "use -s<sym>=<rev> to replace.\n",
				    sym, sc->sfile);
				error = 1; sc->state |= WARNED;
				free(sym);
				continue;
			}
			flags |= NEWCKSUM;
			verbose((stderr,
			    "admin: add symbol %s->%s in %s\n",
			    sym, rev, sc->sfile));
		} else {	/* delete it */
			if (deleteSym(sc, s[i].thing)) {
				verbose((stderr,
				    "admin: delete symbol %s from %s\n",
				    sym, sc->sfile));
				flags |= NEWCKSUM;
			} else {
				verbose((stderr,
				    "admin: symbol %s not found in %s\n",
				    sym, sc->sfile));
				error = 1; sc->state |= WARNED;
				continue;
			}
		}
		free(sym);
	}

	/*
	 * Pathnames.
	 */
	for (i = 0; p[i].flags; ++i) {
		char	*rev;
		delta	*d;
		
		if (p[i].flags & A_ADD) {
			unless (rev = strrchr(p[i].thing, ':')) {
				fprintf(stderr,
				    "admin: no revision in %s\n", p[i].thing);
				error = 1; sc->state |= WARNED;
				continue;
			}
			*rev++ = 0;
		} else {
			rev = p[i].thing;
		}
		unless (d = rfind(sc, rev)) {
			verbose((stderr,
			    "admin: no revision %s in %s\n", rev, sc->sfile));
			error = 1; sc->state |= WARNED;
			continue;
		}
		flags |= NEWCKSUM;
		assert(d);
		if (d->pathname && !(d->flags & DUPPATH)) {
			free(d->pathname);
			d->pathname = 0;
		}
		if (p[i].flags & A_ADD) {
			verbose((stderr, "admin: set path of %s to %s in %s\n",
			    d->rev, p[i].thing, sc->sfile));
			d->pathname = strdup(p[i].thing);
		} else {
			verbose((stderr, "admin: delete path of %s in %s\n",
			    d->rev, sc->sfile));
		}
	}

	if (text) {
		FILE	*desc;
		char	dbuf[200];

		if (!text[0]) {
			if (sc->text) {
				freeLines(sc->text);
				sc->text = 0;
				flags |= NEWCKSUM;
			}
			goto user;
		}
		desc = fopen(text, "r");
		if (!desc) {
			fprintf(stderr, "admin: can't open %s\n", text);
			error = 1; sc->state |= WARNED;
			goto user;
		}
		if (sc->text) {
			freeLines(sc->text);
			sc->text = 0;
		}
		while (fgets(dbuf, sizeof(dbuf), desc)) {
			sc->text = addLine(sc->text, strnonldup(dbuf));
		}
		fclose(desc);
		flags |= NEWCKSUM;
	}

user:	for (i = 0; u[i].flags; ++i) {
		flags |= NEWCKSUM;
		if (u[i].flags & A_ADD) {
			sc->usersgroups =
			    addLine(sc->usersgroups, strdup(u[i].thing));
		} else {
			unless (removeLine(sc->usersgroups, u[i].thing)) {
				verbose((stderr,
				    "admin: user/group %s not found in %s\n",
				    u[i].thing, sc->sfile));
				error = 1; sc->state |= WARNED;
			}
		}
	}

	/*
	 * b	turn on branching support (BRANCHOK)
	 * d	default branch (sc->defbranch)
	 * R	turn on rcs keyword expansion (RCSEXPAND)
	 * Y	turn on 4 digit year printouts 
	 *
	 * Anything else, just eat it.
	 */
	for (i = 0; f[i].flags; ++i) {
		int	add = f[i].flags & A_ADD;
		char	*v = &f[i].thing[1];

		flags |= NEWCKSUM;
		switch (f[i].thing[0]) {
		    char	buf[500];

		    case 'b':	if (add)
		    			sc->state |= BRANCHOK; 
		    		else 
					sc->state &= ~BRANCHOK;
				break;
		    case 'd':	if (sc->defbranch) free(sc->defbranch);
		    		sc->defbranch = *v ? strdup(v) : 0;
				break;
		    case 'R':	if (add) 
		    			sc->state |= RCS;
				else
		    			sc->state &= ~RCS;
				break;
		    case 'Y':	if (add) 
		    			sc->state |= YEAR4;
				else
		    			sc->state &= ~YEAR4;
				break;
		    default:	sprintf(buf, "%c %s", v[-1], v);
		    		if (add) {
		    			sc->flags =
					    addLine(sc->flags, strdup(buf));
				} else {
					unless (removeLine(sc->flags, buf)) {
						verbose((stderr,
						    "admin: flag %s not "
						    "found in %s\n",
						    buf, sc->sfile));
						error = 1; sc->state |= WARNED;
					}
				}
				break;
		}
	}

	if ((flags & NEWCKSUM) == 0) {
		goto out;
	}

	/*
	 * Do the delta table & misc.
	 */
	unless (sfile = fopen(sccsXfile(sc, 'x'), "w")) {
		fprintf(stderr, "admin: can't create %s: ", sccsXfile(sc, 'x'));
		perror("");
		OUT;
	}
	delta_table(sc, sfile, 0);

	seekto(sc, sc->data);
	assert(sc->state & SOPEN);
	while (next(buf, sc)) {
		fputsum(sc, buf, sfile);
	}
	fseek(sfile, 0L, SEEK_SET);
	fprintf(sfile, "\001h%05u\n", sc->cksum);
	/* 
	 * XXX - right here, I should look at old checksum and if no different,
	 * leave well enough alone.
	 */
#ifdef	PARANOID
	t = sccsXfile(sc, 'S');
	unlink(t);
	rename(sc->sfile, t);
#else
	unlink(sc->sfile);		/* Careful. */
#endif
	t = sccsXfile(sc, 'x');
	if (rename(t, sc->sfile)) {
		fprintf(stderr,
		    "delta: can't rename(%s, %s) left in %s\n",
		    t, sc->sfile, t);
		OUT;
	}
	Chmod(sfile, t, 0444);
	if (IS_GEXEC(sc) || IS_SEXEC(sc)) mkexecutable(sfile, t);
	goto out;
#undef	OUT
}


private void
doctrl(sccs *s, char *pre, int val, FILE *out)
{
	char	small[10];

	sertoa(small, val);
	fputsum(s, pre, out); 
	fputsum(s, small, out);
	fputsum(s, "\n", out);
}

#define	nextline(inc)	nxtline(s, &inc, &lines, &print, out, state, slist)

void
nxtline(sccs *s, int *ip, int *lp, int *pp, FILE *out, 
        register serlist *state, ser_t *slist)
{
	register short leave = 0;
	int c, print = *pp, incr = *ip, lines = *lp;
	register BUF	(buf);

	while (!eof(s)) {
		if (leave) { /* eat control lines first */
			peekc(c, s);
			if (c != '\001')  break;
		}
		if (!next(buf, s)) break;
		debug2((stderr, "[%d] ", lines));
		debug2((stderr, "G> %.*s", linelen(buf), buf));
		fputsum(s, buf, out);
		if (buf[0] != '\001') {
			if (print) {
				incr++; lines++;
				leave = 1;
			}
			continue;
		}
		changestate(state, buf[1], atoi(&buf[3]));
		print = printstate(state, slist); 
	}
	*ip = incr;
	*lp = lines;
	*pp = print;
}

int
delta_body(sccs *s, delta *n, FILE *diffs, FILE *out, int *ap, int *dp, int *up)
{
	serlist	*state = 0;
	ser_t	*slist = 0;
	int	print = 0;
	int	lines = 0;
	int	added = 0, deleted = 0, unchanged = 0;
	char	buf2[1024];

	*ap = *dp = *up = 0;
	/*
	 * Do the actual delta.
	 */
	seekto(s, s->data);
	slist = serialmap(s, n, 0, 0, 0, 0);	/* XXX - -gLIST */
	assert(s->state & SOPEN);
	state = allocstate(0, 0, 300);
	while (fnext(buf2, diffs)) {
		int	where;
		char	what;

newcmd:		
		if (scandiff(buf2, &where, &what) != 0) {
		    	fprintf(stderr, "delta: can't figure out '%s'\n", buf2);
			if (state) free(state);
			if (slist) free(slist);
			return (-1);
		}
		debug2((stderr, "where=%d what=%c\n", where, what));

#define ctrl(pre, val)	doctrl(s, pre, val, out)

		if (what != 'a') where--;
		while (lines < where) {
			nextline(unchanged);
		}
		switch (what) {
		    case 'a':
			ctrl("\001I ", n->serial);
		    	while (fnext(buf2, diffs)) {
				if (isdigit(buf2[0])) {
					ctrl("\001E ", n->serial);
					goto newcmd;
				}
				fputsum(s, &buf2[2], out);
				debug2((stderr, "INS %s", &buf2[2]));
				added++;
			}
			break;
		    case 'd':
			ctrl("\001D ", n->serial);
		    	while (fnext(buf2, diffs)) {
				if (isdigit(buf2[0])) {
					ctrl("\001E ", n->serial);
					goto newcmd;
				}
				nextline(deleted);
			}
			break;
		    case 'c':	
			ctrl("\001D ", n->serial);
			/* Toss the old stuff */
		    	while (fnext(buf2, diffs)) {
				if (!strcmp(buf2, "---\n")) break;
				nextline(deleted);
			}
			ctrl("\001E ", n->serial);
			/* add the new stuff */
			ctrl("\001I ", n->serial);
			while (fnext(buf2, diffs)) {
				if (isdigit(buf2[0])) {
					ctrl("\001E ", n->serial);
					goto newcmd;
				}
				fputsum(s, &buf2[2], out);
				debug2((stderr, "INS %s", &buf2[2]));
				added++;
			}
			break;
		}
		ctrl("\001E ", n->serial);
	}
	while (!eof(s)) {
		nextline(unchanged);
	}
	*ap = added;
	*dp = deleted;
	*up = unchanged;
	if (state) free(state);
	if (slist) free(slist);
	return (0);
}

/*
 * Initialize as much as possible from the file.
 * Don't override the revision if it is already set.
 */
private delta *
getInit(delta *d, char *file)
{
	FILE	*f;
	char	*s, *t;
	char	buf[500];

	if (!(f = fopen(file, "r"))) {
		return (d);
	}
/*
D 1.2 93/03/11 00:50:40.000 butthead 2 1  9/2/44
C added 4.x etc targets
S 
*/
	while (fgets(buf, sizeof(buf), f)) {
		chop(buf);
		if (buf[0] == 'D') {
			assert(buf[1] && isdigit(buf[2]));
			if (d) {
				sccs_freetree(d);
				return (0);
			}
			/* skip rev */
			for (s = &buf[2]; *s++ != ' '; );
			t = s;
			while (*s++ != ' ');	/* eat date */
			while (*s++ != ' ');	/* eat time */
			s[-1] = 0;
			d = sccs_parseArg(d, 'D', t);
			t = s;
			while (*s && (*s++ != ' '));	/* eat user */
			if (s[-1] == ' ') s[-1] = 0;
			d = sccs_parseArg(d, 'U', t);
			if (!d->hostname) d->flags |= NOHOST;
		} else if (buf[0] == 'C') {
			d->comments = addLine(d->comments, strdup(&buf[2]));
		} else if (buf[0] == 'P') {
			d->pathname = strdup(&buf[2]);
		} else if (buf[0] == 's') {
			d->state = strdup(&buf[2]);
		}
	}
	fclose(f);
	if (!d->comments) d->flags |= NOCOMMENTS;
	if (!d->pathname) d->flags |= NOPATH;
	if (!d->state) d->flags |= NOSTATE;
	return (d);
}

/*
 * Read the p.file and extract and return the old/new/user/{inc,excl} lists.
 *
 * Returns 0 if OK, -1 on error.  Warns on all errors.
 */
private int
read_pfile(char *who, sccs *s, pfile *pf)
{
	char	*iLst = malloc(512), *xLst = malloc(512);
	char	c1 = 0, c2 = 0;
	int	e;
	FILE	*tmp;
	char	date[10], time[10], user[40];

	bzero(pf, sizeof(*pf));
	unless (tmp = fopen(s->pfile, "r")) {
		fprintf(stderr, "delta: can't open %s\n", s->pfile);
		free(iLst); 
		free(xLst); 
		return (-1);
	}
	iLst[0] = xLst[0] = 0;
	e = fscanf(tmp, "%s %s %s %s %s -%c%s -%c%s",
	    pf->oldrev, pf->newrev, user, date, time, &c1, iLst, &c2, xLst);
	pf->user = strdup(user);
	strcpy(pf->date, date);
	strcat(pf->date, " ");
	strcat(pf->date, time);
	fclose(tmp);
	switch (e) {
	    case 5:	/* No extras, cool. */
		free(iLst); iLst = 0;
		free(xLst); xLst = 0;
		break;
	    case 7:	/* figure out if it was -i or -x */
	    	if (c1 == 'x') {
			free(xLst);
			xLst = iLst;
			iLst = 0;
		} else {
			assert(c1 == 'i');
			free(xLst);
			xLst = 0;
		}
		break;
	    case 9:
		assert(c1 == 'i');
		assert(c2 == 'x');
		break;
	    default:
		free(iLst); 
		free(xLst); 
		fprintf(stderr,
		    "%s: can't get revision info from %s\n", who, s->pfile);
		return (-1);
	}
	pf->iLst = iLst;
	pf->xLst = xLst;
	debug((stderr, "pfile(%s, %s, %s, %s, %s, %s)\n",
	    pf->oldrev, pf->newrev, user, pf->date, iLst, xLst));
	return (0);
}

#ifdef	ANSIC
private	jmp_buf jmp;

void	abort_ci() { longjmp(jmp, 1); }
#endif

/*
 * delta the specified file.
 *
 * Not implemented: 
 *	-g<list>
 *	-m<mr>
 */
int
sccs_delta(sccs *s, int flags, delta *prefilled, char *prsfile)
{
	FILE	*sfile = 0;	/* the new s.file */
	FILE	*diffs = 0;
	int	error = 0;
	char	*t;
	delta	*d = 0, *n;
	char	*tmpfile = tmpnam(0);
	int	added, deleted, unchanged;
	int	locked;
	char	buf2[1024];
	pfile	pf;

	assert(s);
	debug((stderr, "delta %s %x\n", s->gfile, flags));
	if (flags & NEWFILE) mksccsdir(s);
	bzero(&pf, sizeof(pf));
	unless(locked = lock(s, 'z')) {
		fprintf(stderr, "delta: can't get lock on %s\n", s->sfile);
		error = -1; s->state |= WARNED;
out:		
		sccs_freetree(prefilled);
		if (sfile) fclose(sfile);
		if (diffs) fclose(diffs);
		free_pfile(&pf);
		unlink(tmpfile);
		if (locked) unlock(s, 'z');
		debug((stderr, "delta returns %d\n", error));
		return (error);
	}
#define	OUT	{ error = -1; s->state |= WARNED; goto out; }

	if (prsfile) {
		unless (prefilled = getInit(prefilled, prsfile)) {
			fprintf(stderr, "delta: bad init file %s\n", prsfile);
			goto out;
		}
	}

	if (!HAS_SFILE(s) && HAS_GFILE(s)) {
		return (checkin(s, flags, prefilled));
	}

	if (!HAS_PFILE(s) && HAS_SFILE(s) && HAS_GFILE(s) && IS_WRITABLE(s)) {
		fprintf(stderr,
		    "delta: %s writable but not checked out?\n", s->gfile);
		s->state |= WARNED;
		OUT;
	}
	unless (IS_WRITABLE(s)) {
		unless (HAS_PFILE(s)) {
			verbose((stderr, "Clean %s (not edited)\n", s->gfile));
			unedit(s, flags);
			goto out;
		}
		verbose((stderr, "delta: %s locked but not writable?\n",
		    s->gfile));
		OUT;
	}
	unless (s->tree) {
		fprintf(stderr, "delta: bad delta table in %s\n", s->sfile);
		OUT;
	}

	/*
	 * OK, checking done, start the delta.
	 */
	if (read_pfile("delta", s, &pf)) OUT;
	unless (d = findrev(s, pf.oldrev)) {	
		fprintf(stderr,
		    "delta: can't find %s in %s\n", pf.oldrev, s->gfile);
		OUT;
	}
	switch (diff_gfile(s, &pf, tmpfile)) {
	    case 1: 		/* no diffs */
		if (flags & FORCE) break;	/* forced 0 sized delta */
		if (!(flags & SILENT))
			fprintf(stderr, "Clean %s (no diffs)\n", s->gfile);
		unedit(s, flags);
		goto out;
	    case 0: 		/* diffs */
		break;
	    default: OUT;
	}
	unless (diffs = fopen(tmpfile, "r")) {
		fprintf(stderr, "delta: can't open diff file %s\n", tmpfile);
		OUT;
	}
	if (flags & PRINT) {
		fprintf(stdout, "==== Changes to %s ====\n", s->gfile);
		while (fnext(buf2, diffs)) fputs(buf2, stdout);
		fputs("====\n\n", stdout);
		fseek(diffs, 0L, SEEK_SET);
	}

	/*
	 * Add a new delta table entry.  We'll come back and fix up
	 * the add/del/unch lines later.
	 */
	if (prefilled) {
		n = prefilled;
		prefilled = 0;
	} else {
		n = calloc(sizeof(*n), 1);
		assert(n);
	}
	if (pf.iLst) {
		assert(!n->include);
		n->include = getserlist(s, 0, pf.iLst, &error);
	}
	if (pf.xLst) {
		assert(!n->exclude);
		n->exclude = getserlist(s, 0, pf.xLst, &error);
	}
	if (error) OUT;
	n->type = 'D';
	n->serial = s->nextserial++;
	n->parent = d;
	n->next = s->table;
	s->table = n;
	if (flags & USEREV) {
		assert(n->rev);
	} else {
		n->rev = strdup(pf.newrev);
	}
	explode_rev(n);
	assert(d);
	n->pserial = d->serial;
	/* this glop removes zone dups after adding a date */
	if (!n->sdate) {
		char	*z;
		date(n);
		z = n->zone;
		n->zone = 0;
		addzone(n, z);
		free(z);
	}
	if (!n->user) n->user = strdup(getuser());
	if (!n->hostname && gethost()) addhost(n, gethost());
	if (!n->pathname) addpath(n, fullname(s, 0));
	if (!n->comments && !(flags & DONTASK) && !(n->flags & NOCOMMENTS)) {
		fprintf(stderr, 
		    "End comments with \".\" by itself, "
		    "blank line, or EOF.\n");
		fprintf(stderr, "%s %s>>  ", s->gfile, n->rev);
#ifdef	ANSIC
		if (setjmp(jmp)) {
			fprintf(stderr,
			    "\nCheck in aborted due to interrupt.\n");
			sccs_freetree(n);
			OUT;
		}
		signal(SIGINT, abort_ci);
#else
		sig(UNBLOCK, SIGINT);
#endif
		while (fnext(buf2, stdin)) {
			char	*t;

			if (buf2[0] == '\n' || streq(buf2, ".\n"))
				break;
			/* Null the newline */
			for (t = buf2; *t; t++);
			t[-1] = 0;
			n->comments = addLine(n->comments, strdup(buf2));
			fprintf(stderr,
			    "%s %s>>  ", s->gfile, n->rev);
		}
#ifndef ANSIC
		if (sig(CAUGHT, SIGINT)) {
			sig(BLOCK, SIGINT);
			fprintf(stderr, 
			    "\nCheck in aborted due to interrupt.\n");
			sccs_freetree(n);
			OUT;
		}
#endif
	}
	dinsert(s, n);
	s->numdeltas++;

	/*
	 * Do the delta table & misc.
	 */
	unless (sfile = fopen(sccsXfile(s, 'x'), "w")) {
		fprintf(stderr, "delta: can't create %s: ", sccsXfile(s, 'x'));
		perror("");
		OUT;
	}
	delta_table(s, sfile, 1);

	assert(d);
	if (delta_body(s, n, diffs, sfile, &added, &deleted, &unchanged)) {
		OUT;
	}
	end(s, n, sfile, flags, added, deleted, unchanged);
#ifdef	PARANOID
	t = sccsXfile(s, 'S');
	unlink(t);
	rename(s->sfile, t);
	t = sccsXfile(s, 'G');
	unless (flags&SAVEGFILE)
		rename(s->gfile, t);
#else
	unlink(s->sfile);		/* Careful. */
	unless (flags&SAVEGFILE)
		unlink(s->gfile);	/* Careful */
#endif
	t = sccsXfile(s, 'x');
	if (rename(t, s->sfile)) {
		fprintf(stderr,
		    "delta: can't rename(%s, %s) left in %s\n",
		    t, s->sfile, t);
		OUT;
	}
	Chmod(sfile, t, 0444);
	unlink(s->pfile);
	if (IS_GEXEC(s)) mkexecutable(sfile, t);
	goto out;
}

/*
 * Print the summary and go and fix up the top.
 */
private void
end(sccs *s, delta *n, FILE *out, int flags, int add, int del, int same)
{
	char	buf[100];

	unless (flags & SILENT) {
		char	*comma = "";
		int	lines = count_lines(n->parent);

		fprintf(stderr, "%s revision %s: ", s->gfile, n->rev);
		if (add) {
			fprintf(stderr, "%d inserted", add);
			comma = ", ";
		}
		if (del) {
			fprintf(stderr, "%s%d deleted", comma, del);
			comma = ", ";
		}
		if (same) {
			fprintf(stderr, "%s%d unchanged", comma, same);
			comma = ", ";
		}
		lines -= del;
		lines += add;
		fprintf(stderr, "%s%d total\n", comma, lines);
	}

	/*
	 * Now fix up the checksum and summary.
	 */
	fseek(out, 8L, SEEK_SET);
	sprintf(buf, "\001s %05d/%05d/%05d\n", add, del, same);
	fputsum(s, buf, out);
	fseek(out, 0L, SEEK_SET);
	fprintf(out, "\001h%05u\n", s->cksum);
}

/*
 * diffs - diff the gfile or the specified (or implied) rev
 */
int
sccs_diffs(sccs *s, char *r1, char *r2, int flags, char kind)
{
	FILE	*diffs = 0;
	char	*left, *right;
	char	*leftf, *rightf;
	char	tmpfile[30];
	char	diffFile[30];
	char	*columns;
	char	tmp2[32];
	char	buf[1024];
	pfile	pf;
	int	first = 1;
	char	spaces[80];

	bzero(&pf, sizeof(pf));
	GOODSCCS(s);
	if (kind == D_SDIFF) {
		unless (columns = getenv("COLUMNS")) columns = "80";
	}
	if (r1 && r2) {
		left = r1;
		right = r2;
	} else if (r1) {
		left = r1;
		right = HAS_PFILE(s) ? "edited" : 0;
	} else if (HAS_PFILE(s)) {
		if (read_pfile("diffs", s, &pf)) return (-1);
		left = pf.oldrev;
		right = "edited";
	} else {
		unless (HAS_GFILE(s)) {
			verbose((stderr,
			    "diffs: %s not checked out.\n", s->gfile));
			s->state |= WARNED;
			return (-1);
		}
		left = 0;
		right = "?";
	}
	unless (findrev(s, left)) {
		free_pfile(&pf);
		return (-2);
	}
	if (r2 && !findrev(s, r2)) {
		free_pfile(&pf);
		return (-3);
	}
	sprintf(tmpfile, "/tmp/diffget%d", getpid());
	if (sccs_get(s, left, pf.iLst, pf.xLst, flags|SILENT|PRINT, tmpfile)) {
		unlink(tmpfile);
		free_pfile(&pf);
		return (-1);
	}
	if (r2 || !HAS_GFILE(s)) {
		sprintf(tmp2, "%s-2", tmpfile);
		if (sccs_get(s, right, 0, 0, flags|SILENT|PRINT, tmp2)) {
			unlink(tmpfile);
			unlink(tmp2);
			free_pfile(&pf);
			return (-1);
		}
		leftf = tmpfile;
		rightf = tmp2;
	} else {
		tmp2[0] = 0;
		leftf = tmpfile;
		rightf = s->gfile;
	}
	if (!right) right = findrev(s, 0)->rev;
	if (!left) left = findrev(s, 0)->rev;
	if (kind == D_SDIFF) {
		int	i, c = atoi(columns);

		for (i = 0; i < c/2 - 18; ) spaces[i++] = '=';
		spaces[i] = 0;
		sprintf(buf, "%s -w%s %s %s", SDIFF, columns, leftf, rightf);
		diffs = fastPopen(buf, "r");
		if (!diffs) {
			unlink(tmpfile);
			if (tmp2[0]) unlink(tmp2);
			free_pfile(&pf);
			return (-1);
		}
		diffFile[0] = 0;
	} else {
		strcpy(spaces, "=====");
		sprintf(diffFile, "/tmp/diffs%d", getpid());
		diff(leftf, rightf, kind, diffFile);
		diffs = fopen(diffFile, "r");
	}
	while (fnext(buf, diffs)) {
		if (first) {
			printf("%s %s %s vs %s %s\n",
			    spaces, s->gfile, left, right, spaces);
			first = 0;
		}
		fputs(buf, stdout);
	}
	if (kind == D_SDIFF) {
		fastPclose(diffs);
	} else {
		fclose(diffs);
	}
	unlink(tmpfile);
	if (tmp2[0]) unlink(tmp2);
	if (diffFile[0]) unlink(diffFile);
	free_pfile(&pf);
	return (0);
}

/*
 * This routine walks the tree and builds up the host/path/state/zone
 * pointers, essentially "duping" them by sharing the pointer to the
 * pointer to the data in the delta that owns the pointer.
 *
 * Used for prs, sccslog, get, diffs, and smoosh.
 */
private void
inherit(int flags, delta *d)
{
	delta	*p;
	if (!d) return;
	if ((p = d->parent)) {
		if (!d->pathname && p->pathname) {
			d->pathname = p->pathname;
			d->flags |= DUPPATH;
		}
		if (!d->hostname && p->hostname) {
			d->hostname = p->hostname;
			d->flags |= DUPHOST;
		}
		if (!d->state && p->state) {
			d->state = p->state;
			d->flags |= DUPSTATE;
		}
		if (!d->zone && p->zone) {
			d->zone = p->zone;
			d->flags |= DUPZONE;
		}
	}
	if (flags & INHERIT_DATE) {
		d->date = sccs_date2time(d->sdate, d->zone, EXACT);
	}
	inherit(flags, d->kid);
	inherit(flags, d->siblings);
}

private void
do_prs(delta *start, delta *stop, int flags, FILE *out)
{
	int	i;	/* used by EACH */
	char	*z = "";

	if (!start) return;
	if (!stop || (stop != start)) {
		do_prs(start->kid, stop, flags, out);
		do_prs(start->siblings, stop, flags, out);
	}
	if (start->zone && (start->zone[0] != '-')) z = "+";
	fprintf(out, "D %s %s%s%s",
	    start->rev, start->sdate, z, start->zone ? start->zone : "");
	if (flags & TIMET) {
		fprintf(out, " (%lu)", DATE(start));
	}
	fprintf(out, " %s%s%s %d %d  ",
	    start->user, 
	    start->hostname ? "@" : "", 
	    start->hostname ?   start->hostname : "",
	    start->serial, start->pserial);
	fprintf(out, "%d/%d/%d\n", start->added, start->deleted, start->same);
	if (start->state) fprintf(out, "s %s\n", start->state);
	if (start->pathname) fprintf(out, "P %s\n", start->pathname);
	EACH(start->comments) {
		fprintf(out, "C %s\n", start->comments[i]);
	}
	fprintf(out, "------------------------------------------------\n");
}

/*
 * prs - a shallow imitation of the original.
 * XXX - should this have a time ordered output?
 * XXX - if so, then make it deal with the time zone.
 * XXX - an option to only print inherited stuff when it changes.
 */
int
sccs_prs(sccs *s, int flags, FILE *out)
{
	delta	*start = 0, *stop = 0;
	symbol	*sym;
	int	i;

	GOODSCCS(s);
	start = s->rstart ? s->rstart : s->tree;
	stop = s->rstop ? s->rstop : 0;
	unless (flags & SILENT) {
		EACH(s->text) {
			fprintf(out, "T %s\n", s->text[i]);
		}
		if (s->defbranch) fprintf(out, "B %s\n", s->defbranch);
		EACH(s->usersgroups) {
			fprintf(out, "U %s\n", s->usersgroups[i]);
		}
		for (sym = s->symbols; sym; sym = sym->next) {
			fprintf(out, "S %s:%s\n", sym->name, sym->rev);
		}
		fprintf(out,
		    "------------------------------------------------\n");
	}
	do_prs(start, stop, flags, out);
	return (0);
}

#ifndef	ANSIC

#ifndef	_NSIG
#define	_NSIG	32	/* XXX - might be wrong, probably OK */
#endif

private	int	caught[_NSIG];
private	struct	sigaction savesig[_NSIG];
private	void	catch(int sig) { caught[sig]++; }

/*
 * Signal theory of operation:
 *	block signals when initializing,
 *	unblock when we are going to do anything that takes a long time.     
 *	XXX - doesn't stack more than one deep.  Bad for a library.
 */
private int
sig(int what, int sig)
{
	struct sigaction sa;
	sigset_t sigs;

	assert(sig > 0);
	assert(sig < _NSIG);
	switch (what) {
	    case CATCH:
		bzero(&sa, sizeof(sa));
		sa.sa_handler = catch;
		sigaction(sig, &sa, &savesig[sig]);
		break;
	    case UNCATCH:
		sigaction(sig, &savesig[_NSIG], 0);
		break;
	    case BLOCK:
	    	sigemptyset(&sigs);
		sigaddset(&sigs, sig);
		sigprocmask(SIG_BLOCK, &sigs, 0);
		break;
	    case UNBLOCK:
	    	sigemptyset(&sigs);
		sigaddset(&sigs, sig);
		sigprocmask(SIG_UNBLOCK, &sigs, 0);
		break;
	    case PENDING:
	    	sigemptyset(&sigs);
	    	sigpending(&sigs);
		return (sigismember(&sigs, sig));
	    case CAUGHT:
		return (caught[sig]);
	    case CLEAR:
	    	caught[sig] = 0;
		break;
	}
	return (0);
}
#endif /* !ANSIC */

/* --------------------- module smoosh ---------------------------- */

private inline int
strmatch(char *s, char *t)
{
	if (!s && !t) return (1);
	if ((s && !t) || (!s && t) || strcmp(s, t)) return (0);
	return (1);
}

/*
 * Check user/host/pathname/date/lines/comments
 */
private inline int
samedelta(delta *l, delta *r)
{
	return ((l->added == r->added) &&
	    (l->deleted == r->deleted) &&
	    (l->same == r->same) &&
	    strmatch(l->user, r->user) &&
	    strmatch(l->hostname, r->hostname) &&
	    strmatch(l->pathname, r->pathname) &&
	    strmatch(l->sdate, r->sdate));
}

private sccs	*left, *right;	/* globals for graft. */

/*
 * Walk the tree, looking for conflicts.
 * A conflict is where we have matching parents but not
 * matching kids, and the kids have the same revision number.
 */
private void
conflicts(sccs *s, delta *d)
{
	delta	*o;

	if (!d) return;
	if (d->link) {
		conflicts(s, d->kid);
		conflicts(s, d->siblings);
		return;
	}
	/*
	 * So I'm the one that doesn't match.
	 * If any of my parent's links kids have my revision
	 * number, it's a conflict.
	 */
	assert(d->parent);
	for (o = d->parent->link->kid; o; o = o->siblings) {
		if (streq(d->rev, o->rev)) {
			o->flags |= BRANCH;
			markbranches(o);
			move(s, o);
			graft(d);
		}
	}
}

/*
 * Walk through and check out the delta from one file and check it in
 * to the other.
 */
private void
graft(delta *d)
{
	FILE	*f;
	delta	*pre = 0;

	if (!d) return;
	if (!access(right->gfile, F_OK)) {
		fprintf(stderr,
		    "Can't smoosh %s %s into %s because %s exists.\n",
		    left->sfile, d->rev, right->sfile, right->gfile);
		 return;
	}
	if (sccs_get(right, d->parent->rev, 0, 0, SILENT|EDIT|SKIPGET, "-")) {
		sccs_whynot("get", right);
	}
	if (sccs_get(left, d->rev, 0, 0, SILENT|PRINT, right->gfile)) {
		sccs_whynot("get", left);
	}
	f = fopen(SCCSTMP, "w");
	left->rstart = left->rstop = d;
	if (sccs_prs(left, 0, f)) {
		sccs_whynot("prs", left);
	}
	fclose(f);
	pre = calloc(1, sizeof(*pre));
	pre->rev = strdup(d->rev);
	sccs_restart(right);
	if (sccs_delta(right, BRANCHOK|USEREV, pre, SCCSTMP)) {
		sccs_whynot("delta", right);
	}
	unlink(SCCSTMP);
	graft(d->kid);
	graft(d->siblings);
}

/*
 * Walk through and mark all deltas that are not on the
 * same branch as their parent.
 */
private void
markbranches(delta *d)
{
	if (!d) return;
	assert(d->parent);
	if (!samebranch(d, d->parent)) d->flags = BRANCH;
	markbranches(d->kid);
	markbranches(d->siblings);
}

/*
 * Move this delta off to the side, renaming its symbols as well.
 */
private void
move(sccs *s, delta *d)
{
	delta *t;
	symbol	*sym;
	int	c;
	char	buf[50];

	if (!d) return;
	if (d->flags & BRANCH) {
		/*
		 * Find the next unused branch.
		 */
		for (t = d->parent; t->r[2]; t = t->parent);
		c = 1;
		do {
			sprintf(buf, "%s.%d.1", t->rev, c);
			c++;
		} while (rfind(s, buf));
	} else {
		delta	*p = d->parent;

		/*
		 * XXX - I think I must be on a branch 100% of the time,
		 * since we are being pushed off to the side.
		 */
		if (p->r[2]) {
			sprintf(buf, "%d.%d.%d.%d",
			    p->r[0], p->r[1], p->r[2], p->r[3]+1);
		} else {
			sprintf(buf, "%d.%d", p->r[0], p->r[1]+1);
		}
	}
	fprintf(stderr, "Rename %s -> %s in %s\n", d->rev, buf, right->sfile);
	/*
	 * Rewrite the symbols.
	 *
	 * XXX - I need to think about rewriting symbols when I am
	 * shifting a LOD.  After I implement LOD.
	 * For now, just rewrite stuff that matches exactly.
	 */
	for (sym = right->symbols; sym; sym = sym->next) {
		if (streq(sym->rev, d->rev)) {
			free(sym->rev);
			sym->rev = strdup(buf);
			printf("Remap %s\n", sym->name);
		}
	}
	free(d->rev);
	d->rev = strdup(buf);
	explode_rev(d);
	move(s, d->kid);
	move(s, d->siblings);
}

/*
 * The two deltas passed in are the same.   Link them.
 * For each kid/sibling, try linking that one as well.
 */
private void
linktree(delta *l, delta *r)
{
	delta	*a, *b;

	if (l->link || r->link) return;		/* insurance */
	l->link = r;
	r->link = l;

	/*
	 * What I want is a loop like this:
	 *
	 *	foreach lk (l->kid, @l->siblings) {
	 *		foreach rk (r->kid, @r->siblings) {
	 *			if samedelta(lk, rk) ....
	 *		}
	 *	}
	 */
	for (a = l->kid; a; a = a->siblings) {
		if (a->link) continue;	/* XXX - happens? */
		for (b = r->kid; b; b = b->siblings) {
			if (b->link) continue;
			if (samedelta(a, b)) {
				linktree(a, b);
				break;
			}
		}
	}
}

int
sccs_smoosh(char *lfile, char *rfile)
{

	left = sccs_init(lfile, 0);
	right = sccs_init(rfile, 0);
	if (!left || !HAS_SFILE(left) || !right || !HAS_SFILE(right)) {
		return(100);
	}
	if (!samedelta(left->tree, right->tree)) {
		return(101);
	}
	linktree(left->tree, right->tree);
	conflicts(left, left->tree);
	sccs_free(left);
	sccs_free(right);
	left = right = 0;
	return(0);
}

private void
pdelta(sccs *s, delta *d, int dopath, FILE *out)
{
	d->sdate[8] = '-';
	fprintf(out, "%s%s%s:%s:%s%s%s",
	    d->user,
	    d->hostname ? "@" : "", d->hostname ? d->hostname : "",
	    dopath ? (d->pathname ? d->pathname : s->sfile) : "",
	    d->sdate, d->zone ? "-" : "", d->zone ? d->zone : "");
	d->sdate[8] = '0';
}

/*
 * Print user[@host]:path:date
 *
 * If we are doing this for resync, all print the list of leaves.
 * Note: if there is only one delta, i.e., 1.1 is also a leaf, I don't
 * print it twice.
 */
void
sccs_id(sccs *s, int flags, FILE *out)
{
	delta	*d = s->tree;
	int	dopath;
	int	first = 1;

	unless (flags & TOP) {
		pdelta(s, s->tree, 1, out);
		fprintf(out, "\n");
		return;
	}
	for (d = s->table; d; d = d->next) {
		if (!d->kid) {
			dopath = d->pathname &&
		(!s->tree->pathname || strcmp(d->pathname, s->tree->pathname));
			if (!first++) printf(" ");
			pdelta(s, d, dopath, out);
		}
	}
	fprintf(out, "\n");
}

#ifdef	DEBUG
debug_main(char **av)
{
	fprintf(stderr, "---");
	do {
		fprintf(stderr, " %s", av[0]);
		av++;
	} while (av[0]);
	fprintf(stderr, " ---\n");
}
#endif
