/* 
 * smoosh - join two SCCS files that have at least rev 1.1 in common.
 *
 * The two files are known as "left" and "right" and we smoosh left into
 * right.  This means that rights revisions may get renamed.
 *
 * Each file is turned into a graph.  The two graphs are matched delta for
 * delta, until all are matched or there are leftovers.  The matched deltas
 * are marked as such.
 *
 * Exit codes:
 *	0 - files were related, exactly the same.
 *	1 - only left had new deltas, nothing done.
 *	2 - only right had new deltas, nothing done.
 *	100 - some unknown error
 *	101 - there was no common root
 *	102 - one of the files had a p.file (edited)
 *	103 - one of the files had a z.file (locked)
 *	104 - one of the files had a r.file (unresolved)
 *	
 * Copyright (c) 1998 Larry McVoy.  All rights reserved.
 */
#include "sccs.h"
WHATSTR("@(#)smoosh.c 1.7");

int
main(int ac, char **av)
{
	int	err;
	char	*name, *name2;

	debug_main(av);
	name = sfileFirst("smoosh", &av[optind], 0);
	name2 = sfileNext();
	if (!name || !name2 || sfileNext()) {
		fprintf(stderr, "Usage: smoosh leftfile rightfile\n");
		exit(100);
	}
	err = sccs_smoosh(name, name2);
	sfileDone();
#ifndef	NOPURIFY
	purify_list();
#endif
	return(err);
}
